/*
 *  IKEv2.c  -  code for IKEv2 routines / interface
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include"common.h"

#include <stdlib.h>
#include <assert.h>

#ifdef  HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif //HAVE_NETINET_IN_H
#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif // HAVE_WINSOCK_H

#include <openssl/rand.h>

#include "IKEv2.h"
#include "state.h"
#include "session.h"

// IKEv2 implementation

/**
 * @brief Begin new IKEv2 session
 *
 * @param *i2 IKEv2 context, which contains list of all IKEv2 sessions
 * @param **session new IKEv2 session
 * @param type responder or initiator session
 *
 * @return  0 on success, 1 otherwise
 */
int IKEv2BeginSession( struct ikev2_ctx *i2, struct IKEv2Session **session, uint32_t type )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Begin Session...\n" );
  if( !session || ( type != IKEv2_STY_INITIATOR && type != IKEv2_STY_RESPONDER ) )
    return IKEv2_RET_ERRPARAMETERS;            // wrong parameters
  uint64_t MySPI=0;                             // generate SPI
  RAND_bytes( (uint8_t *)&MySPI, sizeof( uint64_t ) );
  if( !( *session = NewSession( i2, type == IKEv2_STY_INITIATOR ? MySPI : 0,
                          type == IKEv2_STY_RESPONDER ? MySPI : 0, type ) ) )
    return IKEv2_RET_ERRMEMALLOC;       // can't alloc session structure
  return IKEv2_RET_OK;
}

/**
 * @brief End an existing  IKEv2 session
 *
 * @param *i2 IKEv2 context which contains list of all sessions
 * @param *session sesion to delete
 *
 * @return always success (0)
 */
int IKEv2EndSession( struct ikev2_ctx *i2, struct IKEv2Session *session )
{
  xlogf( I2L_DBG, "EAP-IKEv2: End Session...\n" );
  if( session )
    DeleteSession( i2, session );
  return IKEv2_RET_OK;
}

/**
 * @brief 
 * Check, if for a packet a new session should be initialized
 * otherwise find the existing session
 *
 * @param *i2 IKEv2 context
 * @param *in ptr to message
 */
void *IKEv2GetSession(struct ikev2_ctx *i2, uint8_t *in )
{
  struct IKEv2Header *hdr = (struct IKEv2Header *)in;
  if( hdr->ExchangeType == IKEv2_EXT_IKE_SA_INIT && hdr->ResponderSPI == 0 &&
      !FindISession( i2, ntoh64( hdr->InitiatorSPI ) ) )        // create a new session?
  {
    struct IKEv2Session *se;
    IKEv2BeginSession( i2, &se, IKEv2_STY_RESPONDER );
    se->iSPI = ntoh64( hdr->InitiatorSPI );
    return (void *)se;
  }
  if( hdr->Flags & IKEv2_HDF_INITIATOR )              // session already exists?
    return FindISession( i2, ntoh64( hdr->InitiatorSPI ) );
  else
    return FindRSession( i2, ntoh64( hdr->ResponderSPI ) );
}

/**
 * @brief Process IKEv2 Message
 *
 * @param *i2       IKEv2 context data
 * @param *in       incoming data
 * @param **out     outcoming data
 * @param *olen     outcoming data length
 * @param *session  current session data
 *
 * @return 0 on success
 */
int IKEv2ProcessMsg(const struct ikev2_ctx *i2, uint8_t *in, uint8_t **out, uint32_t *olen, struct IKEv2Session *session)
{
  assert(session && i2);

  if( in )  // received a packet --> process it
  {
    struct IKEv2Header *hdr = (struct IKEv2Header *)in;   // check header info

    xlogf( I2L_DBG, "EAP-IKEv2: Got an IKEv2 message (%Ld/%Ld).\n", ntoh64(hdr->InitiatorSPI), ntoh64(hdr->ResponderSPI) );

    if( session->Type == IKEv2_STY_INITIATOR && !( hdr->Flags & IKEv2_HDF_INITIATOR ) )
    {   // i am initiator
      if( session->rSPI == 0 )
        session->rSPI = ntoh64( hdr->ResponderSPI );
      if( session->rSPI != ntoh64( hdr->ResponderSPI ) )
      {
        xlogf( I2L_ERR, "EAP-IKEv2: Error: Session and Packet SPIs are not the same.\n" );
        *out = NULL;
        *olen = 0;
        session->Status = IKEv2_SST_FAILED;
        return IKEv2_RET_ERRWRONGSPI;
      }
      DoIState( i2, in, out, olen, session);
    }
    else
    { // i am responder
      if( ( ntoh64( hdr->ResponderSPI ) == session->rSPI || hdr->ResponderSPI == 0 ) &&
          session->Type == IKEv2_STY_RESPONDER && ( hdr->Flags & IKEv2_HDF_INITIATOR ) )
      {
        if( hdr->ResponderSPI == 0 && ( hdr->MessageID != 0 || ( hdr->Flags & IKEv2_HDF_RESPONSE ) ) )
        {         // correct first msg?
          xlogf( I2L_ERR, "EAP-IKEv2: Error: Syntax error in msg, ignoring msg.\n" );
          *out = NULL;
          *olen = 0;
          session->Status = IKEv2_SST_FAILED;
          return IKEv2_RET_ERRWRONGSPI;
        }
        
        if( session->iSPI == 0 )
          session->iSPI = ntoh64( hdr->InitiatorSPI );
        if( session->iSPI != ntoh64( hdr->InitiatorSPI ) )
        {
          xlogf( I2L_ERR, "EAP-IKEv2: Error: Session and Packet SPIs are not the same.\n" );
          *out = NULL;
          *olen = 0;
          session->Status = IKEv2_SST_FAILED;
          return IKEv2_RET_ERRWRONGSPI;
        }
        DoRState( i2, in, out, olen, session);
      }
      else
      { // wrong instance
        xlogf( I2L_ERR, "EAP-IKEv2: Wrong instance or wrong SPIs.\n" );
        *out = NULL;
        *olen = 0;
        session->Status = IKEv2_SST_FAILED;
        return IKEv2_RET_ERRWRONGSPI;
      }
    }
  }
  else      // received no packet --> send first packet
  {
    if( session->State == IKEv2_STATE_CONNECTED && session->Type == IKEv2_STY_RESPONDER )
      DoRState( i2, NULL, out, olen, session);
    else
      DoIState( i2, NULL, out, olen, session);
  }


  return IKEv2_RET_OK;
}
