/*
 *  auth.c  -  functions for authentication and certificates
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include"common.h"

#include <assert.h>
#include<string.h>
#include<openssl/x509.h>
#include<openssl/pem.h>
#include<openssl/rsa.h>

#include"message.h"  //CertList definition
#include"auth.h"
#include"prf.h"
#include"integ.h"
#include"payload.h"
#include "enums.h"


// private prototypes:
uint16_t GenerateCertAuthData(EVP_PKEY *pkey,uint8_t *data,uint32_t dlen,uint8_t *dest);
int CertAuthenticate(const struct ikev2_ctx *i2, struct CertList *cert, 
		uint8_t *adata,uint32_t alen, uint8_t *sigbuf, uint32_t siglen);

#if ! HAVE_OPENSSL_ADD_ALL_ALGORITHMS
#warning Using own implemetation of OpenSSL_add_all_algorithms()
#undef OpenSSL_add_all_algorithms
void OpenSSL_add_all_algorithms(void);

void OpenSSL_add_all_algorithms(void)
{
    OpenSSL_add_all_ciphers();
    OpenSSL_add_all_digests();
}
#endif

// Authentication functions

/**
 * @brief generate data for auth payload
 *
 * @param *i2      ikev2 context data
 * @param type     type of auth  
 * @param *session ptr to sessiot struct
 * @param *id      ptr to id data
 * @param idlen    id data length
 * @param *p       ptr to dest buffer
 *
 * @return legth of auth data
 */
uint16_t GenerateAuthData( const struct ikev2_ctx *i2, uint8_t type, struct IKEv2Session *session, uint8_t *id, uint16_t idlen, uint8_t *p )
{
  uint32_t blen, len = 0;
  uint8_t *buf, *buf2;
  uint8_t *data;
  
  uint32_t nlen = session->Type == IKEv2_STY_INITIATOR ? BN_num_bytes( session->Nr ) :
                                                          BN_num_bytes( session->Ni );
  uint32_t dlen = getDigestLength( GetTransformID( IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA,
      session->UsedTransforms ) );

  if( !( data = (uint8_t *)malloc( session->fmlen + nlen + dlen ) ) )
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for auth data.\n" );
    return 0;
  }
  memcpy( data, session->fmsg, session->fmlen );  // generate data block to sign
  BN_bn2bin( session->Type == IKEv2_STY_INITIATOR ? session->Nr : session->Ni, data + session->fmlen );
  prf( session->Type == IKEv2_STY_INITIATOR ? session->SK_pi : session->SK_pr,
      GetIntegKeyLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
      session->UsedTransforms ) ), id, idlen, &buf, &blen,
      GetTransformID( IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA, session->UsedTransforms ) );
  if ( blen == 0 ) {free(data);return 0;}
  memcpy( data + session->fmlen + nlen, buf, blen );
  
  free( buf );
  buf=NULL;

  switch( type )     // which auth method?
  {
  case IKEv2_AMT_DSS_DS:
  case IKEv2_AMT_RSA_DS:
    if( !i2->rpk ) {
      xlogf( I2L_ERR, "EAP-IKEv2: Error: Private key unavailable.\n" );
      free( data );
      data=NULL;
      return 0;
    }
    len=GenerateCertAuthData(i2->rpk,data,session->fmlen+nlen+dlen,p);
    free(data);
    data=NULL;
    break;

  case IKEv2_AMT_SKMIC:
    assert(session->pwd && session->pwdlen);
    prf( session->pwd, session->pwdlen ,(const uint8_t*) "Key Pad for EAP-IKEv2", 21, &buf, &blen, GetTransformID(
      IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA, session->UsedTransforms ) );
    if( blen )
    {
      prf( buf, blen, data, session->fmlen + nlen + dlen, &buf2, &len, GetTransformID(
        IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA, session->UsedTransforms ) );
      free( data );
      data=NULL;
      free( buf );
      buf=NULL;
      if( len )
      {
        memcpy( p, buf2, len );
        free( buf2 );
	buf2=NULL;
      }
    }
    break;
  }
  return (uint16_t)len;
}



/**
 * @brief Authenticate a peer
 *
 * @param *i2 IKEv2 context
 * @param *cert list of raw certificate data
 * @param type auth payload type
 * @param *data auth data
 * @param dlen auth data length
 * @param *session current IKEv2 session
 * @param *idData peer identification data
 * @param idlen idData length
 *
 * @return 1 on success, 0 otherwise
 */

int Authenticate( const struct ikev2_ctx *i2,struct CertList *cert, uint8_t type, uint8_t *data, uint16_t dlen, struct IKEv2Session *session,
     uint8_t *idData, uint16_t idlen)
{
  struct sharedSecList *list = i2->sslist;
  uint32_t blen, len = 0;
  uint8_t *buf, *buf2;
  uint8_t *adata;
  int res=0;
  uint8_t *id=IKEv2_GET_ID(idData);
  int idt=IKEv2_GET_IDTYPE(idData);


  switch(type) {
      case IKEv2_AMT_RSA_DS:
	  xlogf( I2L_DBG, "EAP-IKEv2: Trying to authenticate (rsa)....\n" );
	  break;
      case IKEv2_AMT_SKMIC:
	  xlogf( I2L_DBG, "EAP-IKEv2: Trying to authenticate (shared secret)....\n" );
	  break;
      case IKEv2_AMT_DSS_DS:
	  xlogf( I2L_DBG, "EAP-IKEv2: Trying to authenticate (dss)....\n" );
	  break;
      default:
	  xlogf( I2L_DBG, "EAP-IKEv2: Trying to authenticate (unknown type)....\n" );
  }

  xlogf(I2L_DBG,"EAP-IKEv2: trying auth: %s\n",bin2str(id,idlen-sizeof(struct IKEv2IdentificationPayloadBody)));
  
  uint32_t nlen = session->Type == IKEv2_STY_INITIATOR ? BN_num_bytes( session->Ni ) :
                                                          BN_num_bytes( session->Nr );
  uint32_t adlen = getDigestLength( GetTransformID( IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA,
      session->UsedTransforms ) );

  if( !( adata = (uint8_t *)malloc( session->rmlen + nlen + adlen ) ) )
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for auth data.\n" );
    return 0;
  }
  memcpy( adata, session->rmsg, session->rmlen );  // generate data block to sign
  BN_bn2bin( session->Type == IKEv2_STY_INITIATOR ? session->Ni : session->Nr, adata + session->rmlen );
  prf( session->Type == IKEv2_STY_INITIATOR ? session->SK_pr : session->SK_pi,
      GetIntegKeyLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
      session->UsedTransforms ) ), idData, idlen , &buf, &blen,
      GetTransformID( IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA, session->UsedTransforms ) );
  
  if ( blen == 0 ) {free(adata);return 0;}
  memcpy( adata + session->rmlen + nlen, buf, blen );
  free( buf );
  buf=NULL;

  int exist;
  uint8_t *pwd;
  int pwdlen;
  switch( type )
  {
      case IKEv2_AMT_RSA_DS:
      case IKEv2_AMT_DSS_DS:
	  res=CertAuthenticate(i2,cert,adata,session->rmlen+nlen+adlen,data,dlen);
	  break;

  case IKEv2_AMT_SKMIC:
    while(list) {
	exist=0;
	if(list->id && list->idtype == idt && list->idlen==(signed)(idlen-sizeof(struct IKEv2IdentificationPayloadBody)) &&
		!memcmp(list->id,id,idlen-sizeof(struct IKEv2IdentificationPayloadBody))) {
	    exist=1;
	    break;
	}
	list=list->next;
    }
    if(exist && (list->authtype & IKEv2_AUTH_SK)==IKEv2_AUTH_SK) {
	if(session->Type==IKEv2_STY_RESPONDER) {
	    pwd=i2->pwd;
	    pwdlen=i2->pwdlen;
	} else {
	    pwd=list->pwd;
	    pwdlen=list->pwdlen;
	}
	prf(pwd, pwdlen, (const uint8_t*)"Key Pad for EAP-IKEv2", 21, &buf, &blen, GetTransformID( IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA, session->UsedTransforms ) );
	if(blen) {
	    prf( buf, blen, adata, session->rmlen + nlen + adlen, &buf2, &len, GetTransformID( IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA, session->UsedTransforms ) );
	    if(len && len == dlen && !memcmp( data, buf2, len ) ) {
		res = 1;
		free( buf2 );
		buf2=NULL;
	    }
	}
    } else if(exist && !((list->authtype & IKEv2_AUTH_SK)==IKEv2_AUTH_SK)) {
        xlogf(I2L_DBG,"EAP-IKEv2: Shared key authentication disabled for the account\n");
    } else {
	xlogf(I2L_DBG,"EAP-IKEv2: User not found on user list\n");
    }
    break;
  }

  if(buf) {
      free(buf);
  }
  free( adata );
  adata=NULL;
  return res;     // default = not authenticated
}

// Certificate functions

/**
 * @brief Load a certificate from a file and convert it to DER encoding.
 *
 * @param type      type of certificate
 * @param *filename name of file containing certificate
 * @param **cert    result
 *
 * @return length in bytes of cert 
 */
int LoadCertData( int type, char *filename, unsigned char **cert )
{
  FILE *fp;
  X509 *x509;
  int len, err;
  unsigned char *ce;

  *cert = NULL;

  switch( type )
  {
  case IKEv2_CET_X509_SIGNATURE:
    if( !( fp = fopen( filename, "r" ) ) )   // open cert file
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't open certificate file.\n" );
      return 0;
    }
    if( !( x509 = PEM_read_X509( fp, NULL, NULL, NULL ) ) )    // load cert from file
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't read certificate file.\n" );
      fclose(fp);
      return 0;
    }
    len = i2d_X509( x509, NULL );             // convert cert to DER encoding
    if( !( ce = (unsigned char*)malloc( len ) ) )
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't alloc mem for certificate.\n" );
      fclose(fp);
      return 0;
    }

    *cert = ce;
    err = i2d_X509( x509, &ce );
    X509_free( x509 );
    x509=NULL;

    if( !err )
      len = 0;
    fclose(fp);
    break;

  default:
    xlogf( I2L_ERR, "EAP-IKEv2: Error: not supported certificate type.\n" );
    len = 0;
    break;
  }
    
  return len;
}

// shared secred, misc functions

/**
 * @brief add a shared secret = id/pwd to the list of acceptables shared secrets of other party
 *
 * @param **list   ptr to list
 * @param idt      id type
 * @param *id      id
 * @param *pwd     password
 * @param authtype allowed authetnication type
 */
void AddSharedSec( struct sharedSecList **list, int idt, char *id, char *pwd, int authtype)
{
    uint8_t *t_id,*t_pwd;
    char pwdlen,idlen;
    struct sharedSecList *sl = (struct sharedSecList *)malloc( sizeof( struct sharedSecList ) );
    if( !sl )
	return;

    idlen=hex2bytes(id, (uint8_t **)&t_id);
    if(!idlen) {                                    //if not binary
	idlen=strlen(id);
	t_id=Clone((uint8_t*)id,idlen);
    }
    if (!idlen||!t_id) return;

    if(pwd){
	pwdlen=hex2bytes(pwd, (uint8_t **)&t_pwd);
	if(!pwdlen) {                   //if not binary
	    pwdlen=strlen(pwd);
	    t_pwd=Clone((uint8_t*)pwd,pwdlen);
	}
    } else {
	t_pwd=NULL;
	pwdlen=0;
    }

    if (!idlen||!t_id) return;
    sl->id = t_id;
    sl->idlen=idlen;
    sl->idtype = idt;
    sl->authtype=authtype;
    sl->pwd = t_pwd;
    sl->pwdlen=pwdlen;
    sl->next = *list;
    *list = sl;
}

/**
 * @brief Update list or add  secret to list
 *
 * @param **list      ptr to shared secret list
 * @param id_type     id type
 * @param *id         id
 * @param *secret     password
 * @param authtype    allowed authetnication type
 *
 * @return  0 on success, non 0 otherwise
 */
int UpdateSharedSec(struct sharedSecList **list,int id_type, char *id, char *secret, int authtype)
{
    
    if(!*list) { // just initiate the list
	AddSharedSec(list,id_type,id,secret,authtype);
    } else {
	int found=0;
	struct sharedSecList *item=*list;
	int idlen,secretlen;
	uint8_t *tid=NULL,*tpwd=NULL;
	idlen=hex2bytes(id,&tid);
	if(!idlen) {
	    idlen=strlen(id);
	}
	while (item) {
	    if(idlen==item->idlen&&!memcmp(id,item->id,idlen)) { // item found, update if nessessary
		found=1;
		free(tid);
		item->idtype=id_type;
		item->authtype=authtype;
		if(secret) {
		    secretlen=hex2bytes(secret,&tpwd);
		    if(!secretlen) {
			secretlen=strlen(secret);
		    }
		    if(secretlen!=item->pwdlen || memcmp(item->pwd,secret,secretlen)) {
			free(item->pwd);
			if(tpwd) {
			    item->pwd=tpwd;
			} else {
			    item->pwd=(uint8_t*)malloc(secretlen);
			    if(!item->pwd) {
				xlogf(I2L_ERR,"EAP-IKEv2: %s: error while allocating memory",__FUNCTION__);
				return 1;
			    }
			    memcpy(item->pwd,secret,secretlen);
			}
			item->pwdlen=secretlen;
		    }
		} else {
		    item->pwd=NULL;
                    item->pwdlen=0;
		}
		break;
	    }
	    item=item->next;
	}
	if(!found) {
	    AddSharedSec(list,id_type,id,secret,authtype);
	}
    }
    return 0;
}


/**
 * @brief check if id/pwd is in list
 * @warning kwz: this function is unused
 *
 * @param *list ptr to list
 * @param *id id
 * @param idlen id length
 * @param *pwd  password
 * @param pwdlen password length
 *
 * @return 1 if found, 0 otherwise
 */
int IsAcceptable( struct sharedSecList *list, char *id, int idlen, char *pwd, int pwdlen )
{
  while( list )
  {
    if( list->id && list->pwd && id && pwd && idlen==list->idlen && !memcmp( list->id, id, idlen ) && pwdlen==list->pwdlen && !memcmp( list->pwd, pwd, pwdlen ) )
      return 1;
    list = list->next;
  }
  return 0;
}



/**
 * @brief free shered secret list memory
 *
 * @param *list ptr to list
 */
void FreeSharedSecList(struct sharedSecList *list)
{
    struct sharedSecList *tmp;
    while(list)  {
	free(list->pwd);
	list->pwd=NULL;
	free(list->id);
	list->id=NULL;
	tmp=list->next;
	free(list);
	list=tmp;
    }
}

/**
 * @brief Check if username on user list match the certifcate common name and
 * user is allowed to use certificate for authetnication.
 *
 * @param sslist shared secret list
 * @param x509 certificate
 *
 * @return 1 on success, 0 on fail
 */
int AuthorizeCert(struct sharedSecList *sslist, X509 *x509) 
{
    int res=0;
    // lets take a commonName from given certifiacate 
    X509_NAME *subject=X509_get_subject_name(x509);
    if(!subject) {
	xlogf(I2L_ERR,"EAP-IKEv2: Can't find name field in cert\n");
    }
    char *name_buf;
    int name_len;
    int nid=OBJ_txt2nid("commonName");
    name_len=X509_NAME_get_text_by_NID(subject,nid,NULL,0);
    if(name_len==-1) {
	xlogf(I2L_ERR,"EAP-IKEv2: CN is not present\n");
    } else {
	name_buf=(char*)malloc(name_len+1);
	if(!name_buf) {
	    xlogf(I2L_ERR,"EAP-IKEv2: Error: Can't allocente mem for CN data\n");
	    return 0;
	}
	name_buf[name_len]='\0';
	X509_NAME_get_text_by_NID(subject,nid,name_buf,name_len+1);
	// additional username  based verification
	while(sslist) {
	    if(name_len==sslist->idlen && !memcmp(sslist->id,name_buf,name_len) && ((sslist->authtype & IKEv2_AUTH_CERT) == IKEv2_AUTH_CERT)) {
		res=1;
		break;
	    }
	    sslist=sslist->next;
	}
	if(!res) {
	    xlogf(I2L_ERR,"EAP-IKEv2: Cert authorization failed - CN=[%s]\n",name_buf);
	}


	free(name_buf);
	name_buf=NULL;
    }
    return res;
}


/**
 * @brief
 * Creates a list of SHA1 hashes of public key trusted CA's as described in 
 * paragraph 3.7 RFC4306 (IKEv2).
 *
 * @param x509_store local cert store
 * @param result place for result
 * 
 * @return length of result, or 0 in case of error
 */
uint32_t CreateCAHashString(X509_STORE *x509_store,uint8_t **result)
{
    int i;
    uint32_t cert_num=0;
    uint32_t current_cert=0;
    X509_OBJECT *xobj;
    uint8_t *dest;
    uint8_t *pkdata;
    uint32_t pklen;
    if(!x509_store->objs) {
	xlogf(I2L_ERR,"EAP-IKEv2: Not enough data to generate hash\n");
	*result=NULL;
	return 0;
    }
    // count certifiactes
    for(i=0;i<sk_num(x509_store->objs);i++) {
	xobj=(X509_OBJECT*)sk_value(x509_store->objs,i);
	if(xobj && xobj->type==X509_LU_X509) {
	    cert_num++;
	}
    }
    if(!(dest=(uint8_t*)malloc(cert_num*SHA_DIGEST_LENGTH))) {
	xlogf(I2L_ERR,"EAP-IKEv2: Can't alloc memory\n");
	*result=NULL;
	return 0;
    }
    *result=dest;
    
    // for each certificate compute sha1(pubkey) and add to result
    for(i=0;i<sk_num(x509_store->objs);i++) {
	xobj=(X509_OBJECT*)sk_value(x509_store->objs,i);
	if(xobj!=NULL && xobj->type==X509_LU_X509) {
	    X509 *x509=xobj->data.x509;
	    pkdata=X509_get_X509_PUBKEY(x509)->public_key->data;
	    pklen=X509_get_X509_PUBKEY(x509)->public_key->length;
	    SHA1(pkdata,pklen,dest+current_cert*SHA_DIGEST_LENGTH);
	    current_cert++;
	}
    }
    return cert_num*SHA_DIGEST_LENGTH;
}


 /**
  * @brief
  * Generate auth data using certificate private key (RSA or DSA)
  *
  * @param pkey private key structure
  * @param data to sign
  * @param dlen data length
  * @param *dest place for sign
  *
  * @return length of sign
  */

uint16_t GenerateCertAuthData(EVP_PKEY *pkey,uint8_t *data,uint32_t dlen,uint8_t *dest) 
{
    uint8_t md[SHA_DIGEST_LENGTH];
    unsigned int len;
    uint8_t *buf;
    int err;
    SHA1(data,dlen,md);
    switch(EVP_PKEY_type(pkey->type)) {
	case EVP_PKEY_RSA:
	    if(!(buf=(uint8_t*)malloc(RSA_size(pkey->pkey.rsa )))) {
		xlogf( I2L_ERR, "EAP-IKEv2: Can't alloc mem for buffer.\n" );
		return 0;
	    }
	    err = RSA_sign( NID_sha1, md, SHA_DIGEST_LENGTH, buf, &len, pkey->pkey.rsa );
	    break;
	case EVP_PKEY_DSA:
	    if(!(buf=(uint8_t*)malloc(DSA_size(pkey->pkey.dsa)))) {
		xlogf( I2L_ERR, "EAP-IKEv2: Can't alloc mem for buffer.\n" );
		return 0;
	    }
	    err = DSA_sign( NID_sha1, md, SHA_DIGEST_LENGTH, buf, &len, pkey->pkey.dsa);
	    break;
	default:
	    xlogf(I2L_ERR,"EAP-IKEv2: Unsupported key type");
	    return 0;
    }
    if( err ) {
	memcpy(dest,buf,len);
    } else {
	len = 0;
    }
    free(buf);
    return (uint16_t)len;
}


/**
 * @brief Get authentication method for auth payload associated with this private
 * key
 *
 * @param *pkey private key data
 *
 * @return IKEv2_AMT_RSA_DS | IKEv2_AMT_DSS_DS | -1 in case of unknown type
 */
int GetCertAuthMethod(EVP_PKEY *pkey) 
{
    switch(EVP_PKEY_type(pkey->type)) {
	case EVP_PKEY_RSA:
	    return IKEv2_AMT_RSA_DS;
	case EVP_PKEY_DSA:
	    return IKEv2_AMT_DSS_DS;
	default:
	    return -1;
    }
}

/**
 * @brief Authenticate peer using data acguired from CERT payload
 * Opertation is performed in three main steps:
 * 	- verify certifcate against trusted CA certificate
 * 	- chek if certificate CommonName is on sslist
 * 	- verify signature
 *
 * @param *i2 IKEv2 context data
 * @param *cert certificate data
 * @param *adata data to verify
 * @param alen adata length
 * @param *sigbuf sigbuf signature to verify
 * @param siglen signature siglen
 *
 * @return 1 on succes, 0 otherwise
 */
int CertAuthenticate(const struct ikev2_ctx *i2, struct CertList *cert, 
		uint8_t *adata,uint32_t alen, uint8_t *sigbuf, uint32_t siglen)
{
    int res=0,err;
    uint8_t md[SHA_DIGEST_LENGTH];
    xlogf( I2L_DBG, "EAP-IKEv2: Verifying certificate(s).\n" );
    if(cert==NULL || !cert->cdata || !cert->cdlen) {
	return 0;
    }
    if(cert->ctype!=IKEv2_CET_X509_SIGNATURE) {
	xlogf(I2L_ERR,"Unsupported certificate type\n");
	return 0;
    }
    X509 *x509=NULL;

    //verify cert
    X509_STORE *ctx = i2->x509_store;
    X509_STORE_CTX *csc;

    uint8_t *cdata=cert->cdata;
    uint32_t cdlen=cert->cdlen;
    x509 = d2i_X509( NULL, (unsigned char **)&cdata, cdlen );
    if( !x509 )
	return 0;
    csc = X509_STORE_CTX_new();
    X509_STORE_CTX_init( csc, ctx, x509, NULL );
    res = X509_verify_cert( csc );
    err = csc->error;
    //X509_STORE_CTX_cleanup( csc );
    X509_STORE_CTX_free(csc);
    if( err ) 
	xlogf( I2L_ERR, "EAP-IKEv2: Certificate verification failed: %s.\n",
		X509_verify_cert_error_string( err ) );
    if(res) { // certifacate is valid
	res=AuthorizeCert(i2->sslist,x509);
    }
    if(!res) {
	X509_free(x509);
	return res;
    }

    //authenticate
    xlogf(I2L_DBG,"EAP-IKEv2: Authenticate ...\n");
    SHA1(adata,alen,md);
    EVP_PKEY *public_key=X509_get_pubkey(x509);
    switch(EVP_PKEY_type(public_key->type)) {
	case EVP_PKEY_DSA:
	    //TODO:check authtype!!!
	    res=DSA_verify(NID_sha1,md,SHA_DIGEST_LENGTH,sigbuf,siglen,public_key->pkey.dsa);
	    break;
	case EVP_PKEY_RSA:
	    //TODO:check authtype!!!
	    res=RSA_verify(NID_sha1,md,SHA_DIGEST_LENGTH,sigbuf,siglen,public_key->pkey.rsa);
	    break;
    }
    if(res==-1) {
	res=0;
    }
    EVP_PKEY_free(public_key);
    X509_free(x509);

    return res;
}
