/*
 *  conf_utils.c - functions used to parse config options
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *
 */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "conf_utils.h"
#include "enums.h"


#define IKEv2_DEFAULT_ENCR_KEY_LEN_AES_CBC 128
#define IKEv2_DEFAULT_ENCR_KEY_LEN_BLOWFISH 128
#define IKEv2_DEFAULT_ENCR_KEY_LEN_CAST 128
/// Available transform types
static const struct TransformConfigItem availableTransformTypes[]= {
    // encryption
#if 0
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_DES_IV64     ,"des_iv64"      ,0,0},
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_DES          ,"des",          ,0,0},
#endif
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_3DES         ,"3des"          ,0,0},
#if 0
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_RC5          ,"rc5"           ,0,0},
#endif
#ifndef OPENSSL_NO_IDEA
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_IDEA         ,"idea"          ,0,0},
#endif
#ifndef OPENSSL_NO_CAST
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_CAST         ,"cast",  
	1, IKEv2_DEFAULT_ENCR_KEY_LEN_CAST},
#endif
#ifndef OPENSSL_NO_BF
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_BLOWFISH     ,"blowfish",
	1, IKEv2_DEFAULT_ENCR_KEY_LEN_BLOWFISH},
#endif
#if 0
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_3IDEA        ,"3idea"         ,0,0},
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_DES_IV32     ,"des_iv32"      ,0,0},
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_RC4          ,"rc4"           ,0,0},
#endif
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_NULL         ,"null"          ,0,0},
#ifndef OPENSSL_NO_AES
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_AES_CBC      ,"aes_cbc",
	1,IKEv2_DEFAULT_ENCR_KEY_LEN_AES_CBC},
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_AES_CBC      ,"aes_128_cbc",
	1,128},
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_AES_CBC      ,"aes_192_cbc",
	1,192},
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_AES_CBC      ,"aes_256_cbc",
	1,256},
#endif

#if 0
    {IKEv2_TRT_ENCRYPTION_ALGORITHM,  IKEv2_TID_ENCR_AES_CTR      ,"aes_ctr"       ,0,0},
#endif

    // integrity
    {IKEv2_TRT_INTEGRITY_ALGORITHM,   IKEv2_TID_AUTH_HMAC_MD5_96  ,"hmac_md5_96"   ,0,0},
    {IKEv2_TRT_INTEGRITY_ALGORITHM,   IKEv2_TID_AUTH_HMAC_SHA1_96 ,"hmac_sha1_96"  ,0,0},
    {IKEv2_TRT_INTEGRITY_ALGORITHM,   IKEv2_TID_AUTH_DES_MAC      ,"des_mac"       ,0,0},
#if 0
    {IKEv2_TRT_INTEGRITY_ALGORITHM,   IKEv2_TID_AUTH_KPDK_MD5     ,"kpdk_md5"      ,0,0},
    {IKEv2_TRT_INTEGRITY_ALGORITHM,   IKEv2_TID_AUTH_AES_XCBC_96  ,"aes_xcbc_96"   ,0,0},
#endif

    // prf
    {IKEv2_TRT_PSEUDO_RANDOM_FUNCTION,IKEv2_TID_PRF_HMAC_MD5      ,"hmac_md5"      ,0,0},
    {IKEv2_TRT_PSEUDO_RANDOM_FUNCTION,IKEv2_TID_PRF_HMAC_SHA1     ,"hmac_sha1"     ,0,0},
    {IKEv2_TRT_PSEUDO_RANDOM_FUNCTION,IKEv2_TID_PRF_HMAC_TIGER    ,"hmac_tiger"    ,0,0},
#if 0
    {IKEv2_TRT_PSEUDO_RANDOM_FUNCTION,IKEv2_TID_PRF_AES_CBC       ,"aes_cbc"       ,0,0},
#endif

    // dhg 
    {IKEv2_TRT_DIFFIE_HELLMAN_GROUP,  IKEv2_TID_DHG_768MODP       ,"modp768"       ,0,0},
    {IKEv2_TRT_DIFFIE_HELLMAN_GROUP,  IKEv2_TID_DHG_1024MODP      ,"modp1024"      ,0,0},
#if 0
    {IKEv2_TRT_DIFFIE_HELLMAN_GROUP,  IKEv2_TID_DHG_155EC2N       ,"155ec2n"       ,0,0},
    {IKEv2_TRT_DIFFIE_HELLMAN_GROUP,  IKEv2_TID_DHG_185EC2N       ,"185ec2n"       ,0,0},
#endif
    {IKEv2_TRT_DIFFIE_HELLMAN_GROUP,  IKEv2_TID_DHG_1536MODP      ,"modp1536"      ,0,0},
    {IKEv2_TRT_DIFFIE_HELLMAN_GROUP,  IKEv2_TID_DHG_2048MODP      ,"modp2048"      ,0,0},
    {IKEv2_TRT_DIFFIE_HELLMAN_GROUP,  IKEv2_TID_DHG_3072MODP      ,"modp3072"      ,0,0},
    {IKEv2_TRT_DIFFIE_HELLMAN_GROUP,  IKEv2_TID_DHG_4096MODP      ,"modp4096"      ,0,0},
    {IKEv2_TRT_DIFFIE_HELLMAN_GROUP,  IKEv2_TID_DHG_6144MODP      ,"modp6144"      ,0,0},
    {IKEv2_TRT_DIFFIE_HELLMAN_GROUP,  IKEv2_TID_DHG_8192MODP      ,"modp8192"      ,0,0},

    // terminator
    {0, 0, NULL ,0,0}
};


/// Available ID types
static const struct IdTypeConfigItem availableIdTypes[]= {
    
    { IKEv2_IDT_IPV4_ADDR   , "IPV4_ADDR"},
    { IKEv2_IDT_FQDN        , "FQDN"},
    { IKEv2_IDT_RFC822_ADDR , "RFC822_ADDR"},
    { IKEv2_IDT_IPV6_ADDR   , "IPV6_ADDR"},
    { IKEv2_IDT_DER_ASN1_DN , "ASN1_DN"},
    { IKEv2_IDT_DER_ASN1_GN , "ASN1_GN"},
    { IKEv2_IDT_KEY_ID      , "KEY_ID"},

    // terminator
    { 0, NULL }
};






/**
 * @brief Get pointer to the array containing list of the available transforms
 * types
 *
 * @return  pointer to available transforms types
 */
const struct TransformConfigItem* GetAvailableTransformsTypes(void)
{
    return availableTransformTypes;
}


/**
 * @brief Get pointer to the array contaning list of the available ID types
 *
 * @return  pointer to the array contaning list  of available ID types
 */
const struct IdTypeConfigItem* GetAvailableIdTypes(void)
{
    return availableIdTypes;
}


/**
 * @brief Get transform id from string, useful in configuration
 *
 * @param name string transform name
 * @param type transform type (IKEv2_TRT_...)
 * @param id returned transform id
 * @param keylen result ptr for key length
 *
 * @return 0 if succes, -1 if fail
 */
int TransformFromName(const char *name, uint8_t type, uint8_t *id,uint16_t *keylen)
{
    int res=-1;
    int i=0;
    const struct TransformConfigItem* at=GetAvailableTransformsTypes();
    uint16_t keyout=0;
    if(!name) {
	return -1;
    }

    for(i=0;at[i].name;i++) {
	if(at[i].type==type && !strncasecmp(name,at[i].name,strlen(at[i].name))){
	  *id=at[i].id; 
	  if(at[i].variable_key_len) {
	      keyout=extractKeyLen(name);
	      if(!keyout) {
		  keyout=at[i].default_key_len;
	      }
	  }
	  res=0;
	  break;
	}
    }

    if(!keylen && keyout) {
	res=-1;
    } else {
	if(keylen) {
	    *keylen=keyout;
	}
    }
    if(!res) {
	xlogf(I2L_DBG,"EAP-IKEv2: name %s, group=%d, id=%d, keylen=%d\n",name,type,*id,keyout);
    }
    return res;
}


/**
 * @brief
 * Get type id  from name string , useful in configuration
 *
 * @param name string idtype name
 *
 * @return -1 if fail,id on success;
 */
int IdTypeFromName(const char *name) 
{
    int i;
    const struct IdTypeConfigItem *ait=GetAvailableIdTypes();
    if(!name) {
	return -1;
    }  
    for(i=0;ait[i].name;i++) {
	if(!strcasecmp(name,ait[i].name)) {
	    return ait[i].type;
	}
    }
    return -1;
}


/**
 * @brief Get authe type id from string literal
 *
 * @param *authtype string literal of authtype
 *
 * @return auth type id
 */
int AuthtypeFromName(const char *authtype)
{
    if(authtype==NULL) {
	return -1;
    }
    if(!strcasecmp(authtype,"secret")) {
	return IKEv2_AUTH_SK;
    }
    if(!strcasecmp(authtype,"cert")) {
	return IKEv2_AUTH_CERT;
    }
    if(!strcasecmp(authtype,"none")) {
	return IKEv2_AUTH_NONE;
    }
    if(!strcasecmp(authtype,"both")) {
	return IKEv2_AUTH_BOTH;
    }
    return -1;
}

/**
 * @brief Extract key length from algorithm name saved in folowing form:
 *        alg_name:n where:
 *          - alg_name - descriptive name of the algorithm
 *          - n - integer value representing key length
 *        Useful for variable key length encryption algorithms.  
 *
 * @param *alg string literal in form "alg_name:n"
 *
 * @return integer value of n (key length)
 */
uint16_t extractKeyLen(const char *alg)
{
    int parseKey=0;
    if(!alg) 
	return 0;
    while(*alg && *alg++!=':' && *alg!=' '  && *alg!='\t') {
	if(*alg==':' && isdigit(alg[1])) {
	    parseKey=1;
	}
    }
    return parseKey? atoi(alg):0;
}


/**
 * @brief
 * Get human readable transform type string.
 * The outut of this function is in format "%s/%s", where first part is
 * transform type, and second part is used algorithm.
 *
 * @param type transform type
 * @param alg used algorithm
 * 
 * @return pointer to the static buffer which contains human readable transform
 * type string
 */

const char* GetTransformName(int type,int alg) 
{
    static char result[256];
    const char* salg="unknown";
    const char* stype="unknown";
    switch(type) {
	case IKEv2_TRT_PSEUDO_RANDOM_FUNCTION:
	    stype="prf";
	    break;
	case IKEv2_TRT_INTEGRITY_ALGORITHM:
	    stype="integrity";
	    break;
	case IKEv2_TRT_ENCRYPTION_ALGORITHM:
	    stype="encr";
	    break;
	case IKEv2_TRT_DIFFIE_HELLMAN_GROUP:
	    stype="dhgroup";
	    break;
    }
    int i;
    const struct TransformConfigItem* at=GetAvailableTransformsTypes();
    for(i=0;at[i].name;i++) {
	if(type==at[i].type && alg==at[i].id) {
	    salg=at[i].name;
	    break;
	}
    }
    sprintf(result,"%s/%s",stype,salg);
    return result;

}
    
