/*
 *  connector.c - public library interface
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include "common.h"

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif // HAVE_NETINET_IN_H
#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif // HAVE_WINSOCK_H

#include <assert.h>
#include <string.h>
#include<openssl/x509.h>
#include<openssl/pem.h>

#include "payload.h"
#include "message.h"
#include "IKEv2.h"
#include "payload.h"
#include "encr.h"
#include "prf.h"
#include "integ.h"
#include "connector.h"
#include "state.h"
#include "enums.h"
#include "auth.h"

/**
 *  @brief Cert, private key, crl init 
 *
 *  @param *i2 ikev2 context structure
 *
 *  @return 0 on success, 1 otherwise 
 **/

int CertInit( struct ikev2_ctx *i2 )
{
  if(i2->trusted){  // load ca cert and crl
          X509_LOOKUP *lookup = NULL;
          OpenSSL_add_all_algorithms();

          i2->x509_store = X509_STORE_new();

          lookup = X509_STORE_add_lookup( i2->x509_store, X509_LOOKUP_file() );
          X509_LOOKUP_load_file( lookup, i2->trusted , X509_FILETYPE_PEM );
          //lookup = X509_STORE_add_lookup( ctx, X509_LOOKUP_hash_dir() );
          //X509_LOOKUP_add_dir( lookup, i2->trusted, X509_FILETYPE_PEM );

          //lookup = X509_STORE_add_lookup( ctx, X509_LOOKUP_file() );
          //X509_LOOKUP_load_file( lookup, trusted, X509_FILETYPE_DEFAULT );
          //lookup = X509_STORE_add_lookup( ctx, X509_LOOKUP_hash_dir() );
          //lookup = X509_STORE_add_lookup( ctx, X509_LOOKUP_file() );
          //X509_LOOKUP_load_file( lookup, trusted, X509_FILETYPE_DEFAULT );
          //lookup = X509_STORE_add_lookup( ctx, X509_LOOKUP_hash_dir() );
          //X509_LOOKUP_add_dir( lookup, trusted, X509_FILETYPE_PEM );

          xlogf( I2L_DBG, "EAP-IKEv2: trusted = %s\n", i2->trusted);
          if(i2->crl_file) {
#if (OPENSSL_VERSION_NUMBER > 0x00907000L)
              if (X509_load_crl_file(lookup, i2->crl_file, X509_FILETYPE_PEM) != 1)
                  xlogf (I2L_ERR,"Error reading the CRL file [%s]",i2->crl_file);

              X509_STORE_set_flags(i2->x509_store, X509_V_FLAG_CRL_CHECK |
                      X509_V_FLAG_CRL_CHECK_ALL);
#else
              xlogf(L_ERR,"CRL is disabled in used OpenSSL version!");
              return 1;
#endif
          }
	  i2->CAsHashStringLen=CreateCAHashString(i2->x509_store,&(i2->CAsHashString));
  }
  if(i2->pkfile){   //load private key
       FILE *pkfp;
       if( !( pkfp = fopen( i2->pkfile, "r" ) ) )
       {
         xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't open private rsa key file.\n" );
         return 1;
       }

       i2->rpk = PEM_read_PrivateKey( pkfp, NULL, NULL, i2->pkfile_pwd);
       fclose( pkfp );
       if( !i2->rpk )
       {
         xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't read private rsa key file.\n" );
         return 1;
       }
       xlogf( I2L_DBG, "EAP-IKEv2: private key loaded.\n" );
   }
   if (i2->certfile){
        i2->certlen=LoadCertData( IKEv2_CET_X509_SIGNATURE, i2->certfile, (unsigned char **)&i2->cert );
        if(!i2->certlen){
                xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't read rsa cert file.\n");
                return 1;
        }
       xlogf( I2L_DBG, "EAP-IKEv2: rsa cert loaded.\n" );
   }
  return 0;
}

/**
 *  @brief Init IKEv2 
 *  @deprecated
 *  
 *  kwz:Czy ta funkcja naprawde jest nam potrzebna ? Chyba nie bardzo
 **/
int IKEv2Init( struct ikev2_ctx *i2 )
{
    i2=i2; // -Wunused
//	i2->SessionList = NULL;     // no sessions at init
//	i2->suppProp = NULL;
//	xlogf(I2L_DBG, LOCATION("EAP-IKEv2: Init IKEv2.\n") );
//
//	FILE *fp;
//	char buf[100];
//	int nb;
//	if( !i2->randomfilename || !( fp = fopen( i2->randomfilename, "r" ) ) )
//		xlogf(I2L_DBG, "EAP-IKEv2: Warning: Can't seed the random number generator: random file not found\n" );
//	else
//	{
//		nb = fread( buf, 1, 100, fp );
//		fclose( fp );
//		RAND_seed( buf, nb );
//	}
//
	return 0;
}

/**
 * @brief shutdown IKEv2  
 * @deprecated
 **/
void IKEv2Shutdown()
{
	RAND_cleanup();   // shutdown RAND
	xlogf( I2L_INFO, "EAP-IKEv2: IKEv2 shutdown.\n" );
	//close_logfile();  // shutdown logging
}

/**
 * @brief Generate MSK (key which is used to protect communication between user
 * and NAS. Function set proper value of s->eapKeyData.
 *
 * @param *s   IKEv2 session data
 * @param len  key len
 *
 * @return 1  when keys successfully generated, 0 otherwise.
 */
int GenEapKeys(struct IKEv2Session *s, int len){
        if(s->State!=IKEv2_STATE_CONNECTED||s->SK_ready==0) return 0;
        xlogf( I2L_DBG, "EAP-IKEv2: Creating EAP keys\n");
        uint32_t nli = BN_num_bytes( s->Ni );
        uint32_t nlr = BN_num_bytes( s->Nr );
        uint8_t *nonces = (uint8_t *)malloc( nli + nlr); 
        if (nonces==NULL) return 0;
        BN_bn2bin( s->Ni, nonces );
        BN_bn2bin( s->Nr, nonces + nli );

        uint32_t elen = GetEncrKeyLen(
	       	GetTransformID( IKEv2_TRT_ENCRYPTION_ALGORITHM, IKEv2_PID_IKE_SA, s->UsedTransforms ),
		GetTransformKeyLen(IKEv2_TRT_ENCRYPTION_ALGORITHM,IKEv2_PID_IKE_SA,s->UsedTransforms)
	       	);
	if(s->eapKeyData) {
	    free(s->eapKeyData);
	}
        prf_plus( s->SK_d, elen, nonces, nli + nlr, &s->eapKeyData, len, GetTransformID(IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA, s->UsedTransforms ) );
        free( nonces );
	nonces=NULL;
	return 1;
}


/**
 * @brief  Creating an IKEv2 message.
 *
 * @param *i2         IKEv2 context data 
 * @param *ikemsg     IKEv2 message to send
 * @param imlen       length of ikemsg
 * @param response    create response? 
 * @param msgid       message id
 * @param *session    session data 
 * @param **out       destination buffer
 *
 * @return  length of outcoming message
 */
uint32_t CreateIKEv2Message(const struct ikev2_ctx *i2, uint8_t *ikemsg, uint32_t imlen, bool response, uint8_t msgid,
		struct IKEv2Session *session, uint8_t **out )
{
	xlogf( I2L_DBG, "EAP-IKEv2: Creating IKE msg.\n" );

	uint32_t ilen = GetChecksumLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
				session->UsedTransforms ) );         // length of integrity checksum

	if( ikemsg )    // did we get a ikemsg (we get it only for first fragment)
	{
		if( imlen != ntohl( ( (struct IKEv2Header *)ikemsg )->Length ) )  // check length fields
		{
			xlogf( I2L_ERR, "EAP-IKEv2: Length fields of IKEv2 msg are inconsistent.\n" );
			*out = NULL;
			return 0;
		}

//		struct IKEv2PayloadHeader *pl = (struct IKEv2PayloadHeader *)( ikemsg + sizeof( struct IKEv2Header ) );
//		uint8_t nextpl = ( ( struct IKEv2Header *)ikemsg )->NextPayload;
//		session->hasinteg = false;   // msg should be integrity protected??
//                printf("-------- state %d    type %d   status %d\n", session->State,session->Type,session->Status);
//                if ( ! (session->State == IKEv2_STATE_NOT_CONNECTED && session->Type == IKEv2_STY_RESPONDER && session->Status != IKEv2_SST_FAILED )){
//                //if ( ! (session->State == IKEv2_STATE_INIT && session->Type == IKEv2_STY_RESPONDER)){
//		        while( nextpl )       // search for encrypted payload
//                        {
//                                if( nextpl == IKEv2_NPT_ENCRYPTED )
//                                {
//                                        session->hasinteg = true;
//                                        break;
//                                }
//                                nextpl = pl->NextPayload;
//                                pl = ( (struct IKEv2PayloadHeader *)( ( (char *)pl ) + ntohs( pl->PayloadLength ) ) );
//                        }
//                }
		session->fragment = 0;          // init fragmentation parameters
		session->fragdata = ikemsg;
		session->fdlen = imlen;
		session->iklen = GetIntegKeyLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
					session->UsedTransforms ) );
                if(session->integkey!=NULL)  {
		    free(session->integkey);
		    session->integkey=NULL;
		}
		session->integkey = Clone( ( ( (struct IKEv2Header *)ikemsg )->Flags & IKEv2_HDF_INITIATOR ) ? session->SK_ai : session->SK_ar, session->iklen );
	       	session->integtype = GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA, session->UsedTransforms );
	}
	if( !( session->fragdata ) )  // is there something to send?
		return 0;

	uint8_t *chksum;

	bool lastfrag = true;
	uint32_t eaplen = sizeof( struct EAPHeader ) + ( session->include_integ ? ilen : 0 ) + ( session->fragment == 0 ? 4 : 0 );                   // length of eap specific data
	uint32_t fraglen = session->fdlen - session->fragment + eaplen; // size of fragment
	if( fraglen > i2->max_fragment_size )       // last fragment??
	{
		fraglen = i2->max_fragment_size;
		lastfrag = false;
	}

	if( !( *out = (uint8_t*)malloc( fraglen ) ) )     // alloc mem for fragment
	{
		xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't alloc mem for IKEv2 message.\n" );
		return 0;
	}

	struct EAPHeader *hdr = (struct EAPHeader *)*out;

	hdr->Code = response ? EAP_CODE_RESPONSE : EAP_CODE_REQUEST;   // create a header
	hdr->Id = response ? msgid : session->eapMsgID; 
	hdr->Length = htons( fraglen );
	hdr->Type = EAPIKEv2_TYPE;
	hdr->Flags = ( session->include_integ ? EAP_HDF_INTEGRITYINCLUDED : 0 ) |    // integrity checksum?
		( lastfrag ? 0 : EAP_HDF_MOREFRAGMENTS ) |                 // is last fragment?
		( session->fragment == 0 ?
		  EAP_HDF_LENGTHINCLUDED : 0 );                            // has length field?

	uint32_t llen = 0;
	if( session->fragment == 0 )    // insert msg length field?
	{
		llen = 4;
		*( (uint32_t *)( *out + sizeof( struct EAPHeader ) ) ) = htonl( session->fdlen );  // insert msg length
	}

	memcpy( *out + sizeof( struct EAPHeader ) + llen, session->fragdata + session->fragment, fraglen - eaplen );                         // add the ikev2 data to msg

	if( session->include_integ )   // insert integrity checksum
	{
		xlogf( I2L_DBG, "EAP-IKEv2: Attaching integrity checksum\n" );
		uint32_t ilen2;
		if( !session->ointegkey )
		{
			session->ointegkey = Clone(session->integkey,session->iklen);
			session->oiklen = session->iklen;
		}
		integ( session->ointegkey, session->oiklen, *out, fraglen - ilen, &chksum, &ilen2, session->integtype );
		if(session->ointegkey!=NULL) {
		    free(session->ointegkey);
		    session->ointegkey=NULL;
		}
		session->ointegkey = Clone(session->integkey,session->iklen);
		session->oiklen = session->iklen;

		assert(ilen==ilen2); // should never happen
		
		memcpy( *out + fraglen - ilen, chksum, ilen );
		free( chksum );
		chksum=NULL;
	}
	session->fragment += fraglen - eaplen;      // next fragment
	if( lastfrag )                              // no more fragments?
	{
		free( session->fragdata );
		session->fragdata = NULL;
	}

	return fraglen;    // length of msg
}

/**
 * @brief
 * extract data from a eap message 
 * 
 * @param in     ptr to message 
 * @param data   ptr to data 
 * @param dlen   length of data 
 * @param ilen   integrity field length
 *
 * @return 0 on success, -1 otherwise
 */
int ExtractData(const uint8_t *in, uint8_t **data, uint32_t *dlen, uint32_t ilen )
{
	const struct EAPHeader *hdr = (const struct EAPHeader *)in;

	*data = NULL; // default
	*dlen = 0;
	int lenfld = 0;
	int flglen = 0;
	if( hdr->Type != EAPIKEv2_TYPE )
		flglen = -1;

	if( !flglen && hdr->Flags & EAP_HDF_LENGTHINCLUDED )
		lenfld = 4;

	if( ntohs( hdr->Length ) - ilen + lenfld > sizeof( struct EAPHeader ) + flglen )    // are there any data
	{
		*dlen = ntohs( hdr->Length ) - sizeof( struct EAPHeader ) + flglen - ilen - lenfld; // length of data
                
		if( !( *data = (uint8_t*)malloc( *dlen ) ) )
		{
			xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't alloc mem message data.\n" );
			return -1;
		}
		memcpy( *data, in + sizeof( struct EAPHeader ) + lenfld + flglen, *dlen );
	}
	return 0;
}


/**
 * @brief
 * parse an IKEv2 message 
 *
 *  - checking eap headers, lenght fields, fragmentation and integration
 *  - setting ikemsg if received full ikev2 message (all fragments), else NULL
 *
 * @param *in       ptr to message 
 * @param **ikemsg  IKEv2 message 
 * @param *imlen    length of ikemsg 
 * @param *session  session data
 *
 * @return 0 on success, -1 otherwise
 */
int ParseIKEv2Message(const uint8_t *in, uint8_t **ikemsg, uint32_t *imlen, struct IKEv2Session *session )
{
	const struct EAPHeader *hdr = (const struct EAPHeader *)in;
	*ikemsg=NULL;
	*imlen=0;

	if( ( hdr->Code != EAP_CODE_REQUEST && hdr->Code != EAP_CODE_RESPONSE ) || hdr->Type != EAPIKEv2_TYPE )
	{
		xlogf( I2L_ERR, "EAP-IKEv2: Error: got an unexpected message code or type.\n" );
		return -1;
	}

	xlogf( I2L_DBG, "EAP-IKEv2: Got an IKEv2 message.\n" );

	//bool has_chksum = false;
	uint32_t ilen = 0;

	uint32_t msglen = (uint32_t)ntohs( hdr->Length );
	int llen = 0;
	if( hdr->Flags & EAP_HDF_LENGTHINCLUDED )   // extract length field
	{
		msglen = ntohl( *( (const uint32_t *)( in + sizeof( struct EAPHeader ) ) ) );
		llen = 4;
	}
	else
	{
		if( ( hdr->Flags & EAP_HDF_MOREFRAGMENTS ) && !( session->fragdata ) )
		{
			xlogf( I2L_ERR, "EAP-IKEv2: Error: Fragmentation, but no msg length in the first fragment.\n" );
			return -1;
		}
	}

	if( !( session->fragdata ) )
	{
		if( ntohs( hdr->Length ) < sizeof( struct EAPHeader ) + sizeof( struct IKEv2Header ) )
		{   // check for minimal length
			xlogf( I2L_ERR, "EAP-IKEv2: Message too small for first fragment of an IKEv2 msg.\n" );
			return -1;
		}

		if( !( session->fragdata = (uint8_t*)malloc( msglen ) ) )
		{
			xlogf( I2L_ERR, "EAP-IKEv2: Error: can't alloc mem for message.\n" );
			return -1;
		}
		session->fdlen = msglen;
		session->fragment = 0;
		session->iklen = GetIntegKeyLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM,
					IKEv2_PID_IKE_SA, session->UsedTransforms ) );
		if(session->integkey) {
		    free(session->integkey);
		    session->integkey=NULL;
		}
		session->integkey = Clone ( ( ( (const struct IKEv2Header *)(in + sizeof( struct EAPHeader ) + llen ) )->Flags & IKEv2_HDF_INITIATOR ) ?session->SK_ai : session->SK_ar, session->iklen );
	}

	if( hdr->Flags & EAP_HDF_INTEGRITYINCLUDED )    // if there is an itegrity checksum, check it...
	{
                //if( !session->expect_integ )
                //{
                //        xlogf( I2L_ERR, "EAP-IKEv2: Message is integrity protected - it shouldn't be.\n\n\n\n\n\n" );
                //        //return -1;
                //}
		//has_chksum = true;
		//session->hasinteg = true;

		uint8_t *chksum;

		//if( !( session->SK_ai ) || !( session->SK_ar ) )  // do we have keys?
		if( !session->SK_ready ) { // do we have keys?
		        
			xlogf( I2L_ERR, "EAP-IKEv2: Error: Inegrity checksum included but no integrity keys avilable.\n" );
                        return -1;
                }

		ilen = GetChecksumLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
					session->UsedTransforms ) );
		
                uint32_t ilen2;
		integ( session->integkey, session->iklen, in, ntohs( hdr->Length ) - ilen,
				&chksum, &ilen2, GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
					session->UsedTransforms ) );     // generate integ checksum
		if( ilen != ilen2 || memcmp( in + ntohs( hdr->Length ) - ilen, chksum, ilen ) )
		{     // checksum ok?
			xlogf( I2L_ERR, "EAP-IKEv2: The integrity checksum is not correct.\n" );
			free( chksum );
			chksum=NULL;
			return -1;   // integrity checksum not ok
		}
		free( chksum ); // checksum is ok!!
		chksum=NULL;
	}
	else {
                if( session->expect_integ )
                {
                        xlogf( I2L_ERR, "EAP-IKEv2: Error: Message is NOT integrity protected - it should be.\n" );
                        return -1;
                }
        }

	uint8_t *frag;
	uint32_t flen;
	int result = ExtractData( in, &frag, &flen, ilen );  // extract the IKEv2 message from packet

	if( result )
		return result;
	if( session->fragment + flen > session->fdlen ) //received ealier + received now > lenght declared in frst framgment
	{
		xlogf( I2L_ERR, "EAP-IKEv2: Message length and fragments' length are inconsistent.\n" );
		return -1;
	}

	memcpy( session->fragdata + session->fragment, frag, flen );  // copy fragment to message
	free(frag);
	frag=NULL;
	session->fragment += flen;

	if( hdr->Flags & EAP_HDF_MOREFRAGMENTS ) // if not last fragment
	{
		*ikemsg = NULL;
		*imlen = 0;
		return 0;
	}

	// we've got all fragments here
	if( session->fdlen < sizeof( struct IKEv2Header ) || session->fdlen != session->fragment ||
			session->fdlen != ntohl( ( (struct IKEv2Header *)session->fragdata )->Length ) )  // check length fields
	{
		xlogf( I2L_ERR, "EAP-IKEv2: Length fields are inconsistent.\n" );
		*imlen = 0;
		free( *ikemsg );
		*ikemsg = NULL;
		return -1;
	}

	*ikemsg = session->fragdata;    // return ike message
	session->fragdata=NULL; // ikemsg jest zwalniane gdzie indziej
	*imlen = session->fdlen;
	 
	struct IKEv2PayloadHeader *pl = (struct IKEv2PayloadHeader *)( *ikemsg + sizeof( struct IKEv2Header ) );
	uint8_t nextpl = ( ( struct IKEv2Header *)*ikemsg )->NextPayload;
	//bool hasinteg = false;   // msg should be integrity protected??
        int encrypted=0; // is there an encypted payload?
                while( nextpl )       // search for encrypted payload
                {
			uint16_t pllen=ntohs(pl->PayloadLength);
			if(!pllen) {
			    // protection against infinite loop
			    xlogf(I2L_ERR,"EAP-IKEv2: Packet corrupted - paylod type %d length is %d\n",nextpl,pllen);
			    *imlen=0;
			    free(*ikemsg);
			    *ikemsg=NULL;
			    return -1;
			}
			if((uint8_t*)pl+pllen>(uint8_t*)*ikemsg+*imlen) {
			    // protection against buffer overflow
			    xlogf(I2L_ERR,"EAP-IKEv2: Packet corrupted - position exceeds packet length\n");
			    xlogf(I2L_DBG,"EAP-IKEv2: difference is=%d\n",(int)(*ikemsg+*imlen) -(int)((uint8_t*)pl +pllen));
			    *imlen=0;
			    free(*ikemsg);
			    *ikemsg=NULL;
			    return -1;
			}

                        if( nextpl == IKEv2_NPT_ENCRYPTED )
                        {
                                encrypted=1;
                                break;
                        }
                        nextpl = pl->NextPayload;
                        pl = ( (struct IKEv2PayloadHeader *)( ( (char *)pl ) + pllen) );
                }
        if (encrypted){
                if ( !(hdr->Flags & EAP_HDF_INTEGRITYINCLUDED) && 
                                session->expect_integ)
                {
                        xlogf( I2L_ERR, "EAP-IKEv2: There is no checksum in the message but should be.\n" );
                        *imlen = 0;
                        if( *ikemsg )
                                free( *ikemsg ); //??
                        *ikemsg = NULL;
                        return -1;
                }
                session->expect_integ=1;
                       //every next message have to be integrity protected
        }
//bobo
//        if (session->SK_ready){
//	        if( !has_chksum && hasinteg )
//                {   // there was no integ checksum, is that a problem??
//                        xlogf( I2L_DBG, "EAP-IKEv2: There is no checksum in the message but should be.\n" );
//                        *imlen = 0;
//                        if( *ikemsg )
//                                free( *ikemsg ); //??
//                        *ikemsg = NULL;
//                        return -1;
//                }
//        }
	return 0;
}

#define  EAP_RESULT_MESSAGE_SIZE 4
/**
 * @brief
 * Create success and failure messages 
 * 
 * @param succ    true success message, false  failure message  
 * @param session ptr to current session 
 * @param out     ptr to out message 
 *
 * @return length of the message ( EAP_RESULT_MESSAGE_SIZE ), or 0 if fail
 */
uint32_t CreateResultMessage( bool succ, struct IKEv2Session *session, uint8_t **out )
{ 
	xlogf( I2L_DBG, "EAP-IKEv2: Creating %s msg.\n", succ ? "success" : "failure" );

	if( !( *out = (uint8_t*)malloc( EAP_RESULT_MESSAGE_SIZE ) ) )
	{     
		xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't alloc mem for message.\n" );
		return 0;
	}
	struct EAPHeader *hdr = (struct EAPHeader *)*out;

	hdr->Code = succ ? EAP_CODE_SUCCESS : EAP_CODE_FAILURE;   // create a header
	hdr->Id = session->eapMsgID;
	hdr->Length = htons( EAP_RESULT_MESSAGE_SIZE );

	return EAP_RESULT_MESSAGE_SIZE;    // length of message (only header)
}


/**
 * @brief 
 * Create a fragment acknoledge message 
 *
 * @param *in       ptr to received message
 * @param **out     ptr to outcoming message
 * @param *session  session data
 *
 * @return length of the created message, or <=0 on failure
 * @todo check error status on return when called
 * 
 */
uint16_t CreateFragmentAck(const uint8_t *in, uint8_t **out, struct IKEv2Session *session )
{
	uint8_t *chksum;
	uint32_t ilen = 0;
	if( session->include_integ )
	{
		ilen = GetChecksumLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
					session->UsedTransforms ) );
	}

	if( !( *out = (uint8_t*)malloc( sizeof( struct EAPHeader ) + ilen ) ) )
	{
		xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't alloc mem for fragment ack msg.\n" );
		return 0;
	}

	struct EAPHeader *hdr = (struct EAPHeader *)*out;
	const struct EAPHeader *ihdr = (const struct EAPHeader *)in;

	hdr->Code = ihdr->Code == EAP_CODE_RESPONSE ? EAP_CODE_REQUEST : EAP_CODE_RESPONSE;
	hdr->Id = ihdr->Code == EAP_CODE_RESPONSE ? session->eapMsgID : ihdr->Id;
	hdr->Length = htons( sizeof( struct EAPHeader ) + ilen );
	hdr->Type = ihdr->Type;
	hdr->Flags = session->include_integ ? EAP_HDF_INTEGRITYINCLUDED : 0;

	if( session->include_integ )
	{
		if( !( session->fragdata ) )  // do we have keys?
			return -1;

		xlogf( I2L_DBG, "EAP-IKEv2: Making integrity checksum.\n" );
		uint32_t ilen2;
		integ( session->integkey, session->iklen, *out, sizeof( struct EAPHeader ),
				&chksum, &ilen2, GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
					session->UsedTransforms ) );     // generate integ checksum
		if( ilen != ilen2 )
		{     // checksum ok?
			xlogf( I2L_ERR, "EAP-IKEv2: Error: The integrity checksum is not correct.\n" );
			free( chksum );
			chksum=NULL;
			return 0;   // integrity checksum not ok
		}
		memcpy( *out + sizeof( struct EAPHeader ), chksum, ilen );
		free( chksum ); // checksum is ok!!
		chksum=NULL;
	}

	return sizeof( struct EAPHeader ) + ilen;
}

/**
 * @brief  Parse fragmen acknowledgement message
 *
 * @param *in       ptr to message 
 * @param *session  session data
 *
 * @return 0 on success, 1 otherwise
 */
int ParseFragmentAck(const uint8_t *in, struct IKEv2Session *session )
{
  uint8_t *chksum;
  uint32_t ilen = 0;
  ilen = GetChecksumLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
      session->UsedTransforms ) );
  const struct EAPHeader *hdr = (const struct EAPHeader *)in;

  if( ( hdr->Code != EAP_CODE_REQUEST && hdr->Code != EAP_CODE_RESPONSE ) ||
      hdr->Type != EAPIKEv2_TYPE || ( hdr->Flags & EAP_HDF_MOREFRAGMENTS ) ||
      ( !( hdr->Flags & EAP_HDF_INTEGRITYINCLUDED ) && session->expect_integ ) )
    return -1;

  if( !( session->fragdata ) )  // do we have keys?
    return -1;

  if( hdr->Flags & EAP_HDF_INTEGRITYINCLUDED )
  {
    uint32_t ilen2;
    integ( session->integkey, session->iklen, in, ntohs( hdr->Length ) - ilen,
        &chksum, &ilen2, GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
        session->UsedTransforms ) );     // generate integ checksum

    if( ilen != ilen2 || memcmp( in + ntohs( hdr->Length ) - ilen, chksum, ilen ) )
    {     // checksum ok?
      free( chksum );
      chksum=NULL;
      return -1;   // integrity checksum not ok
    }
    free( chksum ); // checksum is ok!!
    chksum=NULL;
  }

  return 0;
}

/**
 * @brief Create and init IKEv2Data structure
 *
 * @param *i2         IKEv2 context data 
 * @param *s          session data
 *
 * @return new allocated IKEv2Data structure
 */
struct IKEv2Data * IKEv2Data_new(struct ikev2_ctx *i2, struct IKEv2Session *s)
{
    struct IKEv2Data *result=(struct IKEv2Data*)malloc(sizeof(struct IKEv2Data));
    if(result==NULL) {
	return NULL;
    }
    result->i2=i2;
    result->session=s;
    return result;
}
