/*
 *  dh.c  -  functions supporting the diffie hellman exchange
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include"common.h"

#include <string.h>

#include "dh.h"
#include "dh_groups.h"
#include "enums.h"

DH *DHGenerateKEData_MODP( uint32_t DHGroup );
DH *DHGenerateKEData_EC2N( uint32_t DHGroup );

/**
 * @brief generate a DH public key for exchange  for MODP groups
 *
 * @param DHGroup DH prime modulus ID
 *
 * @return DH structure, or NULL in case of error
 */
DH *DHGenerateKEData_MODP( uint32_t DHGroup )
{
  DH *dh = DH_new();
  BIGNUM *p, *g;
  uint8_t ig = DHGROUP_GENERATOR;
  if(dh==NULL) {xlogf(I2L_ERR, "EAP-IKEv2: Error: DH_new()\n");return NULL;}

  switch( DHGroup )
  {
  case IKEv2_TID_DHG_768MODP:
    p = BN_bin2bn(MODP768_MODULUS,sizeof(MODP768_MODULUS),NULL);
    break;
    
  case IKEv2_TID_DHG_1024MODP:
    p = BN_bin2bn(MODP1024_MODULUS,sizeof(MODP1024_MODULUS),NULL);
    break;
    
  case IKEv2_TID_DHG_1536MODP:
    p = BN_bin2bn(MODP1536_MODULUS,sizeof(MODP1536_MODULUS),NULL);
    break;
    
  case IKEv2_TID_DHG_2048MODP:
    p = BN_bin2bn( MODP2048_MODULUS,sizeof(MODP2048_MODULUS),NULL);
    break;

  case IKEv2_TID_DHG_3072MODP:
    p = BN_bin2bn(MODP3072_MODULUS,sizeof(MODP3072_MODULUS),NULL);
    break;
    
  case IKEv2_TID_DHG_4096MODP:
    p = BN_bin2bn(MODP4096_MODULUS,sizeof(MODP4096_MODULUS),NULL);
    break;
    
  case IKEv2_TID_DHG_6144MODP:
    p = BN_bin2bn(MODP6144_MODULUS,sizeof(MODP6144_MODULUS),NULL);
    break;

  case IKEv2_TID_DHG_8192MODP:
    p = BN_bin2bn(MODP8192_MODULUS,sizeof(MODP8192_MODULUS),NULL);
    break;

  default:
    DH_free(dh);
    return NULL;
  }
  g = BN_bin2bn( &ig, sizeof( ig ), NULL );
  if( p==NULL || g==NULL ) {DH_free(dh);return NULL;}
  dh->p = p;
  dh->g = g;
  if(!DH_generate_key( dh )) {DH_free(dh);return NULL;}

  return dh;
}

/**
 * @brief
 * generate a DH public key for exchange for EC2N groups (elliptic curves)
 *
 * @deprecated
 * Not mentioned in IKEv2 algorithms requirments
 * 
 * @param  DHGroup DH group identifier
 *
 * @return DH group
 * @todo implement EC2N groups
 */
  
DH *DHGenerateKEData_EC2N( uint32_t DHGroup )
{
  xlogf(I2L_ERR,"EAP-IKEv2: Unimplemented DH group usage\n");
  DHGroup=DHGroup; // -Wunused  
  return NULL;                  // TODO: implement EC2N groups
}

/**
 * @brief 
 * generate a DH public key for exchange 
 *
 * @param DHGroup DH group identifier
 *
 * @return DH group
 */
DH *DHGenerateKEData( uint32_t DHGroup )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Generating DH exchange data.\n" );
  if( DHGroup == IKEv2_TID_DHG_155EC2N || DHGroup == IKEv2_TID_DHG_185EC2N )
    return DHGenerateKEData_EC2N( DHGroup );
  else
    return DHGenerateKEData_MODP( DHGroup );
}

/**
 * @brief 
 * compute a DH key from own dh info + the other's public dh key 
 *
 * @param *pubk other side public key
 * @param *se session data with private key
 */
void ComputeDHKey( BIGNUM *pubk, struct IKEv2Session *se )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Computing the shared DH key.\n" );
  if( !se->dh )
    return;
  if( se->DHKey ) {
    free( se->DHKey );
    se->DHKey=NULL;
  }
  se->DHKeyLen = DH_size( se->dh );
  if( !( se->DHKey = (uint8_t *)malloc( se->DHKeyLen + 1 ) ) )
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Can't allocate mem for DH-Key.\n" );
    return;
  }
  memset(se->DHKey,0,se->DHKeyLen);
  se->DHKeyLen=DH_compute_key( se->DHKey, pubk, se->dh );
  if((signed)se->DHKeyLen==-1 ) { //kwz: DHKeyLen is unsigned
    xlogf( I2L_ERR, "EAP-IKEv2: Error: DH_compute_key()\n" );
    free( se->DHKey );
    se->DHKey=NULL;
    se->DHKeyLen=0;
  }
}
