/*
 *  encr.c  -  functions for encryption algorithms
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include"common.h"

#include<openssl/des.h>
#include<openssl/evp.h>
#include<string.h>
#include<assert.h>
 
#include"encr.h"
#include "enums.h"

// private prototypes:

void Do3DES( uint8_t *data, uint32_t dlen, uint8_t *iv, uint8_t *key, int enc );
void DoEVPCrypt(const EVP_CIPHER *cipher, uint8_t *data, uint32_t dlen, uint8_t *iv, uint8_t *key, uint32_t keylen ,int enc);
const EVP_CIPHER* FindEVPCipher(uint32_t type,uint32_t keylen);

/**
 * @brief Get encryption key length.
 *
 * @param type    type of encryption algorithm 
 * @param keylen  key length in bites (optional)
 *
 * @return the key len of crypto algo in bytes
 */
inline uint32_t GetEncrKeyLen( uint32_t type , uint16_t keylen)
{
    if(!keylen) {
	switch( type ) { //constant key size
	    case IKEv2_TID_ENCR_3DES:
		return 24;    // including parity bits!!
	    case IKEv2_TID_ENCR_IDEA:
		return 16;
		
	}
    } else { // variable key size
	if(keylen%8) {
	    xlogf(I2L_ERR,"EAP-IKEv2: Keys which are not 8 multiple are not supported (keylen is %d bits) \n",keylen);
	    return 0;
	}
	return keylen/8;
    }
    return 0;
}

/**
 * @brief  Get length of block for specific encryption algorithm
 *
 * @param type    type of the encryption algorithm
 * @param keylen  key length of encryption algorithm (optional)
 *
 * @return block length of encryption algorithm 
 */
inline uint32_t GetEncrBlockLen( uint32_t type, uint32_t keylen )
{
  keylen=keylen; //-Wunused  
  switch( type ) {
  case IKEv2_TID_ENCR_3DES:
    return 8;
  case IKEv2_TID_ENCR_AES_CBC: 
    return 16;
  case IKEv2_TID_ENCR_BLOWFISH:
    return 8;
  case IKEv2_TID_ENCR_IDEA:
    return 8;
  case IKEv2_TID_ENCR_CAST:
    return 8;
  case IKEv2_TID_ENCR_NULL:
    return 1;
  }
  
  return 0;
}

/**
 * @brief Check for valid key length
 *
 * @param type    encryption algorithm
 * @param keylen  encryption algorithm key length in bites
 *
 * @return  0 on success, 1 otherwise
 */
int EncrKeyLenValid( uint32_t type, uint32_t keylen )
{
    switch(type) {
	case IKEv2_TID_ENCR_AES_CBC:
	    return keylen==128 || keylen==192 || keylen==256;
	case IKEv2_TID_ENCR_BLOWFISH:
	    return !(keylen % 8) && keylen>=8 && keylen<=448;
	case IKEv2_TID_ENCR_CAST:
	    return !(keylen %8) && keylen>=40 && keylen <=128;
	   
    }
    return 0;
}

/**
 * @brief encrypt / decrypt data with 3DES in ECB mode
 *
 * @param *data ptr to buffer to be encrypted
 * @param dlen len of buffer (has to be multiple of 8)
 * @param *iv init vector (unused)
 * @param *key ptr to key (has to be 24 byte = 3 x 64bit)
 * @param enc encrypt/decrypt
 */
void Do3DES( uint8_t *data, uint32_t dlen, uint8_t *iv, uint8_t *key, int enc )
{
  iv=iv; // -Wunused  
  uint32_t i;
  des_key_schedule k1, k2, k3;
  des_set_odd_parity( (des_cblock *)key );    // set parity, generate expanded keys
  des_key_sched( (des_cblock *)key, k1 );
  des_set_odd_parity( (des_cblock *)( key + 8 ) );
  des_key_sched( (des_cblock *)( key + 8 ), k2 );
  des_set_odd_parity( (des_cblock *)( key + 16 ) );
  des_key_sched( (des_cblock *)( key + 16 ), k3 );
  for( i = 0; i < dlen / 8; i++ )
    //des_ecb3_encrypt( (des_cblock *)( data + i * 8 ), (des_cblock *)( data + i * 8 ), k1, k2, k3, enc );
    des_ecb3_encrypt( (uint8_t *)( data + i * 8 ), (uint8_t *)( data + i * 8 ), k1, k2, k3, enc );
}

/**
 * @brief Encrypt/decrypt a block of data
 *
 * @param *data ptr to src buffer = dst buffer
 * @param dlen len of data (length should be a multiple of block len)
 * @param *iv initialization vector
 * @param *key key for cipher
 * @param type type of crypt algo
 * @param keylen key length parameter (optional)
 * @param enc IKEv2_ENCRYPT for encryption, IKEv2_DECRYPT for decryption
 */

void IKEv2Crypt( uint8_t *data, uint32_t dlen, uint8_t *iv, uint8_t *key, uint32_t type, uint32_t keylen,int enc)
{
    const EVP_CIPHER *cipher=FindEVPCipher(type,keylen);
    if(cipher) {
	DoEVPCrypt(cipher,data,dlen,iv,key,keylen,enc);
    } else {
	switch( type ) {
	    case IKEv2_TID_ENCR_3DES:
		Do3DES( data, dlen, iv, key, enc);
		break;

	    case IKEv2_TID_ENCR_NULL:
		xlogf(I2L_INFO,"EAP-IKEv2: Warning!!! Using NULL encryption!\n");
		break;
		
	    default:
		// this should never happen, because we are using previously
		// negotiated encryption algorithm
		xlogf( I2L_ERR, "EAP-IKEv2: Warning: Unsupported encryption algorithm, data will be transmitted unencrypted.\n" );
		break;
	}
    }
}

/**
 * @brief
 * Try to find encryption algorithm, which is supported by OpenSSL EVP interface
 * @param type encryption algorithm type
 * @param keylen length of key used by this alorithm
 * @return NULL if not found, or EVP_CIPHER struct which coresponds to algorithm
 */

const EVP_CIPHER* FindEVPCipher(uint32_t type,uint32_t keylen)
{
    const EVP_CIPHER *result=NULL;
    switch(type) {
#ifndef OPENSSL_NO_AES
	case IKEv2_TID_ENCR_AES_CBC:
	    switch(keylen) {
		case 128:
		    result=EVP_aes_128_cbc();
		    xlogf(I2L_DBG,"EAP-IKEv2: Using AES_128_CBC\n");
		    break;
		case 192:
		    result=EVP_aes_192_cbc();
		    xlogf(I2L_DBG,"EAP-IKEv2: Using AES_192_CBC\n");
		    break;
		case 256:
		    result=EVP_aes_256_cbc();
		    xlogf(I2L_DBG,"EAP-IKEv2: Using AES_256_CBC\n");
		    break;
		default:
		    xlogf(I2L_ERR,"EAP-IKEv2: Invalid AES CBC key length - %d\n",keylen);
	    }
	    break;
#endif
#ifndef OPENSSL_NO_BF
	case IKEv2_TID_ENCR_BLOWFISH:
	    xlogf(I2L_DBG,"EAP-IKEv2: Using BLOWFISH\n");
	    result=EVP_bf_cbc();
	    break;
#endif
#ifndef OPENSSL_NO_IDEA
	case IKEv2_TID_ENCR_IDEA:
	    xlogf(I2L_DBG,"EAP-IKEv2: Using IDEA\n");
	    result=EVP_idea_cbc();
	    break;
#endif
#ifndef OPENSSL_NO_CAST
	case IKEv2_TID_ENCR_CAST:
	    xlogf(I2L_DBG,"EAP-IKEv2: Using CAST\n");
	    result=EVP_cast5_cbc();
	    break;
#endif
    }
    if(result) {
	if(result->flags & EVP_CIPH_VARIABLE_LENGTH) {
	    xlogf(I2L_DBG,"EAP-IKEv2: variable key length cipher\n");
	} else {
	    xlogf(I2L_DBG,"EAP-IKEv2: constant key length cipher\n");
	}
    }
    return result;
}


/**
 * @brief encrypt/decrypt with EVP interface
 *
 * @param *cipher EVP cipher
 * @param *data   ptr to buffer to be encrypted
 * @param dlen    len of buffer (has to be multiple of 8)
 * @param *iv     init vector (unused)
 * @param *key    ptr to key (has to be 24 byte = 3 x 64bit)
 * @param keylen  encryption key length (in bits)
 * @param enc     encrypt/decrypt
 */
void DoEVPCrypt(const EVP_CIPHER *cipher, uint8_t *data, uint32_t dlen, uint8_t *iv, uint8_t *key, uint32_t keylen, int enc)
{
    assert(data && dlen && iv && key);
    assert(enc==IKEv2_ENCRYPT || enc==IKEv2_DECRYPT);
    EVP_CIPHER_CTX ctx;
    uint8_t *dpt;
    int ol;
    
    dpt=(uint8_t*)malloc(dlen);
    assert(dpt);
    assert(!(keylen % 8 ));
    if(enc==IKEv2_ENCRYPT) {
	//EVP_EncryptInit(&ctx,cipher,key,iv);
	EVP_EncryptInit(&ctx,cipher,NULL,NULL);
	EVP_CIPHER_CTX_set_key_length(&ctx,keylen/8);
	EVP_EncryptInit(&ctx,NULL,key,iv);
    } else {
	EVP_DecryptInit(&ctx,cipher,NULL,NULL);
	EVP_CIPHER_CTX_set_key_length(&ctx,keylen/8);
	EVP_DecryptInit(&ctx,NULL,key,iv);
    }
    xlogf(I2L_DBG,"EAP-IKEv2: Key length:%d\n",EVP_CIPHER_CTX_key_length(&ctx)*8);
    EVP_CIPHER_CTX_set_padding(&ctx,0);
    if(enc==IKEv2_ENCRYPT) {
	EVP_EncryptUpdate(&ctx,dpt,&ol,data,dlen);
    } else {
	EVP_DecryptUpdate(&ctx,dpt,&ol,data,dlen);
    }
    if(dlen!=(unsigned)ol) {
	xlogf(I2L_ERR,"EAP-IKEv2: Something wrong with encryption\n");
	free(dpt);
	return;
    }
    memcpy(data,dpt,dlen);
    free(dpt);
    EVP_CIPHER_CTX_cleanup(&ctx);
}
