/*
 *  hmac.c  -  implementation of hmac
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include"common.h"

#include<assert.h>
#include<stdlib.h>
#include<openssl/md5.h>
#include<openssl/sha.h>
#include<openssl/evp.h>
#include<strings.h>
#include<string.h>

#include"hmac.h"
#include"tiger.h"

/**
 * @brief hmac_md5 implementation
 *
 * @param text      pointer to data stream 
 * @param text_len  length of data stream 
 * @param key       pointer to authentication key 
 * @param key_len   length of authentication key 
 * @param digest    caller digest to be filled in 
 */                
void i2_hmac_md5( const unsigned char*  text, int text_len, const unsigned char* key, int key_len, uint8_t *digest)
{
        MD5_CTX context;
        unsigned char k_ipad[65];    /* inner padding -
                                      * key XORd with ipad
                                      */
        unsigned char k_opad[65];    /* outer padding -
                                      * key XORd with opad
                                      */
        unsigned char tk[16];
        int i;
        /* if key is longer than 64 bytes reset it to key=MD5(key) */
        if (key_len > 64) {

                MD5_CTX      tctx;

                MD5_Init(&tctx);
                MD5_Update(&tctx, key, key_len);
                MD5_Final(tk, &tctx);

                key = tk;
                key_len = 16;
        }

        /*
         * the HMAC_MD5 transform looks like:
         *
         * MD5(K XOR opad, MD5(K XOR ipad, text))
         *
         * where K is an n byte key
         * ipad is the byte 0x36 repeated 64 times
         * opad is the byte 0x5c repeated 64 times
         * and text is the data being protected
         */

        /* start out by storing key in pads */
        memset(k_ipad,0,sizeof k_ipad);
        memset(k_opad,0,sizeof k_opad);
        memcpy(k_ipad,key,key_len);
        memcpy(k_opad,key,key_len);

        /* XOR key with ipad and opad values */
        for (i=0; i<64; i++) {
                k_ipad[i] ^= 0x36;
                k_opad[i] ^= 0x5c;
        }
        /*
         * perform inner MD5
         */
        MD5_Init(&context);                   /* init context for 1st
                                                 pass */
        MD5_Update(&context, k_ipad, 64);     /* start with inner pad */
        MD5_Update(&context, text, text_len); /* then text of datagram */
        MD5_Final(digest, &context);          /* finish up 1st pass */
        /*
         * perform outer MD5
         */
        MD5_Init(&context);                   /* init context for 2nd
                                              * pass */
        MD5_Update(&context, k_opad, 64);     /* start with outer pad */
        MD5_Update(&context, digest, 16);     /* then results of 1st
                                              * hash */
        MD5_Final(digest, &context);          /* finish up 2nd pass */
}

/**
 * @brief hmac_sha1 implementation
 *
 * @param text      pointer to data stream 
 * @param text_len  length of data stream 
 * @param key       pointer to authentication key 
 * @param key_len   length of authentication key 
 * @param digest    caller digest to be filled in 
 */                

void i2_hmac_sha1( const unsigned char *text, int text_len, const unsigned char* key, int key_len, uint8_t *digest )
{
        SHA_CTX context;
        unsigned char k_ipad[65];    /* inner padding -
                                      * key XORd with ipad
                                      */
        unsigned char k_opad[65];    /* outer padding -
                                      * key XORd with opad
                                      */
        unsigned char tk[20];
        int i;
        /* if key is longer than 64 bytes reset it to key=SHA1(key) */
        if (key_len > 64) {

                SHA_CTX      tctx;

                SHA1_Init(&tctx);
                SHA1_Update(&tctx, key, key_len);
                SHA1_Final(tk, &tctx);

                key = tk;
                key_len = 20;
        }

        /*
         * the HMAC_SHA1 transform looks like:
         *
         * SHA1(K XOR opad, SHA1(K XOR ipad, text))
         *
         * where K is an n byte key
         * ipad is the byte 0x36 repeated 64 times
         * opad is the byte 0x5c repeated 64 times
         * and text is the data being protected
         */

        /* start out by storing key in pads */
        memset(k_ipad,0,sizeof k_ipad);
        memset(k_opad,0,sizeof k_opad);
        memcpy(k_ipad,key,key_len);
        memcpy(k_opad,key,key_len);

        /* XOR key with ipad and opad values */
        for (i=0; i<64; i++) {
                k_ipad[i] ^= 0x36;
                k_opad[i] ^= 0x5c;
        }
        /*
         * perform inner SHA1
         */
        SHA1_Init(&context);                   /* init context for 1st
                                                 pass */
        SHA1_Update(&context, k_ipad, 64);     /* start with inner pad */
        SHA1_Update(&context, text, text_len); /* then text of datagram */
        SHA1_Final(digest, &context);          /* finish up 1st pass */
        /*
         * perform outer SHA1
         */
        SHA1_Init(&context);                   /* init context for 2nd
                                              * pass */
        SHA1_Update(&context, k_opad, 64);     /* start with outer pad */
        SHA1_Update(&context, digest, 20);     /* then results of 1st
                                              * hash */
        SHA1_Final(digest, &context);          /* finish up 2nd pass */
}



/**
 * @brief
 * Implementation of MAC_DES integrity algorithm,
 * description: FIPS81 appendix F
 *
 * @param msg message
 * @param mlen message length
 * @param key integrity key - must be 8 bytes
 * @param digest MAC DES message digest
 */
void i2_des_mac(const uint8_t *msg,uint32_t mlen,const uint8_t *key, uint8_t *digest)
{
    assert(key && digest);
    assert((msg!=NULL) == (mlen!=0));

    EVP_CIPHER_CTX ctx;
    int ol;
    uint8_t dpt[8];
    uint8_t pad[8];
    const uint8_t *data;
    uint32_t i;
    
    EVP_EncryptInit(&ctx,EVP_des_cbc(),key,NULL);
    EVP_CIPHER_CTX_set_padding(&ctx,0);
    if(msg && mlen) {
	for(i=0;i<mlen;i+=8) {
	    if(mlen%8 && (i>=mlen-8 || mlen<8)) {
		// add 0's padding if needed
		memcpy(pad,msg+i,mlen%8);
		memset(pad+mlen%8,0,8-mlen%8);
		data=pad;
	    } else {
		data=msg+i;
	    }
	    EVP_EncryptUpdate(&ctx,dpt,&ol,data,8);
	}
    } else { 
	// special case (null message)
	memset(pad,0,8);
	EVP_EncryptUpdate(&ctx,dpt,&ol,pad,8);
    }
    memcpy(digest,dpt,8);
    EVP_CIPHER_CTX_cleanup(&ctx);
    
}


/**
 * @brief
 * HMAC_TIGER implementation based on RFC2104 rules.
 * Digest length is 192 bits (24 bytes)
 * @param text message text
 * @param text_len length of the message text
 * @param key HMAC key
 * @param key_len HMAC key length
 * @param digest result digest
 */
void i2_hmac_tiger(const uint8_t *text, int text_len, const uint8_t *key, int key_len, uint8_t *digest)
{
        unsigned char k_ipad[65];    /* inner padding -
                                      * key XORd with ipad
                                      */
        unsigned char k_opad[65+24];  /* outer padding -
                                      * key XORd with opad + extra space for
				      * inner digest
                                      */
        unsigned char tk[24];

        int i;
        /* if key is longer than 64 bytes reset it to key=MD5(key) */
        if (key_len > 64) {
	    tiger(key,key_len,tk);
	    key = tk;
	    key_len = 24;
        }

        /*
         * the HMAC_TIGER transform looks like:
         *
         * TIGER(K XOR opad, TIGER(K XOR ipad, text))
         *
         * where K is an n byte key
         * ipad is the byte 0x36 repeated 64 times
         * opad is the byte 0x5c repeated 64 times
         * and text is the data being protected
         */

        /* start out by storing key in pads */
        memset(k_ipad,0,sizeof k_ipad);
        memset(k_opad,0,sizeof k_opad);
        memcpy(k_ipad,key,key_len);
        memcpy(k_opad,key,key_len);

        /* XOR key with ipad and opad values */
        for (i=0; i<64; i++) {
                k_ipad[i] ^= 0x36;
                k_opad[i] ^= 0x5c;
        }

        /*
         * perform inner TIGER
         */
	uint8_t *buf=(uint8_t*)malloc(text_len+64);
	memcpy(buf,k_ipad,64);
	memcpy(buf+64,text,text_len);
	tiger(buf,text_len+64,k_opad+64);
	free(buf);
	
        /*
         * perform outer TIGER
         */
	tiger(k_opad,64+24,digest);
}
