/*
 *  integ.c  -  functions supporting the integrity protection
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include"common.h"

#include <stdlib.h>

#include "integ.h"
#include "hmac.h"
#include "enums.h"

/**
 * @brief get the key length for a given integrity algorithm
 *
 * @param type type of the integrity algorithm
 *
 * @return  integrity key length in bytes
 */
inline uint32_t GetIntegKeyLen( uint32_t type )
{
  switch( type )
  {
  case IKEv2_TID_AUTH_HMAC_MD5_96:
    return 16;
  case IKEv2_TID_AUTH_HMAC_SHA1_96:
    return 20;
  case IKEv2_TID_AUTH_DES_MAC:
    return 8;
  }
  return 0;
}

/**
 * @brief Get checksum length of a given integrity algorithm
 *
 * @param type type of integrity algorithm
 *
 * @return  length in bytes
 */
inline uint32_t GetChecksumLen( uint32_t type )
{
  switch( type )
  {
  case IKEv2_TID_AUTH_HMAC_MD5_96:
    return 12;
  case IKEv2_TID_AUTH_HMAC_SHA1_96:
    return 12;
  case IKEv2_TID_AUTH_DES_MAC:
    return 8;
  }
  return 0;
}

/**
 * @brief Get digest length used by sepcified integrity algorithm
 *
 * @param type integrity algorithm type
 *
 * @return length of digest in bytes
 */
inline uint32_t GetIntegDigestLen(uint32_t type) 
{
    switch(type)  {
	case IKEv2_TID_AUTH_HMAC_MD5_96:
	    return MD5DIGEST_SIZE;
	case IKEv2_TID_AUTH_HMAC_SHA1_96:
	    return SHA1DIGEST_SIZE;
	case IKEv2_TID_AUTH_DES_MAC:
	    return DESMACDIGEST_SIZE;
    }
    return 0;
}

/**
 *  @brief Check for valid key length 
 *  
 *  @param type integrity algorithm
 *  @param keylen key length in bytes
 *  
 *  @return 1 on succes, 0 otherwise
 */
inline int IntegKeyLenValid( uint32_t type, uint32_t keylen )
{
    return GetIntegKeyLen(type)==keylen;
}


/* @brief Perform integrity algorithm
 * 
 * @param *key       key for algo
 * @param klen       length of key
 * @param *text      data to be integrity checked
 * @param tlen       length of text
 * @param **res      result
 * @param *rlen      length of result (output)
 * @param integtype  type of integ algo (see IKEv2 specification)  
 */
void integ( const uint8_t *key, uint32_t klen, const uint8_t *text, uint32_t tlen,
	uint8_t **res, uint32_t *rlen, uint32_t integtype )
{
    uint8_t *buf;

    if(!IntegKeyLenValid(integtype,klen)) {
	xlogf(I2L_ERR,"EAP-IKEv2: Incorrect key length for integrity algorithm\n");
	*res=NULL;
	*rlen=0;
	return;

    }
    if(!(buf=(uint8_t*)malloc(GetIntegDigestLen(integtype)))) {
	xlogf( I2L_ERR, "EAP-IKEv2: Can't allocate mem for integ checksum.\n" );
	*res = NULL;
	*rlen = 0;
	return;
    }
    *rlen=GetChecksumLen(integtype);
    if(!*rlen) {
	xlogf(I2L_ERR,"EAP-IKEv2: Unsupported integrity algorithm\n");
	*res=NULL;
	*rlen=0;
	return;
    }

    switch( integtype )
    {
	case IKEv2_TID_AUTH_HMAC_MD5_96:
	    i2_hmac_md5( text, tlen, key, klen, buf );
	    break;

	case IKEv2_TID_AUTH_HMAC_SHA1_96:
	    i2_hmac_sha1( text, tlen, key, klen, buf );
	    break;
	case IKEv2_TID_AUTH_DES_MAC:
	    i2_des_mac(text,tlen,key,buf);
	    break;

	default:
	    *res = NULL;
	    *rlen = 0;
	    break;
    }
    *res=Clone(buf,*rlen); //kwz: Clone is not needed?
    free(buf);
    buf=NULL;
}
