/*
 *  message.c  -  functions to generate / parse all msg'es
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include"common.h"

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif //HAVE_NETINET_IN_H
#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif // HAVE_WINSOCK_H

#include <string.h>
#include <assert.h>

#include "message.h"
#include "payload.h"
#include "session.h"
#include "integ.h"
#include "auth.h"
#include "dh.h"
#include "enums.h"


// private prototypes:

uint16_t GenerateErrorData( uint8_t **data, uint16_t *dlen, struct  ParseMsgResult *res );
void MergeParseMsgResults( struct ParseMsgResult *result, struct ParseMsgResult *result2 );
void FreeCertList(struct CertList *l);
struct ParseMsgResult *ParseMsg( const struct ikev2_ctx *i2, struct IKEv2Session *session, uint8_t *p);
struct Proposal *CloneProposal( struct Proposal *prop, struct Proposal *ref );


/**
 * @brief Create an IKEv2 SA_INIT msg for initiator
 *
 * @param *i2       IKEv2 context data
 * @param **out     result message
 * @param *olen     result length
 * @param *session  current IKEv2 session data
 * @param DHGroup   DH group identifier
 *
 * @return  0 on success
 */
int IKEv2IBuild_SA_Init( const struct ikev2_ctx *i2, uint8_t **out, uint32_t *olen, struct IKEv2Session *session, uint16_t DHGroup )
{
  uint8_t *buf;
  uint8_t *p;

  xlogf( I2L_DBG, "EAP-IKEv2: Building SA Init msg.\n" );

  // alloc memory for msg

  *olen = 0;

  if( !( buf = (uint8_t*)malloc( TB_SIZE ) ) )     // temp buffer, don't forget to free it!
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }

  p = buf;
  struct IKEv2Header *hdr = (struct IKEv2Header *)p;

  // create msg header

  hdr->InitiatorSPI = hton64( session->iSPI );
  hdr->ResponderSPI = 0;
  hdr->MjVersion = IKEv2_MAJOR_VERSION;
  hdr->MnVersion = IKEv2_MINOR_VERSION;
  hdr->ExchangeType = IKEv2_EXT_IKE_SA_INIT;
  hdr->Flags = IKEv2_HDF_INITIATOR;
  hdr->MessageID = htonl( session->MsgID++ );

  p += sizeof( struct IKEv2Header );
  uint8_t *next;
  next=&hdr->NextPayload;

  // create payloads

  p += CreateSAPayload( &next, i2->suppProp, p );  // create SAi1
  uint16_t dhg = DHGroup ? DHGroup : GetTransformID(
      IKEv2_TRT_DIFFIE_HELLMAN_GROUP, IKEv2_PID_IKE_SA, i2->suppProp->protocols );
  p += CreateKEPayload( &next, dhg, session, p ); // create KEi
  p += CreateNoncePayload( &next ,session, NONCE_LEN, p );  // create Ni

  // set length and generate packet

  *olen = p - buf;
  hdr->Length = htonl( *olen );

  *out=buf; 
  buf=NULL;

  if( session->fmsg ) {       // a copy of this msg is stored for auth
    free( session->fmsg );
    session->fmsg=NULL;
  }
  session->fmsg = Clone( *out, *olen );
  session->fmlen = *olen;

  return IKEv2_RET_OK;
}

/**
 * @brief Create an IKEv2 SA_INIT msg for responder.
 *
 * @param *i2        IKEv2 context data
 * @param *oldMsg    received message, for which a respond should be created
 * @param **out      result message
 * @param *olen      result length
 * @param *selTrans  selected transforms  
 * @param *session   current IKEv2 session
 * @param *pubk      DH public key
 *
 * @return 0 on success
 */
int IKEv2RBuild_SA_Init( const struct ikev2_ctx *i2, uint8_t *oldMsg, uint8_t **out, uint32_t *olen,
    struct Proposal *selTrans, struct IKEv2Session *session, BIGNUM *pubk)
{
  uint8_t *buf;
  uint8_t *p;

  xlogf( I2L_DBG, "EAP-IKEv2: Building SA Init msg.\n" );

  // alloc memory for msg

  *olen = 0;

  if( !( buf = (uint8_t*)malloc( TB_SIZE ) ) )     // temp buffer, don't forget to free it!
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }

  p = buf;

  memcpy( p, oldMsg, sizeof( struct IKEv2Header ) );
  
  struct IKEv2Header *hdr = (struct IKEv2Header *)p;

  // create msg header

  hdr->ResponderSPI = hton64( session->rSPI );
  hdr->MjVersion = IKEv2_MAJOR_VERSION;
  hdr->MnVersion = IKEv2_MINOR_VERSION;
  hdr->Flags = IKEv2_HDF_RESPONSE;


  p += sizeof( struct IKEv2Header );

  // create payloads

  uint8_t *next;
  next=&hdr->NextPayload;

  p += CreateSAPayload( &next, selTrans, p );  // create SAr1
  p += CreateKEPayload( &next, GetTransformID( IKEv2_TRT_DIFFIE_HELLMAN_GROUP, IKEv2_PID_IKE_SA,
      session->UsedTransforms ), session, p ); // create KEr

  if(i2->sendCertReq && i2->CAsHashStringLen==0) {
      xlogf(I2L_ERR,"EAP-IKEv2: CER_REQ: No data for Certification Authority field!\n");
  }
  p += CreateNoncePayload(&next,session,NONCE_LEN, p );  // create Nr

  if(i2->CAsHashStringLen>0 && i2->sendCertReq) {
      p+=CreateCertReqPayload(i2,&next,p);
  }

   if(session->SK_ready==0){
      ComputeDHKey( pubk, session );   // compute all keys
      ComputeSessionKeys( session );
   }
   if(i2->authtype==IKEv2_AUTH_SK && i2->send_sk_idr) { //include SK{IDr} payload?
       uint8_t encrypted_next_val;
       uint8_t *encrypted_next=&encrypted_next_val;

           uint8_t *innerp;
           int len;
           if( !( innerp = (uint8_t*)malloc( TB_SIZE ) ) )    // buffer for payloads to encrypt
           {
             *out = NULL;
             free(buf);
             xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
             return IKEv2_RET_ERRMEMALLOC;
           }
           len = CreateIDPayload( &encrypted_next, session->Type==IKEv2_STY_INITIATOR,i2->idtype, i2->id, i2->idlen, innerp );

            // encrypt payloads
          
            uint32_t enclen =  CreateEncryptedPayload( &next, encrypted_next_val, innerp, len, buf, session, p );
	    p+=enclen;
            free(innerp);
            
   } 
  
  // set length and generate packet

   *olen = p - buf;
   hdr->Length = htonl( *olen );

   *out=buf;
   buf=NULL;

  if( session->fmsg ) {      // a copy of this msg is stored for auth
    free( session->fmsg );
    session->fmsg=NULL;
  }
  session->fmsg = Clone( *out, *olen );
  session->fmlen = *olen;
  
  return IKEv2_RET_OK;
}

/**
 * @brief Create an IKEv2 AUTH msg for initiator
 *
 * @param *i2       IKEv2 context data
 * @param **out     ptr to a char* variable, which will be filled with the
 *                  buffer start address
 * @param *olen     out data len
 * @param *session  current IKEv2 session
 *
 * @return 0 on success
 */
int IKEv2IBuild_Auth( const struct ikev2_ctx *i2, uint8_t **out, uint32_t *olen, struct IKEv2Session *session )
{
  uint8_t *buf;
  uint8_t *p;

  xlogf( I2L_DBG, "EAP-IKEv2: Building AUTH msg.\n" );

  // alloc memory for msg

  *olen = 0;

  if( !( buf = (uint8_t*)malloc( TB_SIZE ) ) )     // temp buffer, don't forget to free it!
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }

  p = buf;
  struct IKEv2Header *hdr = (struct IKEv2Header *)p;

  // create msg header
  hdr->InitiatorSPI = hton64( session->iSPI );
  hdr->ResponderSPI = hton64( session->rSPI );
  hdr->MjVersion = IKEv2_MAJOR_VERSION;
  hdr->MnVersion = IKEv2_MINOR_VERSION;
  hdr->ExchangeType = IKEv2_EXT_IKE_AUTH;
  hdr->Flags = IKEv2_HDF_INITIATOR;
  hdr->MessageID = htonl( session->MsgID++ );

  p += sizeof( struct IKEv2Header );

  // create payloads

  uint32_t len;
  uint8_t *innerp;
  uint8_t *idData;
  uint16_t idDataLength;
  if( !( innerp = (uint8_t*)malloc( TB_SIZE ) ) )    // buffer for payloads to encrypt
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }
  uint8_t encrypted_next_val;
  uint8_t *encrypted_next=&encrypted_next_val;
  uint8_t *next=&hdr->NextPayload;

  len = CreateIDPayload(&encrypted_next,session->Type==IKEv2_STY_INITIATOR, i2->idtype, i2->id, i2->idlen, innerp );
  idData=innerp+sizeof(struct IKEv2PayloadHeader);
  idDataLength=len-sizeof(struct IKEv2PayloadHeader);

  if(i2->authtype == IKEv2_AUTH_CERT && i2->certfile) {
      len += CreateCertPayload( i2, &encrypted_next, IKEv2_CET_X509_SIGNATURE, innerp + len );
  }
  if( i2->sendCertReq ) {
    len+=CreateCertReqPayload(i2,&encrypted_next,innerp+len);
  }
  
  len += CreateAuthPayload( i2, &encrypted_next, session,
      idData, idDataLength, innerp + len );

  // encrypt payloads

  uint32_t enclen =  CreateEncryptedPayload( &next, encrypted_next_val, innerp, len, buf,  session, p );
  free(innerp);
  innerp=NULL;
    
  // generate packet
  
  *olen = p - buf + enclen ;
  *out=buf;
  return IKEv2_RET_OK;
}

/**
 * @brief Create an IKEv2 AUTH msg for responder
 *
 * @param *i2       IKEv2 context data
 * @param **out     ptr to a char* variable, which will be filled with the
 *                  buffer start address
 * @param *olen     out data len
 * @param *in       msg to respond
 * @param *session  current IKEv2 session
 *
 * @return 0 on success
 */
int IKEv2RBuild_Auth( const struct ikev2_ctx *i2, uint8_t **out, uint32_t *olen, uint8_t *in, struct IKEv2Session *session )
{
  uint8_t *buf;
  uint8_t *p;

  xlogf( I2L_DBG, "EAP-IKEv2: Building AUTH msg.\n" );

  // alloc memory for msg

  *olen = 0;

  if( !( buf = (uint8_t*)malloc( TB_SIZE ) ) )     // temp buffer, don't forget to free it!
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }

  p = buf;
  struct IKEv2Header *hdr = (struct IKEv2Header *)p;

  // create msg header
  memcpy( p, in, sizeof( struct IKEv2Header ) ); // build hdr
  hdr->MjVersion = IKEv2_MAJOR_VERSION;
  hdr->MnVersion = IKEv2_MINOR_VERSION;
  hdr->ExchangeType = IKEv2_EXT_IKE_AUTH;
  hdr->Flags = IKEv2_HDF_RESPONSE;

  p += sizeof( struct IKEv2Header );

  // create payloads

  uint16_t idDataLength;
  uint32_t len;
  uint8_t *innerp;
  uint8_t *idData;
  if( !( innerp = (uint8_t*)malloc( TB_SIZE ) ) )    // buffer for payloads to encrypt
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }
  uint8_t encrypted_next_val;
  uint8_t *encrypted_next=&encrypted_next_val;
  uint8_t *next=&hdr->NextPayload;

  len = CreateIDPayload(&encrypted_next, session->Type==IKEv2_STY_INITIATOR, i2->idtype, i2->id, i2->idlen, innerp );
  idData=innerp+sizeof(struct IKEv2PayloadHeader);
  idDataLength=len-sizeof(struct IKEv2PayloadHeader);
  
  if( i2->certfile ){
      len += CreateCertPayload( i2, &encrypted_next, IKEv2_CET_X509_SIGNATURE, innerp + len );
  }
  len += CreateAuthPayload( i2, &encrypted_next, session,
	  idData,idDataLength, innerp + len );

  // encrypt payloads

  uint32_t enclen =  CreateEncryptedPayload( &next, encrypted_next_val, innerp, len,buf, session, p );
  free(innerp);
  innerp=NULL;

  // generate packet

  *olen = p - buf + enclen ;
  *out=buf;
  return IKEv2_RET_OK;
}

/**
 * @brief Get an error and generate correspondig error data
 *
 * @param **data ptr to error data
 * @param *dlen  length of data
 * @param *res   result struct from parse msg 
 *
 * @return type of error
 */
uint16_t GenerateErrorData( uint8_t **data, uint16_t *dlen, struct ParseMsgResult *res )
{
  *dlen = 0;
  *data=NULL;
  if( res->unsuppCritPl )   // unsupported critical payload (C-Flag set)
  {
    res->unsuppCritPl = 0;
    if( !( *data = (uint8_t*)malloc( sizeof(uint8_t) ) ) ) // allocate 1 byte
      return 0;
    *dlen = sizeof(uint8_t);
    **data = res->plType;
    return IKEv2_NMT_UNSUPPORTED_CRITICAL_PAYLOAD;
  }

  if( res->NoPropAcceptable ) {   // no proposal is acceptable
    return IKEv2_NMT_NO_PROPOSAL_CHOSEN;
  }

  if( res->KEInvalid ) {
    if( !( *data = (uint8_t*)malloc( sizeof(uint16_t) ) ) ) // allocate place for two 16bit dhgroup identifier
      return 0;
    *( (uint16_t *)*data ) = htons( res->ChosenDHGroup );
    *dlen = sizeof(uint16_t);
    return IKEv2_NMT_INVALID_KE_PAYLOAD;
  }
  
  if( res->Authfailed ) {      // authentication failed
    return IKEv2_NMT_AUTHENTICATION_FAILED;
  }

  return IKEv2_NMT_INVALID_SYNTAX;
}

/**
 * @brief Build IKEv2 message with error notification
 *
 * @param **out     ptr to a char* variable, which will be filled with the
 *                  buffer start address
 * @param *olen     out data len
 * @param *session  current IKEv2 session
 * @param *res      result struct from parse msg 
 *
 * @return 0 on success
 */
int IKEv2BuildErrorMsg( uint8_t **out, uint32_t *olen, struct IKEv2Session *session,
    struct ParseMsgResult *res )
{
  xlogf( I2L_INFO, "EAP-IKEv2: Building error msg.\n" );

  uint8_t *pck, *data;
  uint16_t dlen, type;
  if( !( pck = (uint8_t*)malloc( 512 + dlen ) ) )
  {
    *out = NULL;
    *olen = 0;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }
  if( !( type = GenerateErrorData( &data, &dlen, res ) ) )
  {     // there is no error, nothing to send....
    free( pck );
    pck=NULL;
    *out = NULL;
    *olen = 0;
    return IKEv2_RET_OK;
  }
  
  uint8_t informational_next_val;
  uint8_t *informational_next=&informational_next_val;
  uint32_t len = CreateNotifyPayload( &informational_next, type, data, dlen, pck );
  int result = IKEv2Build_Informational( out, olen, session, informational_next_val, pck, len );
  free( pck );
  pck=NULL;
  return result;
}
    
/**
 * @brief Build a response to a msg containing an error notification
 *
 * @param **out     ptr to a char* variable, which will be filled with the
 *                  buffer start address
 * @param *olen     out data len
 * @param *in       msg to respond
 * @param *res      result struct for parsing 'in'
 * @param *session  current IKEv2 session
 * @param enc       encrypt?
 *
 * @return 0 on success
 */
int IKEv2BuildErrorResponseMsg( uint8_t **out, uint32_t *olen, uint8_t *in, struct ParseMsgResult *res,
    struct IKEv2Session *session, int enc )
{
  uint8_t *buf;
  uint8_t *p;

  xlogf( I2L_INFO, "EAP-IKEv2: Building error response msg.\n" );

  // alloc memory for msg

  *olen = 0;

  if( !( buf = (uint8_t*)malloc( 1024 ) ) )     // temp buffer, don't forget to free it!
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }

  p = buf;
  struct IKEv2Header *hdr = (struct IKEv2Header *)p;

  // create msg header
  memcpy( p, in, sizeof( struct IKEv2Header ) );
  hdr->MjVersion = IKEv2_MAJOR_VERSION;
  hdr->MnVersion = IKEv2_MINOR_VERSION;
  hdr->Flags = IKEv2_HDF_RESPONSE;

  p += sizeof( struct IKEv2Header );
  uint8_t *next=&hdr->NextPayload;
  uint8_t encrypted_next_val;
  uint8_t *encrypted_next=&encrypted_next_val;

  // create payloads
  uint8_t *innerp;
  if( !( innerp = (uint8_t*)malloc( 1024 ) ) )    // buffer for payloads to encrypt
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }
  uint8_t *data;
  uint16_t dlen, type;
  uint32_t len;
  type = GenerateErrorData( &data, &dlen, res );
  len = CreateNotifyPayload( &encrypted_next, type, data, dlen, innerp );
  if( data ) {
    free( data );
    data=NULL;
  }
  
  if( !len )           // no error, nothing to be done
  {
    free( innerp );
    innerp=NULL;
    free( buf );
    buf=NULL;
    *out = NULL;
    return IKEv2_RET_OK;
  }

  // encrypt payloads, if necessary
  uint32_t enclen;
  if( enc )
  {
    enclen =  CreateEncryptedPayload( &next,encrypted_next_val, innerp, len,buf, session, p );
    p+=enclen;
  }
  else
  {
    memcpy( p, innerp, len );
    hdr->NextPayload = encrypted_next_val;
    p+=len;
    hdr->Length = htonl( p-buf );
  }
  free(innerp);

  // generate packet

  *olen = p - buf;
  *out=buf;
  return IKEv2_RET_OK;
}

/**
 * @brief Create an IKEv2 informational message
 *
 * @param **out     ptr to a char* variable, which will be filled with the
 *                  buffer start address
 * @param *olen     out data len
 * @param *session  current IKEv2 session
 * @param first     type of first payload
 * @param *payloads ptr to payloads for informational exchange
 * @param plsize    size of payloads
 *
 * @return 0 on success
 */
int IKEv2Build_Informational( uint8_t **out, uint32_t *olen, struct IKEv2Session *session,
    uint8_t first, uint8_t *payloads, uint32_t plsize )
{
  uint8_t *buf;
  uint8_t *p;

  xlogf( I2L_DBG, "EAP-IKEv2: Building Informational msg.\n" );

  // alloc memory for msg

  *olen = 0;

  if( !( buf = (uint8_t*)malloc( 1024 + plsize ) ) )     // temp buffer, don't forget to free it!
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }

  p = buf;
  struct IKEv2Header *hdr = (struct IKEv2Header *)p;

  // create msg header
  hdr->InitiatorSPI = hton64( session->iSPI );
  hdr->ResponderSPI = hton64( session->rSPI );
  hdr->MjVersion = IKEv2_MAJOR_VERSION;
  hdr->MnVersion = IKEv2_MINOR_VERSION;
  hdr->ExchangeType = IKEv2_EXT_INFORMATIONAL;
  hdr->Flags = IKEv2_HDF_INITIATOR;
  hdr->MessageID = htonl( session->MsgID++ );

  p += sizeof( struct IKEv2Header );

  // encrypt payloads

  uint8_t *next=&hdr->NextPayload;
  uint32_t enclen =  CreateEncryptedPayload( &next,first, payloads, plsize,buf, session, p );

  // generate packet

  *olen = p - buf + enclen;
  *out=buf;
  return IKEv2_RET_OK;
}

/**
 * @brief Create a response to a IKEv2 informational message
 *
 * @param **out     ptr to a char* variable, which will be filled with the
 *                  buffer start address
 * @param *olen     out data len
 * @param *inf      received informational message
 *
 * @return 0 in success
 */
int IKEv2RespondInformational( uint8_t **out, uint32_t *olen, uint8_t *inf )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Building informational response msg.\n" );

  *olen = sizeof( struct IKEv2Header );
  if( !( *out = (uint8_t*)malloc( *olen ) ) )
  {
    *out = NULL;
    *olen = 0;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }

  memcpy( *out, inf, *olen );
  struct IKEv2Header *hdr = (struct IKEv2Header *)*out;
  hdr->NextPayload = IKEv2_NPT_NONE;
  hdr->MjVersion = IKEv2_MAJOR_VERSION;
  hdr->MnVersion = IKEv2_MINOR_VERSION;
  hdr->Flags = IKEv2_HDF_RESPONSE;
  hdr->Length = htonl( *olen );
  
  return IKEv2_RET_OK;
}

// some helper functions for informational exchanges.....

#if 0
/**
 * @brief Create an IKEv2 informational message containing a delete payload for
 * one SPI
 *
 * @param **out     ptr to a char* variable, which will be filled with the
 *                  buffer start address
 * @param *olen     out data len
 * @param *session  current IKEv2 session
 * @param SPI       SPI 
 *
 * @return 0 in success
 */
int IKEv2BuildEndSession( uint8_t **out, uint32_t *olen, struct IKEv2Session *session, uint64_t SPI )
{
  uint8_t *pck;
  if( !( pck = (uint8_t*)malloc( 512 ) ) )
  {
    *out = NULL;
    *olen = 0;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }
  uint32_t len = CreateDeletePayload( IKEv2_NPT_NONE, IKEv2_PID_IKE_SA, (uint8_t *)&SPI, 1, pck );
  int res = IKEv2Build_Informational( out, olen, session, IKEv2_NPT_DELETE, pck, len );
  free( pck );
  pck=NULL;
  return res;
}
#endif

/**
 * @brief Create an IKEv2 CREATE_CHILD_SA initiator message for rekeying
 *
 * @param *i2       IKEv2 context data
 * @param **out     ptr to a char* variable, which will be filled with the
 *                  buffer start address
 * @param *olen     out data len
 * @param *session  current IKEv2 session
 * @param DHGroup   which DH group should we use for dh exchange?
 *
 * @return 0 in success
 */
int IKEv2ICreateChildSA( const struct ikev2_ctx *i2, uint8_t **out, uint32_t *olen, struct IKEv2Session *session, int DHGroup )
{
  uint8_t *buf;
  uint8_t *p;
  int dhex=session->fDHExchange;

  xlogf( I2L_DBG, "EAP-IKEv2: Building create_child_sa msg.\n" );

  // alloc memory for msg

  *olen = 0;

  if( !( buf = (uint8_t*)malloc( TB_SIZE ) ) )     // temp buffer, don't forget to free it!
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }

  p = buf;
  struct IKEv2Header *hdr = (struct IKEv2Header *)p;

  // create msg header
  hdr->InitiatorSPI = hton64( session->iSPI );
  hdr->ResponderSPI = hton64( session->rSPI );
  hdr->MjVersion = IKEv2_MAJOR_VERSION;
  hdr->MnVersion = IKEv2_MINOR_VERSION;
  hdr->ExchangeType = IKEv2_EXT_CREATE_CHILD_SA;
  hdr->Flags =  IKEv2_HDF_INITIATOR;
  hdr->MessageID = htonl( session->MsgID++ );

  p += sizeof( struct IKEv2Header );

  // create payloads
  uint8_t *next=&hdr->NextPayload;
  uint8_t encrypted_next_val;
  uint8_t *encrypted_next=&encrypted_next_val;

  uint32_t len;
  uint8_t *innerp;
  if( !( innerp = (uint8_t*)malloc( TB_SIZE ) ) )    // buffer for payloads to encrypt
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }
  len = CreateSAPayload( &encrypted_next, i2->suppProp, innerp );

  len += CreateNoncePayload( &encrypted_next,session, NONCE_LEN, innerp + len );

  if( dhex )
  {
    uint16_t dhg = DHGroup ? DHGroup : GetTransformID(
        IKEv2_TRT_DIFFIE_HELLMAN_GROUP, IKEv2_PID_IKE_SA, i2->suppProp->protocols );
    uint16_t kelen;
    kelen=CreateKEPayload( &encrypted_next, dhg, session, innerp + len );
    if(!kelen) {
	return -1;
    } else {
	len+=kelen;
    }
  }

  // encrypt payloads

  uint32_t enclen =  CreateEncryptedPayload( &next,encrypted_next_val, innerp, len, buf,session, p );
  free(innerp);

  // generate packet
  *olen = p - buf + enclen;
  *out=buf;
  return IKEv2_RET_OK;
}


/**
 * @brief Create an IKEv2 CREATE_CHILD_SA responder message for rekeying
 *
 * @param *oldMsg   ??ptr to the old IKEv2 message 
 * @param **out     ptr to a char* variable, which will be filled with the
 *                  buffer start address
 * @param *olen     out data len
 * @param *session  current IKEv2 session
 * @param *selTrans selected transforms  
 * @param DHGroup   which DH group should we use for dh exchange?
 *
 * @return 0 on success 
 */
int IKEv2RCreateChildSA( uint8_t *oldMsg, uint8_t **out, uint32_t *olen, struct IKEv2Session *session,
    struct Proposal *selTrans, int DHGroup )
{
  DHGroup=DHGroup; // -Wunused  
  uint8_t *buf;
  uint8_t *p;
  int dhex=session->fDHExchange;

  xlogf( I2L_DBG, "EAP-IKEv2: Building create_child_sa response.\n" );

  // alloc memory for msg

  *olen = 0;

  if( !( buf = (uint8_t*)malloc( TB_SIZE ) ) )     // temp buffer, don't forget to free it!
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }

  p = buf;
  struct IKEv2Header *hdr = (struct IKEv2Header *)p;

  // create msg header
  memcpy( p, oldMsg, sizeof( struct IKEv2Header ) );

  hdr->InitiatorSPI = hton64( session->iSPI );
  hdr->ResponderSPI = hton64( session->rSPI );
  hdr->MjVersion = IKEv2_MAJOR_VERSION;
  hdr->MnVersion = IKEv2_MINOR_VERSION;
  hdr->ExchangeType = IKEv2_EXT_CREATE_CHILD_SA;
  hdr->Flags = IKEv2_HDF_RESPONSE;

  p += sizeof( struct IKEv2Header );

  uint8_t *next=&hdr->NextPayload;
  uint8_t encrypted_next_val;
  uint8_t *encrypted_next=&encrypted_next_val;

  // create payloads

  uint32_t len;
  uint8_t *innerp;
  if( !( innerp = (uint8_t*)malloc( TB_SIZE ) ) )    // buffer for payloads to encrypt
  {
    *out = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for msg buffer.\n" );
    return IKEv2_RET_ERRMEMALLOC;
  }
  len = CreateSAPayload( &encrypted_next, selTrans, innerp );
  len += CreateNoncePayload( &encrypted_next ,session, NONCE_LEN, innerp + len );

  if( dhex )
  {
    len += CreateKEPayload( &encrypted_next, GetTransformID( IKEv2_TRT_DIFFIE_HELLMAN_GROUP, IKEv2_PID_IKE_SA,
       session->UsedTransforms ), session, innerp + len );
  }

  // encrypt payloads

  uint32_t enclen =  CreateEncryptedPayload( &next,encrypted_next_val, innerp, len,buf, session, p );
  free(innerp);
  innerp=NULL;

  // generate packet

  *olen = p - buf + enclen;
  *out=buf ;
  return IKEv2_RET_OK;
}



/**
 * @brief Merges two ParseMsgResult structs
 *
\verbatim
   if( result2.x )
     result.x = result2.x  
\endverbatim
 *
 * @param *result   first ParseMsgResult struct to merge and result
 * @param *result2  second ParseMsgResult struct to merge
 */
void MergeParseMsgResults( struct ParseMsgResult *result, struct ParseMsgResult *result2 )
{
  if( result2->SA )
    result->SA = result2->SA;
  if( result2->KE )
    result->KE = result2->KE;
  if( result2->DHGroup )
    result->DHGroup = result2->DHGroup;
  if( result2->Nonce )
    result->Nonce = result2->Nonce;
  if( result2->authData )
    result->authData = result2->authData;
  if( result2->adlen )
    result->adlen = result2->adlen;
  if( result2->authMeth )
    result->authMeth = result2->authMeth;
  if( result2->Authfailed )
    result->Authfailed = result2->Authfailed;
  if( result2->cert )
    result->cert = result2->cert;
  if( result2->enc )
    result->enc = result2->enc;
  if( result2->failed )
    result->failed = result2->failed;
  if( result2->error )
    result->error = result2->error;
  if( result2->unsuppCritPl )
    result->unsuppCritPl = result2->unsuppCritPl;
  if( result2->plType )
    result->plType = result2->plType;
  if( result2->NoPropAcceptable )
    result->NoPropAcceptable = result2->NoPropAcceptable;
  if( result2->KEInvalid )
    result->KEInvalid = result2->KEInvalid;
  if( result2->ChosenDHGroup )
    result->ChosenDHGroup = result2->ChosenDHGroup;

  if( result2->IDiLen )
      result->IDiLen = result2->IDiLen;
  if( result2->IDrLen )
      result->IDrLen = result2->IDrLen;
  if( result2->IDrData ) 
      result->IDrData = result2->IDrData;
  if( result2->IDiData ) 
      result->IDiData = result2->IDiData;
}

void FreeCertList(struct CertList *l){
        if(l==NULL) return;
        else
        {
                if(l->cdata!=NULL) {
		    free(l->cdata);
		    l->cdata=NULL;
		}
                FreeCertList(l->next);
                free(l);
		l=NULL;
        }
        return;
}

void FreeMsgResult(struct ParseMsgResult *res){
        if(res==NULL) return;
        if(res->SA!=NULL) {
	    DeleteProposalList(res->SA);
	    res->SA=NULL;
	}
        if(res->cert!=NULL) {
	    FreeCertList(res->cert);
	    res->cert=NULL;
	}
        if(res->KE!=NULL) {
	    BN_free(res->KE);
	    res->KE=NULL;
	}
        if(res->Nonce!=NULL)  {
	    BN_free(res->Nonce);
	    res->Nonce=NULL;
	}
        if(res->authData!=NULL) {
	    free(res->authData);
	    res->authData=NULL;
	}
	if(res->IDiData) {
	    free(res->IDiData);
	    res->IDiData=NULL;
	}
	if(res->IDrData) {
	    free(res->IDrData);
	    res->IDrData=NULL;
	}
	
        free(res);
	res=NULL;
        return;
}

/**
 * @brief  Parses a message and returns result
 *
 * @param *i2       IKEv2 context data
 * @param *session  current IKEv2 session
 * @param *p        ptr to message
 *
 * @return filled ParseMsgResult structure or NULL on error
 */
struct ParseMsgResult *ParseMsg( const struct ikev2_ctx *i2, struct IKEv2Session *session, uint8_t *p)
{
  struct CertList *cl;
  uint8_t *data, *newmsg, *schksum, *cchksum;
  struct ParseMsgResult *result2;
  uint32_t len, ilen, chklen;
  uint16_t notData;
  struct IKEv2PayloadHeader *pl;
  struct IKEv2Header *hdr2, *hdr = (struct IKEv2Header *)p;

  struct ParseMsgResult *result = (struct ParseMsgResult *)malloc( sizeof( struct ParseMsgResult ) );
  memset( result, 0, sizeof( struct ParseMsgResult ) );
  p += sizeof( struct IKEv2Header );
  uint8_t next = hdr->NextPayload;
  while( next )
  {
    pl = (struct IKEv2PayloadHeader *)p;   // first check length fields...
    if( (uint32_t)( p - (uint8_t *)hdr ) + sizeof( struct IKEv2PayloadHeader ) > ntohl( hdr->Length ) ||
      (uint32_t)( p - (uint8_t *)hdr ) + ntohs( pl->PayloadLength ) > ntohl( hdr->Length ) )
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Error: Length fields are not correctly set (1).\n" );
      result->error = 1;
      return result;
    }
    switch( next ) // payload type
    {
    case IKEv2_NPT_SA:
      if(result->SA) {
          DeleteProposalList(result->SA);
      }
      result->SA = ParseSAPayload( p );
      break;
      
    case IKEv2_NPT_KE:
      if(result->KE) {
	  BN_free(result->KE);
      }
      result->KE = ParseKEPayload( &result->DHGroup, p );
      break;
      
    case IKEv2_NPT_NONCE:
      if(result->Nonce) {
	  BN_free(result->Nonce);
      }
      result->Nonce = ParseNoncePayload( p );
      break;

    case IKEv2_NPT_IDI:
      if(result->IDiData) {
	  free(result->IDiData);
      }
      result->IDiLen = ParseIDPayload( &result->IDiData, p );
      break;

    case IKEv2_NPT_IDR:
      if(result->IDrData) {
	  free(result->IDrData);
      }
      result->IDrLen = ParseIDPayload( &result->IDrData, p );
      break;

    case IKEv2_NPT_AUTH:
      if(result->authData) {
	  free(result->authData);
      }
      result->adlen = ParseAuthPayload( &result->authMeth, &result->authData, p );
      break;

    case IKEv2_NPT_CERT:
      if( !( cl = (struct CertList *)malloc( sizeof( struct CertList ) ) ) )
        break;
      if(!(ParseCertPayload(&cl->cdata, &cl->cdlen,&cl->ctype,p)))
      {
        cl->next = result->cert;
        result->cert = cl;
      }
      else
      {
        free( cl );
        cl=NULL;
      }
      break;

    case IKEv2_NPT_CERTREQ:
      ParseCertReqPayload(i2,p);
      break;
      
    case IKEv2_NPT_NOTIFY:
      if( ParseNotifyPayload( p, &notData ) )
        result->failed = 1;         // last request has failed
      if( notData )                 // invalid ke payload... store dh group
      {
        result->KEInvalid = 1;
        result->ChosenDHGroup = notData;
      }
      break;

    case IKEv2_NPT_ENCRYPTED:

      //bobo: check it
      if( result->failed || result->KEInvalid ) return result;
      if (session->SK_ready == 0){
	  struct Proposal *prop;
         if( !(prop=FindFirstSupportedProposal( result->SA, i2->suppProp )) || !CheckNeededTransforms( result->SA ) ) {
              xlogf( I2L_ERR, "EAP-IKEv2: Error: SA_INIT response contains an inacceptable proposal.\n" );
              session->Status = IKEv2_SST_FAILED;
              //bobo: check this break!!!
              break;
          }
	 DeleteProposal(prop);
	 prop=NULL;
          session->UsedTransforms = result->SA->protocols; // save res to session
          //bobo: check it:
          //DeleteProposalList( result->SA );
          //result->SA=NULL;
          session->Nr = result->Nonce;

        ComputeDHKey( result->KE, session );     // compute all keys
        ComputeSessionKeys( session );
      }
      
      len = ntohs( pl->PayloadLength ); // get position of checksum
      ilen = GetChecksumLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
      session->UsedTransforms ) );
      schksum = p + len - ilen;
      integ( session->Type == IKEv2_STY_INITIATOR ? session->SK_ar : session->SK_ai,
             GetIntegKeyLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
             session->UsedTransforms ) ), (uint8_t *)hdr, ntohl( hdr->Length ) - ilen,
             &cchksum, &chklen, GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
             session->UsedTransforms ) );
      if( ilen!= chklen || memcmp( schksum, cchksum, ilen ) )  // is checksum correct?
      {
        xlogf( I2L_ERR, "EAP-IKEv2: Got a message with incorrect integrity checksum.\n" );
        result->error = 1;
	free(cchksum);
	cchksum=NULL;
        return result;
      }
      free( cchksum );
      cchksum=NULL;
      pl->PayloadLength = htons( ntohs( pl->PayloadLength ) - ilen ); // length -= integrity len
      len = ParseEncryptedPayload( &data, session, p );  // decrypt the payload
      if( !( newmsg = (uint8_t *)malloc( sizeof( struct IKEv2Header ) + len ) ) )
      {
        xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't alloc mem to decrypt payloads.\n" );
        FreeMsgResult(result);
	result=NULL;
        return NULL;
      }
      memcpy( newmsg, hdr, sizeof( struct IKEv2Header ) );
      memcpy( newmsg + sizeof( struct IKEv2Header ), data, len );
      free(data);
      data=NULL;
      hdr2 = (struct IKEv2Header *)newmsg;
      hdr2->Length = htonl( len +
          sizeof( struct IKEv2Header ) );          // set length correct length of msg
      hdr2->NextPayload = pl->NextPayload;         // first encrypted payload content type
      pl->NextPayload = IKEv2_NPT_NONE;            // encrypted payload is always last!
      result2 = ParseMsg( i2, session, newmsg);    // recursively parse decrypted msg
      MergeParseMsgResults( result, result2 );     // merge the result of encrypted pld into current
      free( result2 );
      result2=NULL;
      free( newmsg );
      newmsg=NULL;
      result->enc = 1;
      p += ilen;      // add integrity len which has been subracted from payloadLength field in plhdr
      break;
      
    case IKEv2_NPT_NONE:  // no payload --> unreachable case... ;)
      break;
      
    default:    // unknown payload type
      xlogf( I2L_ERR, "EAP-IKEv2: Got an unknown Payload ID: %d\n", next );
      if( pl->C )     // if C flag is set, send a reject msg, otherwise ignore payload
      {
        result->unsuppCritPl = 1;
        result->error = 1;
        result->plType = next;
      }
      break;
    }

    p += ntohs( pl->PayloadLength );   // adr of next payload
    next = pl->NextPayload;   // type of next payload
  }
  if( (uint32_t)( p - (uint8_t *)hdr ) != ntohl( hdr->Length ) )  // check hdr length field
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Length fields are not correctly set (2).\n" );
    result->error = 1;
    return result;
  }
  return result;
}

/**
 * @brief Parses a SA_INIT message and returns result
 *
 * @param *i2       IKEv2 context data
 * @param *session  current IKEv2 session
 * @param *p        ptr to message
 *
 * @return filled ParseMsgResult structure or NULL on error
 */
struct ParseMsgResult *ParseSAInitMsg( const struct ikev2_ctx *i2, struct IKEv2Session *session, uint8_t *p )
{
  struct IKEv2Header *hdr = (struct IKEv2Header *)p;
  session->rmlen = ntohl( hdr->Length );          // save msg for auth
  if(session->rmsg) {
      free(session->rmsg);
  }
  session->rmsg = Clone( p, session->rmlen );
  struct ParseMsgResult *res = ParseMsg( i2, session, p );
  
  if( hdr->ExchangeType != IKEv2_EXT_IKE_SA_INIT )
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: ExchangeType incorrect.\n" );
    FreeMsgResult(res);
    res=NULL;
    return NULL;
  }
    
  xlogf( I2L_DBG, "EAP-IKEv2: Got an SA_INIT message.\n" );

  if( res->failed || res->KEInvalid )
    return res;
      
  if( !res->SA ) {
    xlogf( I2L_ERR, "EAP-IKEv2: SA payload is missing.\n" );
    res->error = 1;
  }
  if( !res->KE ) {
    xlogf( I2L_ERR, "EAP-IKEv2: KE payload is missing.\n" );
    res->error = 1;
  }
  if( !res->Nonce ) {
    xlogf( I2L_ERR, "EAP-IKEv2: Nonce payload is missing.\n" );
    res->error = 1;
  }
  
  return res;
}

/**
 * @brief Parses a AUTH message and returns result
 *
 * @param *i2       IKEv2 context data
 * @param *session  current IKEv2 session
 * @param *p        ptr to message
 *
 * @return filled ParseMsgResult structure or NULL on error
 */
struct ParseMsgResult *ParseAuthMsg( const struct ikev2_ctx *i2, struct IKEv2Session *session, uint8_t *p )
{
  struct ParseMsgResult *res = ParseMsg( i2, session, p );
  struct IKEv2Header *hdr = (struct IKEv2Header *)p;

  if( hdr->ExchangeType != IKEv2_EXT_IKE_AUTH )
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: ExchangeType incorrect.\n" );
    FreeMsgResult(res);
    res=NULL;
    return NULL;
  }
  if( session->IDr && res->IDrData) {
        if( session->IDrLen != res->IDrLen || memcmp(session->IDr,res->IDrData,session->IDrLen)){
                xlogf( I2L_ERR, "EAP-IKEv2: Error: IDr inconsistent!!!\n" );
                FreeMsgResult(res);
                res=NULL;
                return NULL;
        }
  }

  xlogf( I2L_DBG, "EAP-IKEv2: Got an AUTH message.\n" );
  
  return res;
}

/**
 * @brief Parses an informational message and returns result
 *
 * @param *i2       IKEv2 context data
 * @param *session  current IKEv2 session
 * @param *p        ptr to message
 *
 * @return filled ParseMsgResult structure or NULL on error
 */
struct ParseMsgResult *ParseInformationalMsg( const struct ikev2_ctx *i2, struct IKEv2Session *session, uint8_t *p )
{
  struct ParseMsgResult *res = ParseMsg( i2, session, p );
  struct IKEv2Header *hdr = (struct IKEv2Header *)p;

  if( hdr->ExchangeType != IKEv2_EXT_INFORMATIONAL )
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: ExchangeType incorrect.\n" );
    FreeMsgResult(res);
    res=NULL;
    return NULL;
  }

  xlogf( I2L_DBG, "EAP-IKEv2: Got an informational message.\n" );

  return res;
}

/**
 * @brief  Parses a CREATE_CHILD_SA message for rekeying and returns result
 *
 * @param *i2       IKEv2 context data
 * @param *session  current IKEv2 session
 * @param *p        ptr to message
 *
 * @return filled ParseMsgResult structure or NULL on error
 */
struct ParseMsgResult *ParseChildSAMsg( const struct ikev2_ctx *i2, struct IKEv2Session *session, uint8_t *p )
{
  struct IKEv2Header *hdr = (struct IKEv2Header *)p;

  struct ParseMsgResult *res = ParseMsg( i2, session, p );

  if( hdr->ExchangeType != IKEv2_EXT_CREATE_CHILD_SA )
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: ExchangeType incorrect.\n" );
    FreeMsgResult(res);
    res=NULL;
    return NULL;
  }

  xlogf( I2L_DBG, "EAP-IKEv2: Got an Create Child SA message.\n" );

  if( res->failed || res->KEInvalid )
    return res;

  if( !res->SA ) {
    xlogf( I2L_ERR, "EAP-IKEv2: SA payload is missing.\n" );
    res->error = 1;
  }
  if( !res->Nonce ) {
    xlogf( I2L_ERR, "EAP-IKEv2: Nonce payload is missing.\n" );
    res->error = 1;
  }

  return res;
}
