/*
 *  prf.c  -  pseudo random functions
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include"common.h"

#include<stdlib.h>
#include<string.h>
#include<strings.h>

#include"prf.h"
#include"hmac.h"
#include "enums.h"

/**
 * @brief Get length of digest for a given prf algorithm
 *
 * @param type type of prf function
 *
 * @return length of digest (in bytes)
 */
inline int32_t getDigestLength( uint32_t type )
{
  switch( type )
  {
  case IKEv2_TID_PRF_HMAC_MD5:
    return MD5DIGEST_SIZE;
  case IKEv2_TID_PRF_HMAC_SHA1:
    return SHA1DIGEST_SIZE;
  case IKEv2_TID_PRF_HMAC_TIGER:
    return TIGERDIGEST_SIZE;
  default:
    return 0;
  }
}

/**
 * @brief Check for valid prf key length.
 * Current implamentation always return success.
 *
 * @param type    type of prf algorithm
 * @param keylen  key length 
 *
 * @return 0 on success, 1 otherwise
 */
inline int PrfKeyLenValid( uint32_t type, uint32_t keylen )
{
    type=type; // -Wunused
    keylen=keylen; //-Wunused
    return 0;
}

/**
 * @brief  Pseudo random function (prf)
 *
 * @param *key  key for hmac
 * @param klen  length of key
 * @param *text  data to be hashed                                          
 * @param tlen  length of text
 * @param **res  result                                                     
 * @param *rlen  length of result (output)                                  
 * @param prftype  type of pseudo random function (see IKEv2 specification)
 */
void prf( uint8_t *key, uint32_t klen, const uint8_t *text, uint32_t tlen,
    uint8_t **res, uint32_t *rlen, uint32_t prftype )
{
    *rlen=getDigestLength(prftype);
    if(!*rlen) {
	xlogf(I2L_ERR,"EAP-IKEv2: Unsupported prf algorithm\n");
	*res=NULL;
	return;
    }
    *res=(uint8_t*)malloc(*rlen);
    if(!*res) {
	xlogf(I2L_ERR,"EAP-IKEv2: Can't alloc memory for prf result.\n");
	*rlen=0;
	return;
    }
    switch( prftype ) {
	case IKEv2_TID_PRF_HMAC_MD5:
	    i2_hmac_md5( text, tlen, key, klen, *res );
	    break;
	case IKEv2_TID_PRF_HMAC_SHA1:
	    i2_hmac_sha1( text, tlen, key, klen, *res );
	    break;
	case IKEv2_TID_PRF_HMAC_TIGER:
	    i2_hmac_tiger(text, tlen, key, klen, *res);
	    break;
	default:
	    *res = NULL;
	    *rlen = 0;
	    break;
    }
}

/**
 * @brief Pseudo random function for longer results (prf+)
 *
 * @param *key     key for hmac                                            
 * @param klen     length of key                                           
 * @param *text    data to be hashed                                       
 * @param tlen     length of text                                          
 * @param **res    result                                                  
 * @param rlen     length of result (input)                                
 * @param prftype  type of pseudo random function (see IKEv2 specification)
 */
void prf_plus( uint8_t *key, uint32_t klen, uint8_t *text, uint32_t tlen,
    uint8_t **res, uint32_t rlen, uint32_t prftype )
{
  uint32_t dummy;
  int32_t ds = getDigestLength( prftype );
  int32_t i, nr = rlen / ds + 1;
  *res = (uint8_t *)malloc( nr * ds );
  uint8_t *buf = (uint8_t *)malloc( ds + tlen + 1 );  // buffer for prf data
  uint8_t *dbuf;

  memcpy( buf, text, tlen );
  buf[tlen] = 1;
  prf( key, klen, buf, tlen + 1, &dbuf, &dummy, prftype ); // 1st part
  memcpy( *res, dbuf, ds );
  free( dbuf );
  dbuf=NULL;
  
  for( i = 2; i <= nr; i++ ) // 2nd to last part
  {
    memcpy( buf, *res + ( i - 2 ) * ds, ds );
    memcpy( buf + ds, text, tlen );
    buf[ds + tlen] = (uint8_t)i;
    prf( key, klen, buf, ds + tlen + 1, &dbuf, &dummy, prftype );
    memcpy( *res + ( i - 1 ) * ds, dbuf, ds );
    free( dbuf );
    dbuf=NULL;
  }
  free( buf );
  buf=NULL;
}
