/*
 *  session.c  -  functions for handling sessions
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include <assert.h>

#include"common.h"

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#else
#define pthread_mutex_init(x,y)
#define pthread_mutex_lock(x)
#define pthread_mutex_unlock(x)
#define pthread_mutex_destroy(x)
#endif //HAVE_PTHREAD_H
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/bio.h>
#include <string.h>
#include "transforms.h" 
#include "session.h"
#include "state.h"
#include "prf.h"
#include "encr.h"
#include "integ.h"
#include "auth.h" 
#include "enums.h"

// private prototypes
static char *base64_encode(const uint8_t *msg,int msglen, int *olen);

/**
 * @brief Begin new IKEv2 session
 *
 * @param *i2       IKEv2 context data
 * @param iSPI      initiator SPI
 * @param rSPI      responder SPI
 * @param Type      session type - initiator or responder
 *
 * @return new IKEv2 session struct
 */
struct IKEv2Session *NewSession( struct ikev2_ctx *i2, uint64_t iSPI, uint64_t rSPI, int Type )
{
  struct IKEv2Session *ns;    // alloc mem for new session struct
  if( !( ns = (struct IKEv2Session *)malloc( sizeof( struct IKEv2Session ) ) ) )
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for new session.\n" );
    return NULL;
  }
  //set all fields to zero
  memset(ns,0,sizeof(*ns));

  pthread_mutex_lock(&i2->mutex);
  ns->next = i2->SessionList;    // insert into list
  i2->SessionList = ns;
  pthread_mutex_unlock(&i2->mutex);

  ns->Type = Type & 0x3;     // init session struct
  ns->DHCounter=i2->DHCounterMax;
  ns->iSPI = iSPI;
  ns->rSPI = rSPI;
  ns->State = IKEv2_STATE_INIT;
  ns->Status = IKEv2_SST_NORMAL;
  ns->eapMsgID=Type==IKEv2_STY_INITIATOR?1:0;
  ns->fDHExchange=i2->enableFastDHEx;
  if(Type==IKEv2_STY_RESPONDER) {
      ns->pwd=Clone(i2->pwd,i2->pwdlen);
      ns->pwdlen=i2->pwdlen;
  }

  return ns;
}

/**
 * @brief Search a session with a given initiator SPI
 *
 * @param *i2       IKEv2 context data
 * @param iSPI      initiator SPI
 *
 * @return founded IKEv2 session or NULL if not found       
 */
struct IKEv2Session *FindISession(struct ikev2_ctx *i2, uint64_t iSPI )
{
  pthread_mutex_lock(&((struct ikev2_ctx*)i2)->mutex);
  struct IKEv2Session *h = i2->SessionList;
  while( h )
  {
    if( h->iSPI == iSPI )
      break;
    h = h->next;
  }
  pthread_mutex_unlock(&i2->mutex);
  return h;
}

/**
 * @brief Search a session with a given responder SPI
 *
 * @param *i2       IKEv2 context data
 * @param rSPI      responder SPI
 *
 * @return founded IKEv2 session or NULL if not found       
 */
struct IKEv2Session *FindRSession( struct ikev2_ctx *i2, uint64_t rSPI )
{
  pthread_mutex_lock(&i2->mutex); 
  struct IKEv2Session *h = i2->SessionList;
  while( h )
  {
    if( h->rSPI == rSPI )
      break;
    h = h->next;
  }
  pthread_mutex_unlock(&i2->mutex);
  return h;
}

/**
 * @brief Find session with given fastID which is ready to fast reconnect
 *
 * @param *i2       IKEv2 context data
 * @param fastid    fastID string received from respondent
 *
 * @return founded IKEv2 session or NULL if not found       		
 */
struct IKEv2Session *FindSessionByFastid(struct ikev2_ctx *i2, const char* fastid)
{
    if(!fastid) {
	return NULL;
    }
    pthread_mutex_lock(&i2->mutex); 
    struct IKEv2Session *h=i2->SessionList;
    while(h) {
	if(h->fFastReconnect && h->fastID && !strncmp(h->fastID,fastid,strlen(h->fastID)))
	    break;
	h = h->next;
    }
    pthread_mutex_unlock(&i2->mutex);
    return h;
    
}


/**
 * @brief Search a session with a given SPI.
 *
 * @param *i2       IKEv2 context data
 * @param SPI       SPI
 *
 * @return founded IKEv2 session or NULL if not found       
 */
struct IKEv2Session *FindSession( struct ikev2_ctx *i2, uint64_t SPI )
{
  pthread_mutex_lock(&i2->mutex);   
  struct IKEv2Session *h = i2->SessionList;
  while( h )
  {
    if( ( h->Type == IKEv2_STY_INITIATOR && h->iSPI == SPI ) ||
        ( h->Type == IKEv2_STY_RESPONDER && h->rSPI == SPI ) )
      break;
    h = h->next;
  }
  pthread_mutex_unlock(&i2->mutex);
  return h;
}

/**
 * @brief Release all resources of IKEv2 session
 *
 * @param *s session to free
 */
void FreeSession( struct IKEv2Session *s )
{
  if( !s )  // session exists?
    return;
    
  if( s->dh ) {           // release resources
    DH_free( s->dh );
  }
  if( s->DHKey ) {
    free( s->DHKey );
  }
  if( s->Ni ) {
    BN_free( s->Ni );
  }
    
  if( s->Nr ) {
    BN_free( s->Nr );
  }
  if( s->fmsg ) {
    free( s->fmsg );
  }
  if( s->rmsg ) {
    free( s->rmsg );
  }
  if( s->IDr ) {
    free( s->IDr );
  }
  if( s->pwd ) {
      free( s->pwd );
  }
  if( s->SK_d ) {
    free( s->SK_d );
  }
  if( s->SK_ai ) {
    free( s->SK_ai );
  }
  if( s->SK_ar ) {
    free( s->SK_ar );
  }
  if( s->SK_ei ) {
    free( s->SK_ei );
  }
  if( s->SK_er ) {
    free( s->SK_er );
  }
  if( s->SK_pi ) {
    free( s->SK_pi );
  }
  if( s->SK_pr ) {
    free( s->SK_pr );
  }
  if( s->UsedTransforms) {
      DeleteProtocol(s->UsedTransforms);
  }
  if(s->fragdata) {
      free(s->fragdata);
  }
  if(s->integkey)  {
      free(s->integkey);
  }
  if(s->ointegkey) {
      free(s->ointegkey);
  }
  if(s->eapKeyData) {
      free(s->eapKeyData);
  }
  if(s->fastID) {
      free(s->fastID);
  }
    
  free( s );            // free the session struct
}


/**
 * @brief Free whole session list
 *
 * @param *s head of the session list
 */
void FreeSessionList(struct IKEv2Session *s)
{
    struct IKEv2Session *next;
    while(s) {
	next=s->next;
	FreeSession(s);
	s=next;
    }
}

/**
 * @brief Free any session from list if last activity time was greater than ikev2_ctx.fastExpire
 *
 * @param *i2       IKEv2 context data
 * @param currentTime current timestamp (i.e. time() result)
 *
 * @return number of deleted sessions
 */
int FreeSessionIfExpired(struct ikev2_ctx *i2,uint32_t currentTime)
{
    assert(i2);
    int deleted=0;
    struct IKEv2Session *next;
    struct IKEv2Session *s;
    s=i2->SessionList;
    pthread_mutex_lock(&i2->mutex);   
    if(!s) {
	pthread_mutex_unlock(&i2->mutex);
	return deleted;
    }
    while(s->next) {
	next=s->next->next;
	assert(currentTime>=s->next->timestamp);
	if(currentTime - s->next->timestamp > i2->fastExpire) {
	    FreeSession(s->next);
	    s->next=next;
	    deleted++;
	    continue;
	}
	s=s->next;
    }
    assert(currentTime>=i2->SessionList->timestamp);
    if(currentTime - i2->SessionList->timestamp > i2->fastExpire) {
	next=i2->SessionList->next;
	FreeSession(i2->SessionList);
	i2->SessionList=next;
	deleted++;
    }
    pthread_mutex_unlock(&i2->mutex);
    return deleted;
}

/**
 * @brief Delete a session entry.
 *
 * @param *i2       IKEv2 context data
 * @param *s        session to delete
 */
void DeleteSession( struct ikev2_ctx *i2, struct IKEv2Session *s )
{
  pthread_mutex_lock(&i2->mutex);     
  struct IKEv2Session *h = i2->SessionList;
  if( h == s )
  {
    i2->SessionList = h->next;
    FreeSession( h );
    h=NULL;
    pthread_mutex_unlock(&i2->mutex);
    return;
  }
  while( h->next )
  {
    if( h->next == s )
    {
      h->next = h->next->next;
      FreeSession( s );
      pthread_mutex_unlock(&i2->mutex);
      return;
    }
    h = h->next;
  }
  pthread_mutex_unlock(&i2->mutex);
}

/**
 * @brief Compute all needed session keys
 *
 * @param *s IKEv2 session data
 */
void ComputeSessionKeys( struct IKEv2Session *s )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Computing all needed session keys.\n" );
  assert(s->Ni && s->Nr);
  uint32_t nli = BN_num_bytes( s->Ni );
  uint32_t nlr = BN_num_bytes( s->Nr );
  uint8_t *nonces = (uint8_t *)malloc( nli + nlr + 16 );    // 16 = len of 2x SPI
  BN_bn2bin( s->Ni, nonces );
  BN_bn2bin( s->Nr, nonces + nli );
  uint8_t *SKEYSEED;
  uint32_t kslen;
  prf( nonces, nli + nlr, s->DHKey, s->DHKeyLen, &SKEYSEED, &kslen, GetTransformID(
      IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA, s->UsedTransforms ) );

  uint32_t elen = GetEncrKeyLen( 
	  GetTransformID( IKEv2_TRT_ENCRYPTION_ALGORITHM, IKEv2_PID_IKE_SA, s->UsedTransforms ),
	  GetTransformKeyLen(IKEv2_TRT_ENCRYPTION_ALGORITHM,IKEv2_PID_IKE_SA,s->UsedTransforms)
	  );
  uint32_t ilen = GetIntegKeyLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
      s->UsedTransforms ) );
  uint32_t keylen = 3 * elen + 4 * ilen + RAW_FASTID_LEN;

  uint8_t *key;
  *( (uint64_t *)( nonces + nli + nlr ) ) = s->iSPI;
  *( (uint64_t *)( nonces + nli + nlr + 8 ) ) = s->rSPI;
  prf_plus( SKEYSEED, kslen, nonces, nli + nlr + 16, &key, keylen, GetTransformID(
      IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA, s->UsedTransforms ) );
  free( nonces );
  free(SKEYSEED);
  SKEYSEED=NULL;
  nonces=NULL;

  uint8_t *p = key;
  if(s->SK_d) {
      free(s->SK_d);
  }
  s->SK_d = Clone( p, elen );
  p += elen;
  if(s->SK_ai) {
      free(s->SK_ai);
  }
  s->SK_ai = Clone( p, ilen );
  p += ilen;
  if(s->SK_ar) {
      free(s->SK_ar);
  }
  s->SK_ar = Clone( p, ilen );
  p += ilen;
  if(s->SK_ei) {
      free(s->SK_ei);
  }
  s->SK_ei = Clone( p, elen );
  p += elen;
  if(s->SK_er) {
      free(s->SK_er);
  }
  s->SK_er = Clone( p, elen );
  p += elen;
  if(s->SK_pi) {
      free(s->SK_pi);
  }
  s->SK_pi = Clone( p, ilen );
  p += ilen;
  if(s->SK_pr) {
      free(s->SK_pr);
  }
  s->SK_pr = Clone( p, ilen );
  p += ilen;
  
  int base64_fastid_len;
  if(s->fastID) {
      free(s->fastID);
  }
  s->fastID = base64_encode(p,RAW_FASTID_LEN,&base64_fastid_len);
  s->SK_ready=1;

      
  
  if(
	  (elen && !(s->SK_d && s->SK_ei && s->SK_er )) ||
	  (ilen && !(s->SK_ai && s->SK_ar && s->SK_pi && s->SK_pr)) 
    ) {
      s->SK_ready=0;
      xlogf( I2L_ERR, "EAP-IKEv2: Error: Keys not set! Out of mem (critical).\n" );
  }
  free( key );
  key=NULL;
}

/**
 * @brief Compute all needed session keys in case of rekeying
 *
 * @param *s   IKEv2 session data
 * @param dhex is there a new dh exchange?
 */
void ComputeNewKeys( struct IKEv2Session *s, int dhex)
{
  xlogf( I2L_DBG, "EAP-IKEv2: Recomputing all needed session keys.\n" );
  //DumpIKEv2SessionKeys(s);  
  assert(s->Ni && s->Nr);
  uint32_t nli = BN_num_bytes( s->Ni );
  uint32_t nlr = BN_num_bytes( s->Nr );

  uint32_t elen = GetEncrKeyLen(
	  GetTransformID( IKEv2_TRT_ENCRYPTION_ALGORITHM, IKEv2_PID_IKE_SA, s->UsedTransforms ),
	  GetTransformKeyLen(IKEv2_TRT_ENCRYPTION_ALGORITHM,IKEv2_PID_IKE_SA,s->UsedTransforms)
	  );
  uint32_t ilen = GetIntegKeyLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
      s->UsedTransforms ) );
  uint32_t keylen = 3 * elen + 4 * ilen;

  uint8_t *pnonces = (uint8_t *)malloc( ( dhex ? s->DHKeyLen : 0 ) + nli + nlr + 16 );    // 16 = len of 2x SPI
  uint8_t *nonces = pnonces;
  if( dhex )
  {
    memcpy( nonces, s->DHKey, s->DHKeyLen );
    nonces += s->DHKeyLen;
  }
  BN_bn2bin( s->Ni, nonces );
  BN_bn2bin( s->Nr, nonces + nli );
  uint8_t *SKEYSEED;
  uint32_t kslen;
  prf( s->SK_d, elen, pnonces, ( dhex ? s->DHKeyLen : 0 ) + nli + nlr, &SKEYSEED, &kslen,
      GetTransformID( IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA, s->UsedTransforms ) );

  uint8_t *key;
  *( (uint64_t *)( nonces + nli + nlr ) ) = s->iSPI;
  *( (uint64_t *)( nonces + nli + nlr + 8 ) ) = s->rSPI;
  prf_plus( SKEYSEED, kslen, nonces, nli + nlr + 16, &key, keylen, GetTransformID(
      IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA, s->UsedTransforms ) );
  free( pnonces );
  free(SKEYSEED);
  pnonces=NULL;

  uint8_t *p = key;
  if( s->SK_d ) {
    free( s->SK_d );
    s->SK_d=NULL;
  }
  s->SK_d = Clone( p, elen );
  p += elen;
  if( s->SK_ai ) {
    free( s->SK_ai );
    s->SK_ai=NULL;
  }
  s->SK_ai = Clone( p, ilen );
  p += ilen;
  if( s->SK_ar ) {
    free( s->SK_ar );
    s->SK_ar=NULL;
  }
  s->SK_ar = Clone( p, ilen );
  p += ilen;
  if( s->SK_ei ) {
    free( s->SK_ei );
    s->SK_ei=NULL;
  }
  s->SK_ei = Clone( p, elen );
  p += elen;
  if( s->SK_er ) {
    free( s->SK_er );
    s->SK_er=NULL;
  }
  s->SK_er = Clone( p, elen );
  p += elen;
  if( s->SK_pi ) {
    free( s->SK_pi );
    s->SK_pi=NULL;
  }
  s->SK_pi = Clone( p, ilen );
  p += ilen;
  if( s->SK_pr ) {
    free( s->SK_pr );
    s->SK_pr=NULL;
  }
  s->SK_pr = Clone( p, ilen );
  
  if(s->fastID) {
      free(s->fastID);
      s->fastID=NULL;
  }
  int base64_fastid_len;
  s->fastID = base64_encode(p,RAW_FASTID_LEN,&base64_fastid_len);

  if(
	  (elen && !(s->SK_d && s->SK_ei && s->SK_er )) ||
	  (ilen && !(s->SK_ai && s->SK_ar && s->SK_pi && s->SK_pr)) 
    ) {
      s->SK_ready=0;
      xlogf( I2L_ERR, "EAP-IKEv2: Error: Keys not set! Out of mem (critical).\n" );
  }

  free( key );
  key=NULL;
//  xlogf(I2L_ERR,"XXX:key recomputed\n");
//  DumpIKEv2SessionKeys(s);  
}


/**
 * @brief Create an empty IKEv2 context structure with initialized mutex
 * variable.
 *
 * @return ptr to created IKEv2 context structure
 */
struct ikev2_ctx * Create_ikev2_ctx()
{
    struct ikev2_ctx *res=(struct ikev2_ctx*)malloc(sizeof(struct ikev2_ctx));
    if(res) {
	memset(res,0,sizeof(struct ikev2_ctx));
	pthread_mutex_init(&res->mutex,NULL);
    }
    return res;
}

/**
 * @brief Free IKEv2 context data
 *
 * @param *i2 IKEv2 context data
 */
void Free_ikev2_ctx(struct ikev2_ctx *i2)
{
    if(i2==NULL) return;
    if(i2->id) {
       free(i2->id);
    }
    if(i2->certfile) {
       free(i2->certfile);
    }
    if(i2->crl_file) {
	free(i2->crl_file);
    }
    if(i2->trusted) {
       free(i2->trusted);
    }
    if(i2->pwd) {
       free(i2->pwd);
    }
    if(i2->pkfile) {
       free(i2->pkfile);
    }
    if(i2->pkfile_pwd) {
	free(i2->pkfile_pwd);
    }
    if(i2->cert) {
       free(i2->cert);
    }
    if(i2->suppProp) {
	DeleteProposalList(i2->suppProp);
    }
    if(i2->SessionList) {
	FreeSessionList(i2->SessionList);
    }
    if(i2->sslist) {
	FreeSharedSecList(i2->sslist);
    }
    if(i2->x509_store) {
	X509_STORE_free(i2->x509_store);
    }
    if(i2->CAsHashString) {
	free(i2->CAsHashString);
    }
    if(i2->rpk) {
	EVP_PKEY_free(i2->rpk);
    }
    pthread_mutex_destroy(&i2->mutex);
    free(i2);
}


/**
 * @brief Encode byte string into base64 form.
 *
 * @param *msg 	  input message
 * @param msglen  input message length
 * @param *olen   output message length
 *
 * @return Ptr to result message. User is responsible to free that memory.
 */
static char *base64_encode(const uint8_t *msg,int msglen, int *olen)
{
        if(!msg) {
	    return NULL;
	}
	BIO *mem, *b64;
	b64 = BIO_new(BIO_f_base64());
	BIO_set_flags(b64,BIO_FLAGS_BASE64_NO_NL);
	mem = BIO_new(BIO_s_mem());
	mem = BIO_push(b64, mem);
	BIO_write(mem, msg, msglen);
	BIO_flush(mem);
	BUF_MEM *bm;
	BIO_get_mem_ptr(mem,&bm);
	if(olen) {
	    *olen=bm->length;
	}
	char *result=malloc(bm->length+1);
	result[bm->length]=0;
	memcpy(result,bm->data,bm->length);
	BIO_free_all(mem);
	return result;
    
}


void DumpIKEv2SessionKeys(const struct IKEv2Session *s) 
{

  xlogf(I2L_ERR,"SK_ready=%d\n",s->SK_ready);
  xlogf(I2L_ERR,"fFastReconnect=%d\n",s->fFastReconnect);
  uint32_t elen = GetEncrKeyLen( 
	  GetTransformID( IKEv2_TRT_ENCRYPTION_ALGORITHM, IKEv2_PID_IKE_SA, s->UsedTransforms ),
	  GetTransformKeyLen(IKEv2_TRT_ENCRYPTION_ALGORITHM,IKEv2_PID_IKE_SA,s->UsedTransforms)
	  );
  uint32_t ilen = GetIntegKeyLen( GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA,
      s->UsedTransforms ) );
  xlogf(I2L_ERR,"ilen=%d, elen=%d\n",ilen,elen);
  xlogf(I2L_ERR,"SK_d\n");
  PrintBuffer(s->SK_d,elen);
  xlogf(I2L_ERR,"SK_ai\n");
  PrintBuffer(s->SK_ai,ilen);
  xlogf(I2L_ERR,"SK_ar\n");
  PrintBuffer(s->SK_ar,ilen);
  xlogf(I2L_ERR,"SK_ei\n");
  PrintBuffer(s->SK_ei,elen);
  xlogf(I2L_ERR,"SK_er\n");
  PrintBuffer(s->SK_er,elen);
  xlogf(I2L_ERR,"SK_pi\n");
  PrintBuffer(s->SK_pi,ilen);
  xlogf(I2L_ERR,"SK_pr\n");
  PrintBuffer(s->SK_pr,ilen);
  xlogf(I2L_ERR,"Fast ID = [%s]\n",s->fastID);
}
