/*
 *  state.c  -  state machine for initiator and responder
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include"common.h"

#include<string.h>

#ifdef HAVE_NETINET_IN_H
#include<netinet/in.h>
#endif // HAVE_NETINET_IN_H
#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif // HAVE_WINSOCK_H
 
#include<openssl/rand.h>

#include"state.h"
#include"message.h"
#include"session.h"
#include"payload.h"
#include"dh.h"
#include"prf.h"
#include"encr.h"
#include"integ.h"
#include"auth.h"
#include "enums.h"

// Initiator states:
#define SEND_SA_INIT 			0
#define WAIT_SA_INIT_RESP 		101
#define SEND_AUTH_REQ 			103
#define WAIT_AUTH_RESP 			104
#define SEND_CLIENT_AUTH_FAIL 		106
#define WAIT_SEND_CLIENT_AUTH_FAIL_RES 	107
#define GOT_INVALID_KE 			108
#define GOT_FAILED 			109
#define SEND_ERROR_NOTIFY 		110

// for fast reconect:
#define SEND_CHILDSA_INIT 111
#define WAIT_CHILDSA_INIT_RESP 112
#define WAIT_CHILDSA_INIT 113
#define SEND_CHILDSA_INIT_RESP 114
#define GOT_INVALID_REKEYING_KE         115

#define WAIT_SA_INIT 		        0
#define SEND_INVALID_KE 		203
#define SEND_SA_INIT_RESP		        200
#define WAIT_AUTH 			204
#define SEND_AUTH 			205
//#define SEND_SERVER_AUTH_FAIL 		206
//#define WAIT_FOR_RESP 	                207
//#define GOT_AUTH_FAILED 		208
//#define SEND_ERROR_NOTIFY 		210


void print_I_state(int state);
void print_R_state(int state);


#define PRINT_STATE(STATENAME)  \
    case STATENAME:  \
    xlogf(I2L_DBG, "EAP-IKEv2: ----------- %s state: %s\n",role,#STATENAME); \
    break;


/**
 * @brief Print descripitve name of current initiator state.
 *
 * @param state current state
 */
void print_I_state(int state)
{
    static const char role[]="initiator";
    switch(state) {
	PRINT_STATE(SEND_SA_INIT);
	PRINT_STATE(WAIT_SA_INIT_RESP);
	PRINT_STATE(SEND_AUTH_REQ);
	PRINT_STATE(WAIT_AUTH_RESP);
	PRINT_STATE(SEND_CLIENT_AUTH_FAIL);
	PRINT_STATE(WAIT_SEND_CLIENT_AUTH_FAIL_RES);
	PRINT_STATE(GOT_INVALID_KE);
	PRINT_STATE(GOT_FAILED);
	PRINT_STATE(SEND_ERROR_NOTIFY);
	PRINT_STATE(SEND_CHILDSA_INIT);
	PRINT_STATE(WAIT_CHILDSA_INIT_RESP);
	PRINT_STATE(GOT_INVALID_REKEYING_KE);
	default:
	xlogf(I2L_DBG,"EAP-IKEv2: ----------- %s state: %d\n",role,state); 
    }
}

/**
 * @brief Print descripitve name of current responder state.
 *
 * @param state current state
 */
void print_R_state(int state)
{
    static const char role[]="responder";
    switch(state) {
	PRINT_STATE(WAIT_SA_INIT);
	PRINT_STATE(WAIT_SA_INIT_RESP);
	PRINT_STATE(SEND_AUTH_REQ);
	PRINT_STATE(WAIT_AUTH_RESP);
	PRINT_STATE(SEND_CLIENT_AUTH_FAIL);
	PRINT_STATE(WAIT_SEND_CLIENT_AUTH_FAIL_RES);
	PRINT_STATE(GOT_INVALID_KE);
	PRINT_STATE(GOT_FAILED);
	PRINT_STATE(SEND_ERROR_NOTIFY);
	PRINT_STATE(SEND_INVALID_KE);
	PRINT_STATE(SEND_SA_INIT_RESP);
	PRINT_STATE(WAIT_AUTH);
	PRINT_STATE(SEND_AUTH);
	PRINT_STATE(WAIT_CHILDSA_INIT);
	PRINT_STATE(SEND_CHILDSA_INIT_RESP);
	default:
	xlogf(I2L_DBG,"EAP-IKEv2: ----------- %s state: %d\n",role,state); 
    }

}


/**
 * @brief The responder state machine
 *
 * @param *i2       IKEv2 context data
 * @param *in       received packet
 * @param **out     generated packet to send
 * @param *olen     out length
 * @param *s        current IKEv2 session
 */
void DoRState(const struct ikev2_ctx *i2,  uint8_t *in, uint8_t **out, uint32_t *olen, struct IKEv2Session *s)
{
  struct ParseMsgResult *res=NULL;
  struct IKEv2Header *hdr = (struct IKEv2Header *)in;
  struct Proposal *prop=NULL;
//  int dhg;

  if( in && hdr->MjVersion != 2 )
  {
    xlogf( I2L_INFO, "EAP-IKEv2: Got a message with wrong major version, ignoring...\n" );
    *out = NULL;
    *olen = 0;
    return;
  }
  
  if( s->WaitResponse == 1 )
  {     // check msg ID for responses
    if( !in || s->respMsgID != hdr->MessageID || !( hdr->Flags & IKEv2_HDF_RESPONSE ) )
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Wrong message ID or response flag, ignoring msg (3).\n" );
      *out = NULL;
      *olen = 0;
      return;
    }
  }
  else
  {  // should not be a response
    if( in && ( ntohl( hdr->MessageID ) != s->recvMsgID || ( hdr->Flags & IKEv2_HDF_RESPONSE ) ) )
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Wrong message ID or response flag, ignoring msg (4).\n" );
      *out = NULL;
      *olen = 0;
      return;
    }
    else
      s->recvMsgID++;
  }

  if( in && hdr->ExchangeType == IKEv2_EXT_INFORMATIONAL )  // got an informational, respond...
  {
    if( s->State != IKEv2_STATE_CONNECTED )
    {
      xlogf( I2L_ERR, "EAP-IKEv2: unexpected informational msg, ignoring.\n" );
      *out = NULL;
      *olen = 0;
      return;
    }
    if( !( res = ParseInformationalMsg( i2, s, in ) ) )
    {
      *olen = 0;
      *out = NULL;
      return;
    }
    
    if( IKEv2RespondInformational( out, olen, in ) != IKEv2_RET_OK )   // respond to informational
    {
      *olen = 0;
      *out = NULL;
      FreeMsgResult(res);
      return;
    }
    
    if( res->failed )
      s->Status = IKEv2_SST_FAILED;

    FreeMsgResult(res);
    res=NULL;
    //free( res );
    return;
  }

CHANGE_R_STATE:
  print_R_state(s->State);

  switch( s->State )
  {
  case WAIT_SA_INIT:
    if( !( res = ParseSAInitMsg( i2, s, in ) ) )
    {
      *out = NULL;
      *olen = 0;
      break;
    }
    if( res->failed )
    {
      s->Status = IKEv2_SST_FAILED;
      *out = NULL;
      *olen = 0;
      FreeMsgResult(res);
      res=NULL;
      return;
    }
    if( res->error )
    {
      s->Status = IKEv2_SST_FAILED;
      IKEv2BuildErrorResponseMsg( out, olen, in, res, s, 0 );
      FreeMsgResult(res);
      res=NULL;
      return;
    }

    if( !( prop = FindFirstSupportedProposal( res->SA, i2->suppProp ) ) ||
        !CheckNeededTransforms( prop ) ) // search for an acceptable proposal
    {
      xlogf( I2L_ERR, "EAP-IKEv2: No acceptable proposal....rejecting connection.\n" );
      res->NoPropAcceptable = 1;
      s->Status = IKEv2_SST_FAILED;
      IKEv2BuildErrorResponseMsg( out, olen, in, res, s, 0 );
      break;
    }
    DeleteProposalList( res->SA );
    res->SA=NULL;
    if(s->UsedTransforms) {
	DeleteProtocol(s->UsedTransforms);
    }
    s->UsedTransforms = prop->protocols; // save res to session
    if(s->Ni) {
	BN_free(s->Ni);
    }
    s->Ni = res->Nonce;
    res->Nonce=NULL;
    if( res->DHGroup != GetTransformID( IKEv2_TRT_DIFFIE_HELLMAN_GROUP, IKEv2_PID_IKE_SA,
            s->UsedTransforms ) )   // correct DH group in KE payload??
    {
	free(prop);
	prop=NULL;
        s->State = SEND_INVALID_KE;
        goto CHANGE_R_STATE;
    }
    
    s->State = SEND_SA_INIT_RESP;
    goto CHANGE_R_STATE;
 
  case SEND_SA_INIT_RESP:
    if( IKEv2RBuild_SA_Init( i2, in, out, olen, prop, s, res->KE ) != IKEv2_RET_OK )
            // SK_* keys computed inside IKEv2RBuild_SA_Init
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't generate msg.\n" );
      s->State = WAIT_SA_INIT;
      *out = NULL;
      *olen = 0;
      break;
    }
    free(prop);
    prop=NULL;
        
    s->State = WAIT_AUTH;
    break;

  case SEND_INVALID_KE:
    xlogf( I2L_INFO, "EAP-IKEv2: Invalid KE payload.\n" );
    res->KEInvalid = 1;
    res->ChosenDHGroup = GetTransformID( IKEv2_TRT_DIFFIE_HELLMAN_GROUP, IKEv2_PID_IKE_SA,
        s->UsedTransforms );
    IKEv2BuildErrorResponseMsg( out, olen, in, res, s, 0 );
    s->Status = IKEv2_SST_FAILED;
    s->recvMsgID--;
    
    s->State = WAIT_SA_INIT;
    break;

  case WAIT_AUTH:
    s->expect_integ=1;
    s->include_integ=1;
    if( !( res = ParseAuthMsg( i2, s, in ) ) )
    {
      *out = NULL;
      *olen = 0;
      break;
    }
    if( res->failed )
    {
      s->Status = IKEv2_SST_FAILED;
      *out = NULL;
      *olen = 0;
      FreeMsgResult(res);
      res=NULL;
      return;
    }
    if( res->error )
    {
      s->Status = IKEv2_SST_FAILED;
      IKEv2BuildErrorResponseMsg( out, olen, in, res, s, 1 );
      FreeMsgResult(res);
      res=NULL;
      return;
    }
    if( !res->enc )   // not encrypted, ignoring...
    {
      *out = NULL;
      *olen = 0;
      FreeMsgResult(res);
      res=NULL;
      return;
    }

    if(!Authenticate( i2, res->cert ,res->authMeth, res->authData, res->adlen, s, res->IDiData, res->IDiLen))        // athentication succeeded?
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Authentication failed.\n" );
      res->Authfailed = 1;
      IKEv2BuildErrorResponseMsg( out, olen, in, res, s, 1 );
      s->Status = IKEv2_SST_FAILED;
      FreeMsgResult(res);
      res=NULL;
      return;
    }
    s->State = SEND_AUTH;
    goto CHANGE_R_STATE;

  case SEND_AUTH:
    xlogf( I2L_INFO, "EAP-IKEv2: Authenticated.\n" );
    s->Status = IKEv2_SST_ESTABLISHED;
    s->State = IKEv2_STATE_CONNECTED;

    if( IKEv2RBuild_Auth( i2, out, olen, in, s ) != IKEv2_RET_OK )
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't generate msg.\n" );
      *out = NULL;
      *olen = 0;
      break;
    }
    hdr = (struct IKEv2Header *)*out;
    s->respMsgID = hdr->MessageID;
    s->WaitResponse = 0;  // no more response
    break;

  case IKEv2_STATE_CONNECTED:
    hdr=(struct IKEv2Header*) in;
    if(hdr->ExchangeType==IKEv2_EXT_IKE_SA_INIT) {
	s->State=WAIT_SA_INIT;
	s->include_integ=0;
	s->SK_ready=0;
	if(s->ointegkey) {
	    free(s->ointegkey);
	    s->ointegkey=NULL;
	}
    } else {
	s->State=WAIT_CHILDSA_INIT;
    }
    goto CHANGE_R_STATE;  

  case WAIT_CHILDSA_INIT:
    *out=NULL;
    *olen=0;
    s->Status=IKEv2_SST_REKEYING;
    if(!(res=ParseChildSAMsg(i2,s,in))) {
	break;
    }
    if(res->failed) {
	s->Status=IKEv2_SST_FAILED;
	break;
    }
    if(res->error) {
	s->Status=IKEv2_SST_FAILED;
	IKEv2BuildErrorResponseMsg(out,olen,in,res,s,0);
    }
    if(!(prop=FindFirstSupportedProposal(res->SA, i2->suppProp)) || !CheckNeededTransforms(prop)) {
	xlogf( I2L_INFO,"EAP-IKEv2: No acceptable proposal found - rejecting connection.\n");
	res->NoPropAcceptable=1;
	s->Status=IKEv2_SST_FAILED;
	IKEv2BuildErrorResponseMsg(out,olen,in,res,s,1);
    }
    DeleteProposalList(res->SA);
    res->SA=NULL;
    if(s->UsedTransforms) {
	DeleteProtocol(s->UsedTransforms);
    }
    s->UsedTransforms=prop->protocols;
    if(s->Ni) {
	BN_free(s->Ni);
    }
    s->Ni=res->Nonce;
    res->Nonce=NULL;
    if(res->KE && res->DHGroup!=GetTransformID(IKEv2_TRT_DIFFIE_HELLMAN_GROUP,IKEv2_PID_IKE_SA,s->UsedTransforms)) {
	xlogf(I2L_INFO,"EAP-IKEv2: Invalid KE payload.\n");
	res->KEInvalid=1;
	res->ChosenDHGroup=GetTransformID(IKEv2_TRT_DIFFIE_HELLMAN_GROUP,IKEv2_PID_IKE_SA,s->UsedTransforms);
	IKEv2BuildErrorResponseMsg(out,olen,in,res,s,1);
	s->Status=IKEv2_SST_FAILED;
	s->recvMsgID--;
	s->State=WAIT_CHILDSA_INIT;
	break;
    }
    s->State=SEND_CHILDSA_INIT_RESP;
    goto CHANGE_R_STATE;
    
  case SEND_CHILDSA_INIT_RESP:
    if(IKEv2RCreateChildSA(in,out,olen,s,prop,res->DHGroup)!=IKEv2_RET_OK) {
	xlogf(I2L_ERR,"EAP-IKEv2: Can't generate msg.\n");
	*out=NULL;
	*olen=0;
	break;
    }
    free(prop);
    prop=NULL;
    if(res->KE) {
	if(s->fDHExchange) {
	ComputeDHKey(res->KE,s);
	} else {
	    xlogf(I2L_INFO,"EAP-IKEv2: DH exchange unsupproted\n");
	}
    }
    ComputeNewKeys(s,res->KE && s->fDHExchange);
    s->MsgID=0;
    s->recvMsgID=0;
    s->WaitResponse=0;
    s->State=IKEv2_STATE_CONNECTED;
    s->Status=IKEv2_SST_ESTABLISHED;
    break;

  default:
    xlogf( I2L_ERR, "EAP-IKEv2: Internal Error: Unknown current state for session.\n" );
    *out = NULL;
    *olen = 0;
    break;
  }
  FreeMsgResult(res);
  res=NULL;
}




/**
 * @brief The initiator state machine
 *
 * @param *i2       IKEv2 context data
 * @param *in       received packet
 * @param **out     generated packet to send
 * @param *olen     out length
 * @param *s        current IKEv2 session
 */
void DoIState(const struct ikev2_ctx *i2, uint8_t *in, uint8_t **out, uint32_t *olen, struct IKEv2Session *s)
{
  struct ParseMsgResult *res=NULL;
  struct IKEv2Header *hdr = (struct IKEv2Header *)in;
  struct Proposal *prop;
  //int dhg;

  if( in )
  {
    if( hdr->MjVersion != 2 )
    {
      xlogf( I2L_INFO, "EAP-IKEv2: Got a message with wrong major version, ignoring...\n" );
      *out = NULL;
      *olen = 0;
      return;
    }
    
    if( s->WaitResponse == 1 )
    {
      if( s->respMsgID != hdr->MessageID || !( hdr->Flags & IKEv2_HDF_RESPONSE ) )
      {
        xlogf( I2L_ERR, "EAP-IKEv2: Wrong message ID or response flag, ignoring msg! (1)\n" );
        *out = NULL;
        *olen = 0;
        return;
      }
    }
    else
    {
      if( ntohl( hdr->MessageID ) != s->recvMsgID || ( hdr->Flags & IKEv2_HDF_RESPONSE ) )
      {
        xlogf( I2L_ERR, "EAP-IKEv2: Wrong message ID or response flag, ignoring msg. (2)\n" );
        *out = NULL;
        *olen = 0;
        return;
      }
      else
        s->recvMsgID++;
    }
  }
  else
  {
    if( s->WaitResponse )
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Internal Error: An received msg expected.\n" );
      *out = NULL;
      *olen = 0;
      return;
    }
  }

 
CHANGE_I_STATE:
  print_I_state(s->State);
  switch (s->State) {
      case  SEND_SA_INIT:
	  if( IKEv2IBuild_SA_Init( i2, out, olen, s, 0 ) != IKEv2_RET_OK ) {
	      xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't generate msg.\n" );
	      *out = NULL;
	      *olen = 0;
	      break;
	  }
	  hdr = (struct IKEv2Header *)*out;
	  s->respMsgID = hdr->MessageID;
	  s->State=WAIT_SA_INIT_RESP;
	  s->WaitResponse=1;
	  break;

	  
      case  WAIT_SA_INIT_RESP:
	  if( !( res = ParseSAInitMsg( i2, s, in ) ) ) {
	      *out = NULL;
	      *olen = 0;
	      break;
	  }
	  s->WaitResponse=0;
	  if( res->failed || res->error) {
	      s->State = GOT_FAILED;
	      goto CHANGE_I_STATE;
	  }
	  if( res->KEInvalid ){      // KE payload contained wrong DH group, resend first msg with correct KE
	      s->State=GOT_INVALID_KE;
	      goto CHANGE_I_STATE;
	  }
	  
	  if(s->SK_ready==0){          
	      if( !(prop=FindFirstSupportedProposal(res->SA,i2->suppProp)) || !CheckNeededTransforms( res->SA ) ) {
		  xlogf( I2L_ERR, "EAP-IKEv2: Error: SA_INIT response contains an inacceptable proposal.\n" );
		  *out = NULL;
		  *olen = 0;
		  s->Status = IKEv2_SST_FAILED;
		  break;
	      }
	      DeleteProposal(prop);
	      prop=NULL;
	      s->UsedTransforms = res->SA->protocols; // save res to session
	      s->Nr = res->Nonce;

	      ComputeDHKey( res->KE, s );     // compute all keys
	      ComputeSessionKeys( s );
	  }
          
          res->SA->protocols = NULL;
          DeleteProposalList( res->SA );
          res->SA=NULL;
          res->Nonce=NULL;

	  s->State=SEND_AUTH_REQ;
	  goto CHANGE_I_STATE;

	  
      case GOT_FAILED:
	  s->Status = IKEv2_SST_FAILED;
	  *out = NULL;
	  *olen = 0;
	  break;

	  
      case GOT_INVALID_KE: 
#ifdef  IKE_FORCE_SAME_DHG
#define IKE_RES_CHOSEN_DH_GROUP 0
#warning Initiator forced to use same DHGroup
#else
#define IKE_RES_CHOSEN_DH_GROUP res->ChosenDHGroup
#endif//IKE_FORCE_SAME_DHG
	  xlogf(I2L_DBG,"EAP-IKEv2: DH counter = %d\n",s->DHCounter);
	  if(s->DHCounter) {
	      s->MsgID--;
	      if( IKEv2IBuild_SA_Init( i2, out, olen, s, IKE_RES_CHOSEN_DH_GROUP) != IKEv2_RET_OK ) {
		  xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't generate msg.\n" );
		  *out = NULL;
		  *olen = 0;
		  break;
	      }
	      hdr = (struct IKEv2Header *)*out;
	      s->respMsgID = hdr->MessageID;
	      s->State=WAIT_SA_INIT_RESP;
	      s->WaitResponse=1;
	      s->DHCounter--;
	  } else {
	      xlogf(I2L_ERR,"EAP-IKEv2: Too many 'Invalid KE'");
	      s->State=GOT_FAILED;
	      s->WaitResponse=0;
	      goto CHANGE_I_STATE;
	  }
	  break;


      case  SEND_AUTH_REQ:
          if (res->IDrData!=NULL){
              if (i2->authtype==IKEv2_AUTH_SK){
                      //if using sharedkey auth and received SK{IDr} payload
                  struct sharedSecList *list = i2->sslist;
                  uint8_t *id=IKEv2_GET_ID(res->IDrData);
                  int idlen=res->IDrLen-4; //bobo: clean this -4
		  if(s->pwd) {
		      free(s->pwd);
		      s->pwd=NULL;
		  }
                  while( list ){
                    if(list->idlen==idlen && !memcmp(id,list->id,idlen)){
	                    xlogf( I2L_DBG, "EAP-IKEv2: IDr matched\n" );
                            s->pwd=Clone(list->pwd,list->pwdlen);
                            s->pwdlen=list->pwdlen;
                            break;
                    }
                    list = list->next;
                  }
                  if(!list) {
		      xlogf( I2L_DBG, "EAP-IKEv2: IDr NOT matched!!!!!\n" );
		  }
  		  if(!s->pwd) {
  		      xlogf(I2L_DBG,"EAP-IKEv2: Password not available - generating random\n");
  		      s->pwd=(uint8_t*) malloc(20);
  		      s->pwdlen=20;
  		      RAND_bytes(s->pwd,s->pwdlen);
		  }
              }
              if(s->IDr){
                        free(s->IDr);
                        s->IDr=NULL;
                        s->IDrLen=0;
              }
              if((s->IDr=Clone(res->IDrData,res->IDrLen))){
                      s->IDrLen=res->IDrLen;
              }
          }
	  if( IKEv2IBuild_Auth( i2, out, olen, s ) != IKEv2_RET_OK ) {
	      xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't generate msg.\n" );
	      *out = NULL;
	      *olen = 0;
	      break;
	  }
          s->include_integ=1;
          s->expect_integ=1;
	  hdr = (struct IKEv2Header *)*out;
	  s->respMsgID = hdr->MessageID;
	  s->State=WAIT_AUTH_RESP;
	  s->WaitResponse=1;
	  break;

	  
      case  WAIT_AUTH_RESP:
	  if( !( res = ParseAuthMsg( i2, s, in ) ) ) {
	      *out = NULL;
	      *olen = 0;
	      break;
	  }
	  s->WaitResponse=0;
	  if( res->error ) {
	      s->State=SEND_ERROR_NOTIFY;
	      goto CHANGE_I_STATE;

	  }
	  if(res->failed) {
	      s->State=GOT_FAILED;
	      goto CHANGE_I_STATE;
	  }
	  if(!res->enc) {
	      break;
	  }

	  if(!Authenticate(i2,res->cert,res->authMeth,res->authData,res->adlen,s,res->IDrData,res->IDrLen)){
	      xlogf( I2L_ERR, "EAP-IKEv2: Authentication failed.\n" );   // initiate an informational exchange??
	      res->Authfailed = 1;
	      s->State=SEND_ERROR_NOTIFY;
	      goto CHANGE_I_STATE;
	  }
	  xlogf( I2L_INFO, "EAP-IKEv2: Authenticated.\n" );
	  s->WaitResponse = 0;  // got SA_INIT response
	  s->Status = IKEv2_SST_ESTABLISHED;
	  s->State = IKEv2_STATE_CONNECTED;
	  *out=NULL;
	  *olen=0;
	  break;

	  
       case SEND_ERROR_NOTIFY:
	      IKEv2BuildErrorMsg( out, olen, s, res );
	      s->State = IKEv2_STATE_FAILED;
	      if(out) {
		  s->respMsgID = ( (struct IKEv2Header *)*out )->MessageID;
	      }
	      s->WaitResponse = 1;
	      break;

	      
       case  IKEv2_STATE_FAILED:
	      if( !( res = ParseInformationalMsg( i2, s, in ) ) ) {
		  *out = NULL;
		  *olen = 0;
		  break;
	      }
	      s->WaitResponse = 0;
	      s->Status = IKEv2_SST_FAILED;
	      break;
       case IKEv2_STATE_CONNECTED:
	      if(s->fFastReconnect) {
		  s->State=SEND_CHILDSA_INIT;
		  goto CHANGE_I_STATE;
	      }
	      
       case SEND_CHILDSA_INIT:
	      //kwz: dirty hack for keys mismatch
	      if(s->ointegkey) {
		  free(s->ointegkey);
		  s->ointegkey=NULL;
	      }
	      //initiate new EAP sequence numbers
	      s->eapMsgID=0;
	      s->recvMsgID=0;
	      s->respMsgID=0;
	      s->MsgID=0;
	      s->Status=IKEv2_SST_REKEYING;
	      s->DHCounter=i2->DHCounterMax;
	      if(IKEv2ICreateChildSA(i2,out,olen,s,0)!=IKEv2_RET_OK) {
		  xlogf(I2L_ERR,"Failed to create CHILDSA message\n");
		  *out=NULL;
		  *olen=0;
		  break;
	      }

	      s->WaitResponse=1;
	      s->State=WAIT_CHILDSA_INIT_RESP;
	     break; 
       case WAIT_CHILDSA_INIT_RESP:
	  if( !( res = ParseChildSAMsg( i2, s, in ) ) ) {
	      *out = NULL;
	      *olen = 0;
	      break;
	  }
	  s->WaitResponse=0;
	  if( res->error ) {
	      s->State=SEND_ERROR_NOTIFY;
	      goto CHANGE_I_STATE;

	  }
	  if(res->failed) {
	      s->State=GOT_FAILED;
	      goto CHANGE_I_STATE;
	  }
	  if(res->KEInvalid) {
	      s->State=GOT_INVALID_REKEYING_KE;
	      goto CHANGE_I_STATE;
	  }
	  if(!res->enc) {
	      break;
	  }
	  if(s->Nr) {
	      BN_free(s->Nr);
	  }
	  s->Nr=res->Nonce;
	  res->Nonce=NULL;
	  if(res->KE) {
	      if(s->fDHExchange) {
		  ComputeDHKey(res->KE,s);
	      } else {
		  xlogf(I2L_INFO,"EAP-IKEv2: DH exchange unsupproted\n");
	      }
	  }
          ComputeNewKeys(s,res->KE && s->fDHExchange);
	  xlogf( I2L_INFO, "EAP-IKEv2: Authenticated.\n" );
	  s->DHCounter=i2->DHCounterMax;
	  s->Status = IKEv2_SST_ESTABLISHED;
	  s->State = IKEv2_STATE_CONNECTED;
	  break;
	      
      	     	      
      case GOT_INVALID_REKEYING_KE: 
#ifdef  IKE_FORCE_SAME_REKEYING_DHG
#define IKE_RES_CHOSEN_REKEYING_DH_GROUP 5
#warning Initiator forced to use same DHGroup in fast rekeying procedure
#else
#define IKE_RES_CHOSEN_REKEYING_DH_GROUP res->ChosenDHGroup
#endif//IKE_FORCE_SAME_DHG
	  xlogf(I2L_DBG,"EAP-IKEv2: DH counter = %d\n",s->DHCounter);
	  if(s->DHCounter) {
	      s->MsgID--;
	      if( IKEv2ICreateChildSA(i2,out,olen,s,IKE_RES_CHOSEN_REKEYING_DH_GROUP) != IKEv2_RET_OK ) {
		  xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't generate msg.\n" );
		  *out = NULL;
		  *olen = 0;
		  break;
	      }
	      hdr = (struct IKEv2Header *)*out;
	      s->respMsgID = hdr->MessageID;
	      s->State=WAIT_CHILDSA_INIT_RESP;
	      s->WaitResponse=1;
	      s->DHCounter--;
	  } else {
	      xlogf(I2L_ERR,"EAP-IKEv2: Too many 'Invalid KE'");
	      s->State=GOT_FAILED;
	      s->WaitResponse=0;
	      goto CHANGE_I_STATE;
	  }
	  break;

       default:
	      xlogf( I2L_ERR, "EAP-IKEv2: Internal Error: Unknown current state for session.\n" );
	      *out = NULL;
	      *olen = 0;
	      break;
  }
  if(res) {
      FreeMsgResult(res);
      res=NULL;
  }
	      
}

