/*
 *  transforms.c  -  functions supporting the negotiation of crypto-algorithms
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include"common.h"

#include <stdlib.h>
#include <string.h>
 
#include "enums.h"
#include "transforms.h"
#include "integ.h"
#include "encr.h"
#include "prf.h"

// private prototypes:

struct Proposal *CloneProposal( struct Proposal *prop, struct Proposal *ref );


/**
 * @brief Add a new proposal to proposal list
 *
 * @param **list proposal list
 *
 * @return ptr to added proposal or NULL when fail
 */
struct Proposal *AddProposal( struct Proposal **list )
{
  struct Proposal *new;
  if( !( new = (struct Proposal *)malloc( sizeof( struct Proposal ) ) ) ) // alloc new
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't alloc mem for proposal.\n" );
    return NULL;
  }
  new->protocols = NULL;
  new->next = NULL;   // append always at the end
  if( !( *list ) )    // empty list?
    *list = new;
  else
  {
    struct Proposal *h = *list;
    while( h->next )    // search for end of list
      h = h->next;
    h->next = new;      // insert
  }
  return new;
}

/**
 * @brief Removes a proposal from propsals list
 *
 * @param *prop proposal to remove
 * @param **list proposal list
 */
void RemoveProposal( struct Proposal *prop, struct Proposal **list )
{
  if( !( *list ) )    // empty list?
    return;
    
  if( *list == prop )   // is first?
    *list = ( *list )->next;
  else
  {
    struct Proposal *h = *list;
    while( h->next )      // search for item
    {
      if( h->next == prop )   // found?
      {
        h->next = h->next->next;
        return;
      }
      h = h->next;
    }
  }
}

/**
 * @brief Free proposal with protocols and transforms
 *
 * @param *prop proposal to free
 */
void DeleteProposal( struct Proposal *prop )
{
  struct Protocol *h = prop->protocols;
  while( h )
  {
    h = h->next;
    DeleteProtocol( prop->protocols );
    prop->protocols = h;
  }
  free( prop );
  prop=NULL;
}

/**
 * @brief Free a list of proposals
 *
 * @param *prop ptrr to proposals list
 */
void DeleteProposalList( struct Proposal *prop )
{
  struct Proposal *h = prop;
  while( prop )
  {
    h = prop->next;
    DeleteProposal( prop );
    prop = h;
  }
}

/**
 * @brief Add a protocol to a given proposal
 *
 * @param *prop     target proposal
 * @param protocol  protocol ID
 * @param SPISize   SPI size
 * @param SPI       SPI
 *
 * @return ptr to new added protocol or NULL when fail
 */
struct Protocol *AddProtocol( struct Proposal *prop, uint8_t protocol, uint8_t SPISize, uint32_t SPI )
{
  struct Protocol *new;
  if( !( new = (struct Protocol *)malloc( sizeof( struct Protocol ) ) ) ) // alloc new
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't alloc mem for protocol.\n" );
    return NULL;
  }
  new->next = NULL;   // append always at the end
  new->transforms = NULL;
  new->protocol = protocol;
  new->SPISize = SPISize;   // for initial exchanges always 0!!
  new->SPI = SPI;
  if( !prop->protocols )    // empty list?
    prop->protocols = new;
  else
  {
    struct Protocol *h = prop->protocols;
    while( h->next )    // search for end of list
      h = h->next;
    h->next = new;      // insert
  }
  return new;
}

/* removes a protocol from a list */
/**
 * @brief Removes a protocol from proposals list
 *
 * @param *protocol protocol to remove
 * @param *prop proposals list
 */
void RemoveProtocol( struct Protocol *protocol, struct Proposal *prop )
{
  if( !prop->protocols )    // empty list?
    return;

  if( prop->protocols == protocol )   // is first?
    prop->protocols = prop->protocols->next;
  else
  {
    struct Protocol *h = prop->protocols;
    while( h->next )      // search for item
    {
      if( h->next == protocol )   // found?
      {
        h->next = h->next->next;
        return;
      }
      h = h->next;
    }
  }
}

/**
 * @brief Free a protocol with transforms
 *
 * @param *protocol protocol to free
 */
void DeleteProtocol( struct Protocol *protocol )
{
  struct Transform *h = protocol->transforms;
  while( h )
  {
    h = h->next;
    free( protocol->transforms );
    protocol->transforms = h;
  }
  free( protocol );
  protocol=NULL;
}

/* add a transform to a protocol */
/**
 * @brief Add a transform to a protocol
 *
 * @param *protocol protocol struct
 * @param type      transform type
 * @param id        transform id
 * @param keyLen    key length if required, 0 otherwise
 *
 * @return 1 on success, 0 otherwise
 */
int AddTransform( struct Protocol *protocol, uint8_t type, uint16_t id, uint16_t keyLen )
{
  if( keyLen )
  {
    switch( type )    // does key length make a sense?
    {
    case IKEv2_TRT_ENCRYPTION_ALGORITHM:
      if( !EncrKeyLenValid( id, keyLen ) )
        return 0;
      break;

    case IKEv2_TRT_PSEUDO_RANDOM_FUNCTION:
      if( !PrfKeyLenValid( id, keyLen ) )
        return 0;
      break;

    case IKEv2_TRT_INTEGRITY_ALGORITHM:
      if( !IntegKeyLenValid( id, keyLen ) )
        return 0;
      break;

    case IKEv2_TRT_DIFFIE_HELLMAN_GROUP:      // no key len for dh and esn
    case IKEv2_TRT_EXTENDED_SEQUENCE_NUMBERS:
      return 0;
    }
  }

  struct Transform *new;
  if( !( new = (struct Transform *)malloc( sizeof( struct Transform ) ) ) ) // alloc new
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't alloc mem for trasnform.\n" );
    return 0;
  }
  new->next = NULL;   // append always at the end
  new->type = type;
  new->id = id;
  new->keyLen = keyLen;
  if( !protocol->transforms )    // empty list?
    protocol->transforms = new;
  else
  {
    struct Transform *h = protocol->transforms;
    while( h->next )    // search for end of list
      h = h->next;
    h->next = new;      // insert
  }
  return 1;
}

/**
 * @brief Remova transform
 *
 * @param *trans     transform to remove
 * @param *protocol  protocol with given transform
 */
void RemoveTransform( struct Transform *trans, struct Protocol *protocol )
{
  if( !protocol->transforms )    // empty list?
    return;

  if( protocol->transforms == trans )   // is first?
    protocol->transforms = protocol->transforms->next;
  else
  {
    struct Transform *h = protocol->transforms;
    while( h->next )      // search for item
    {
      if( h->next == trans )   // found?
      {
        h->next = h->next->next;
        return;
      }
      h = h->next;
    }
  }
}

/**
 * @brief Get a transform id for a given transform typa and protocol
 *
 * @param type      transform type
 * @param protocol  protocol ID
 * @param *list     protocol list
 *
 * @return transform ID, or 0 when not found
 */
uint16_t GetTransformID( uint8_t type, uint8_t protocol, struct Protocol *list )
{
  while( list )
  {
    if( list->protocol == protocol )
    {
      struct Transform *trans = list->transforms;
      while( trans )
      {
        if( trans->type == type )
          return trans->id;
        trans = trans->next;
      }
    }
    list = list->next;
  }
  return 0;
}

/**
 * @brief Get a transform key length for a given transform type and protocol
 *
 * @param type      transform type
 * @param protocol  protocol ID
 * @param *list     protocol list
 *
 * @return key length
 */
uint16_t GetTransformKeyLen( uint8_t type, uint8_t protocol, struct Protocol *list )
{
  while( list )
  {
    if( list->protocol == protocol )
    {
      struct Transform *trans = list->transforms;
      while( trans )
      {
        if( trans->type == type )
          return trans->keyLen;
        trans = trans->next;
      }
    }
    list = list->next;
  }
  return 0;
}

/**
 * @brief Check for required transform types
 *
 * @param *prop proposals
 *
 * @return 1 on success, 0 otherwise
 */
int CheckNeededTransforms( struct Proposal *prop )
{
  if( !GetTransformID( IKEv2_TRT_ENCRYPTION_ALGORITHM, IKEv2_PID_IKE_SA, prop->protocols ) )
    return 0;
  if( !GetTransformID( IKEv2_TRT_PSEUDO_RANDOM_FUNCTION, IKEv2_PID_IKE_SA, prop->protocols ) )
    return 0;
  if( !GetTransformID( IKEv2_TRT_INTEGRITY_ALGORITHM, IKEv2_PID_IKE_SA, prop->protocols ) )
    return 0;
  if( !GetTransformID( IKEv2_TRT_DIFFIE_HELLMAN_GROUP, IKEv2_PID_IKE_SA, prop->protocols ) )
    return 0;
  return 1;
}

/**
 * @brief Clone proposal prop with first transform of each type, which is also
 * in ref
 *
 * @param *prop source
 * @param *ref  reference
 *
 * @return result proposal 
 */
struct Proposal *CloneProposal( struct Proposal *prop, struct Proposal *ref )
{
  int next;
  uint8_t i;
  struct Protocol *prot;
  struct Proposal *erg = (struct Proposal *)malloc( sizeof( struct Proposal ) );
  struct Protocol *p1 = prop->protocols;
  struct Protocol *p2 = ref->protocols;
  
  memset( erg, 0, sizeof( struct Proposal ) );
  while( p1 && p2 )
  {
    if( p1->protocol != p2->protocol )
    {
      DeleteProposal( erg );
      return NULL;
    }

    prot = AddProtocol( erg, p1->protocol, p1->SPISize, p1->SPI );
    for( i = 1; i <= 5; i++ )    // transform types
    {
      next = 0;
      struct Transform *tr1 = p1->transforms;
      while( !next && tr1 )
      {
        if( tr1->type == i )
        {
          struct Transform *tr2 = p2->transforms;
          while( !next && tr2 )
          {
            if( tr2->type == i && tr1->id == tr2->id && tr1->keyLen == tr2->keyLen )
            {
              AddTransform( prot, i, tr1->id, tr1->keyLen );
              next = 1;
            }
            tr2 = tr2->next;
          }
        }
        tr1 = tr1->next;
      }
    }
    p1 = p1->next;
    p2 = p2->next;
  }
  if( p1 != p2 )
  {
    DeleteProposal( erg );
    return NULL;
  }
  return erg;
}

/**
 * @brief Find a supported proposal, if there is any
 *
 * @param *prop proposal to find
 * @param *list proposal list
 *
 * @return copy of founded proposal, or NULL if not found
 */
struct Proposal *FindFirstSupportedProposal( struct Proposal *prop, struct Proposal *list )
{
  struct Proposal *myprop;
  while( prop )
  {
    if( ( myprop = FindProposal( prop, list ) ) )
    {
      struct Proposal *cp = CloneProposal( prop, myprop );
      if( cp )
        cp->next = NULL;
      return cp;
    }
    prop = prop->next;
  }
  return NULL;
}

/**
 * @brief Compare two proposals
 *
 * @param *p1 argument 1
 * @param *p2 argument 2
 *
 * @return 1 if equal, 0 otherwise
 */
int CmpProposal( struct Proposal *p1, struct Proposal *p2 )
{
  int next, ex;
  uint8_t i;
  struct Protocol *pr1 = p1->protocols;
  struct Protocol *pr2 = p2->protocols;
  while( pr1 && pr2 )
  {
    if( pr1->protocol != pr2->protocol )
      return 0;

    for( i = 1; i <= 5; i++ )    // transform types
    {
      ex = 0;
      next = 0;
      struct Transform *tr1 = pr1->transforms;
      while( !next && tr1 )
      {
        if( tr1->type == i )
        {
          ex = 1;
          struct Transform *tr2 = pr2->transforms;
          while( !next && tr2 )
          {
            if( tr2->type == i && tr1->id == tr2->id && tr1->keyLen == tr2->keyLen )
              next = 1;
            tr2 = tr2->next;
          }
        }
        tr1 = tr1->next;
      }
      if( ( GetTransformID( i, pr1->protocol, pr1 ) == 0 ) ^
          ( GetTransformID( i, pr2->protocol, pr2 ) == 0 ) || ( ex && !next ) )
        return 0;      // a transform type is in p1 (p2) but not in p2 (p1)
    }
    pr1 = pr1->next;
    pr2 = pr2->next;
  }
  if( pr1 != pr2 )
    return 0;
  return 1;
}

/**
 * @brief Find proposal in proposal list
 *
 * @param *prop proposal to find
 * @param *list proposal list
 *
 * @return ptr to founded proposal
 */
struct Proposal *FindProposal( struct Proposal *prop, struct Proposal *list )
{
  while( list )
  {
    if( CmpProposal( prop, list ) )
      return list;
    list = list->next;
  }
  return NULL;
}




