/*
 *  utils.c  -  some useful functions (data conversion, ...)
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include"common.h"

#include<stdlib.h>
#include<string.h>
#include<assert.h>
 

int validate_hexstring(char *s);
uint8_t gethalfbyte(uint8_t c);


/**
 * @brief check hex string for validity
 *
 * @param *s hex string in format "0x..."
 *
 * @return 1 on success, 0 otherwise
 */
int validate_hexstring(char *s){
        unsigned i=0;
        uint8_t c;
        while((c=s[i++])!='\0')
                if((c<'0'||c>'9')&&(c<'A'||c>'F')) return 0;
        
        return 1;
}


/**
 * @brief Get half byte from hex string representation
 *
 * @param c character value
 *
 * @return integer value
 */
uint8_t gethalfbyte(uint8_t c){
        if (c>='0'&&c<='9') return c-'0';
        else return 10+c-'A';
}

/**
 * @brief Create printable representation of byte string.
 * If data in set are ASCII characters, then representation is given string,
 * else the format is "0x...".
 *
 * @param data byte string
 * @param len length of data
 * 
 * @return pointer to static buffer which contains printable representation of
 * data
 */
const char* bin2str(const uint8_t *data,uint32_t len)
{
    assert(len<256/2-4);
    static char result[256];
    unsigned int i;
    int isbinary=0;
    for(i=0;i<len;i++) {
	if(data[i]<32 || data[i]>127) {
	    isbinary=1;
	}
    }
    char *p;
    if(isbinary) {
	p=result;
	sprintf(result,"0x");
	p+=2;
	for(i=0;i<len;i++) {
	    sprintf(p,"%02X",data[i]);
	    p+=2;
	}
    } else {
	p=result;
	for(i=0;i<len;i++) {
	    sprintf(p,"%c",data[i]);
	    p++;
	}
    }
    return result;
}


/**
 * @brief convert string representation of hex value into array of bytes
 *
 * @param *s     string representation
 * @param **dest result array of bytes
 *
 * @return length of dest
 */
unsigned hex2bytes(char *s, uint8_t **dest)
{
        unsigned si=0,di=0,len,pair;

        if(s[0]!='0'||s[1]!='x') return 0;
        s=s+2;
        len=strlen(s);
        pair=len&1;
        len=len>>1;
        
        if((len==0&&pair==0)||!validate_hexstring(s)) return 0;

        *dest=(uint8_t*)malloc(len+pair);
        if(pair) (*dest)[di++]=gethalfbyte(s[si++])&0xF;

        if(!*dest) {
                xlogf(I2L_ERR,"EAP-IKEv2: Malloc error\n");
                return 0;
        }

        for(;di<len+pair;di++,si+=2)
                (*dest)[di]=((unsigned)gethalfbyte(s[si])<<4)|((gethalfbyte(s[si+1]))&0xFF);
        
        return len+pair;
}

/**
 * @brief If data passed to function are in format "0x..." then convert such
 * valu to correspondig byte array, else leave text string.
 *
 * @param **data  data to convert/result
 * @param *len    len of result data
 */
void hexalize(uint8_t **data,unsigned *len) 
{
    uint8_t *tdata=NULL;
    if(!data||!len) {
	return;
    }
    char *udata=(char*)*data;
    if(*data) {
	*len=hex2bytes(udata,&tdata);
	if(*len) {
	    free(*data);
	    *data=tdata;
	} else {
	    *len=strlen((char*)*data);
	}
    }
}

/**
 * @brief Convert a hex representation to BIGNUM
 *
 * @param *hex hex representation
 * @param len  length of hex
 *
 * @return BIGNUM value of hex 
 */
BIGNUM *HexToBignum( const uint8_t *hex, uint32_t len )
{
  int32_t i, j = ( len + 1 ) / 2 - 1;
  int h = 1;
  uint8_t *buf = (uint8_t *)malloc( j + 1 );
  if(buf==NULL) return NULL;
  
  for( i = len - 1; i >= 0; i-- )
  {
    if( !ishex( hex[i] ) )
      continue;
    if( h )
    {
      buf[j] = hextobin( hex[i] );
      h = 0;
    }
    else
    {
      buf[j] |= hextobin( hex[i] ) << 4;
      h = 1;
      j--;
    }
  }

  BIGNUM *result=BN_bin2bn( buf + j + 1, ( len + 1 ) / 2 - j - 1, NULL );
  free(buf);
  buf=NULL;
  return result;
}


/**
 * @brief Clone a buffer
 *
 * @param *data data to clone
 * @param len data length
 *
 * @return pointer to new allocated memory with copy of data.
 */
uint8_t *Clone( const uint8_t *data, uint32_t len )
{
  uint8_t *buf;

  if( !data || !len )
    return NULL;
    
  if((buf=(uint8_t *)malloc(len))) {
      memcpy( buf, data, len );
  }
  return buf;
}

/**
 * @brief Print a binary buffer in hex string format
 *
 * @param *buf buffer
 * @param len  buffer length
 */
void PrintBuffer( unsigned char *buf, unsigned long len )
{
  unsigned long i;
  for( i = 0; i < len; i++ )
  {
    if( i % 16 == 0 )
      printf( "\n" );
    if( i % 2 == 0)
      printf( " " );
    printf( "%02X", buf[i] );
  }

  printf( "\n" );
}

void PrintBignum(const BIGNUM *a){
        unsigned char *buf;
        if(a==NULL) return;
        if(!(buf=(unsigned char*)malloc(BN_num_bytes(a)))) return; 
        
        BN_bn2bin(a,buf);
        PrintBuffer(buf,BN_num_bytes(a));
        free(buf);
}

/**
 * @brief Convert IP address from string to binary format
 * e.g. 1.2.34 --> 0x04030201
 *
 * @param *str_ip string with IP address
 *
 * @return 32-bit value of IP address
 */
uint32_t StringToIP( char *str_ip )
{
  uint32_t ip = 0;
  char num[4];
  int i, j = 0, k;
  for( i = 0; i < 4; i++ )
  {
    for( k = 0; k < 3 && str_ip[j] && str_ip[j] >= '0' && str_ip[j] <= '9'; j++, k++ )
      num[k] = str_ip[j];
    num[k] = 0;
    j++;
    ip |= (char)atoi( num ) << ( i * 8 );
  }
  return ip;
}

/**
 * @brief Convert an IP addr from binary format to string
 * e.g. 0x04030201 --> 1.2.3.4
 * 
 * @param ip 32-bit value of IP address
 *
 * @return string representation of IP address
 */
char *IPToString( uint32_t ip )
{
  char *str_ip;
  if( !( str_ip = (char *)malloc( 16 ) ) )
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate mem for buffer in func 'IPToString'.\n" );
    return NULL;
  }
  sprintf( str_ip, "%d.%d.%d.%d", ( (char *)&ip )[0], ( (char *)&ip )[1],
                                  ( (char *)&ip )[2], ( (char *)&ip )[3] );
  return str_ip;
}
