/*
 *  payload.c  -  generate / parse all payloads
 *  
 *  EAP-IKEv2
 *
 *  This library implements ideas of draft-tschofenig-eap-ikev2-10.txt
 *  Internet-Draft. Version of the document that this library corresponds
 *  to you can find at
 *  http://tools.ietf.org/wg/eap/draft-tschofenig-eap-ikev2-10.txt.
 *
 *  This file is part of libeap-ikev2.
 *
 *  libeap-ikev2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  libeap-ikev2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libeap-ikev2; if not, write to the Free Software
 *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Copyright (C) 2005-2006 Krzysztof Rzecki <krzysztof.rzecki@ccns.pl>      
 *  Copyright (C) 2005-2006 Rafal Mijal <rafal.mijal@ccns.pl>                
 *  Copyright (C) 2005-2006 Piotr Marnik <piotr.marnik@ccns.pl>              
 *  Copyright (C) 2005-2006 Pawel Matejski <pawel.matejski@ccns.pl>          
 *  Copyright (C) 2003      Udo Schilcher <udo.schilcher@edu.uni-klu.ac.at>
 *  Copyright (C) 2003      Thomas Hambrusch <thambrus@edu.uni-klu.ac.at>          
 *
 */

#include"common.h"

#include <assert.h>
#ifdef HAVE_NETINET_IN_H
#include<netinet/in.h>
#endif //HAVE_NETINET_IN_H
#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif // HAVE_WINSOCK_H

#include<openssl/rand.h>
#include<string.h>

#include"payload.h"
#include"transforms.h"
#include"session.h"
#include"encr.h"
#include"message.h"
#include"auth.h"
#include"dh.h"
#include"conf_utils.h"
#include "enums.h"
#include "integ.h"

// private prototypes


static void fillFixedPayloadHdr(uint8_t **next, uint8_t type,uint8_t *p);

// Create payloads

/**
 * @brief Create an encrypted payload.
 *
 * This is the most complex payload creation function. The work is performed in
 * three main steps:
 * 	- encrypt given data,
 * 	- setup all remainig fields of IKEv2 message (hdr length, next payload
 * 	field),
 * 	- compute and attach integrity checksum of whole IKEv2 message.
 * This function is ALWAYS called as last when chain of payloads is builded. 	
 *
 * @param next       next payload type. 
 * @param inner_next value for 'NextPayload' field
 * @param *sigdata   ptr to the begining of the IKEv2 message
 * @param *data      payloads to be encrypted
 * @param dlen       length of data
 * @param *session   IKEv2 session data
 * @param *p         destination buffer 
 *
 * @return size of created payload (in bytes)
 *
 * @see For more details about 'next' param usage see CreateKEPayload  description.
 */
uint32_t CreateEncryptedPayload( uint8_t **next, uint8_t inner_next, uint8_t *data, uint32_t dlen, uint8_t *sigdata, 
    struct IKEv2Session *session, uint8_t *p )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Creating encrypted payload.\n" );
  uint8_t itid=GetTransformID(IKEv2_TRT_INTEGRITY_ALGORITHM,IKEv2_PID_IKE_SA, session->UsedTransforms);
  uint8_t etid=GetTransformID(IKEv2_TRT_ENCRYPTION_ALGORITHM,IKEv2_PID_IKE_SA,session->UsedTransforms);
  uint16_t ekeylen=GetTransformKeyLen(IKEv2_TRT_ENCRYPTION_ALGORITHM,IKEv2_PID_IKE_SA, session->UsedTransforms);
  uint32_t ilen=GetChecksumLen(itid);
  uint16_t pllen=0;
  
  uint8_t *iv;
  struct IKEv2PayloadHeader *ph=(struct IKEv2PayloadHeader*)p;
  fillFixedPayloadHdr(next,IKEv2_NPT_ENCRYPTED,p);
  ph->NextPayload=inner_next;

  uint32_t bs = GetEncrBlockLen(etid ,ekeylen);
  if( !bs )
    return 0;
  p += sizeof( struct IKEv2PayloadHeader );
  iv=p; // initialization vector
  RAND_bytes( iv, bs );

  p += bs;  // encrypt data
  pllen+=bs;
  uint32_t enclen = ( ( (int)( ( dlen + 1 ) / bs ) ) + 1 ) * bs; // len of data to be encrypted (incl. padding)
  uint8_t padlen = enclen - dlen - 1;  // length of padding (8 bit!!)
  memcpy( p, data, dlen );
  RAND_bytes( p + dlen, padlen );   // padding
  p[dlen + padlen] = padlen;        // len of padding
  //encrypt
  IKEv2Crypt( p, enclen, iv, session->Type == IKEv2_STY_INITIATOR ? session->SK_ei : session->SK_er,
      etid,ekeylen, IKEv2_ENCRYPT );
  
   
  p+=enclen;
  // set lengths
  struct IKEv2Header *hdr=(struct IKEv2Header*)sigdata;
  hdr->Length=htonl(p-sigdata+ilen);
  pllen=p-(uint8_t*)ph+ilen;
  ph->PayloadLength = htons(pllen);   // size of payload
  
  // attach chksum
  uint8_t *chksum;
  integ(session->Type==IKEv2_STY_INITIATOR ? session->SK_ai : session->SK_ar,
          GetIntegKeyLen(itid),sigdata,p-sigdata,&chksum,&ilen,itid);
  memcpy(p,chksum,ilen);
  free(chksum);
  
  
  return pllen;
}

/**
 * @brief Function creates an KE-Payload
 *
 * @param next      next payload type.
 *                  This variable is used as input/output data. On input *next must
 *                  contain ptr to the 'next' field of the previous payload.
 *                  After function execution *next will contain ptr to the
 *                  'next' field of current payload. This sollution is very
 *                  useful when creating a chain of payloads.
 * @param DHGroup   group ID for DH
 * @param *session  IKEv2 session data
 * @param *p        destination buffer
 *
 * @return length of created payload
 */
uint32_t CreateKEPayload( uint8_t **next, uint16_t DHGroup, struct IKEv2Session *session, uint8_t *p )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Creating KE payload.\n" );

  struct IKEv2KeyExchangePayload *ke = (struct IKEv2KeyExchangePayload *)p;

  fillFixedPayloadHdr(next,IKEv2_NPT_KE,p);
  ke->DHGroupNr = htons( DHGroup );
  ke->Reserved2 = 0;
  p += sizeof( struct IKEv2KeyExchangePayload );

  if( session->dh )
    DH_free( session->dh );
  DH *dh;
  if( !( dh = DHGenerateKEData( DHGroup ) ) )
  {
        xlogf( I2L_ERR, "EAP-IKEv2: Error: DHGenerateKEData()\n" );
        return 0;
  }
  session->dh = dh;
  uint16_t modsize=BN_num_bytes(dh->p);
  uint16_t size=BN_num_bytes(dh->pub_key);
  assert(modsize-size>=0);
  memset(p,0,modsize-size);
  BN_bn2bin(dh->pub_key,p+modsize-size);
  p += modsize;
  ke->Length = htons( p - (uint8_t *)ke );
  return p - (uint8_t *)ke;
}

/**
 * @brief Creates SA payload
 *
 * @param next      next payload type. 
 * @param *suppProp list of supported transforms
 * @param *p        destination buffer
 *
 * @return size of created payload (in bytes)
 *
 * @see For more details about 'next' param usage see CreateKEPayload  description.
 */
uint32_t CreateSAPayload( uint8_t **next, struct Proposal *suppProp, uint8_t *p )
{
  uint32_t i;
  struct IKEv2PayloadHeader *plhdr = (struct IKEv2PayloadHeader *)p;

  xlogf( I2L_DBG, "EAP-IKEv2: Creating SA payload.\n" );
  
  // create payload hdr
  fillFixedPayloadHdr(next,IKEv2_NPT_SA,p);

  p += sizeof( struct IKEv2PayloadHeader );
  struct IKEv2Proposal *pr = (struct IKEv2Proposal *)p;
  struct IKEv2Transform *t;
  struct IKEv2Attribute *att;

  for( i = 0; suppProp; i++ )     // proposals
  {
    struct Protocol *prot = suppProp->protocols;
    while( prot )
    {// first create proposal hdr
      pr->IsLast = suppProp->next ? 2 : 0;  // is last?
      pr->Reserved = 0;
      pr->ProposalLength = htons( sizeof( struct IKEv2Proposal ) + prot->SPISize );
      pr->ProposalNr = (uint8_t)( i + 1 );
      pr->ProtocolId = prot->protocol;
      pr->SPISize = prot->SPISize;
      p += sizeof( struct IKEv2Proposal );

      if( prot->SPISize == 8 )                // fill in SPI
        *( (uint64_t *)p ) = hton64( prot->SPI );
      if( prot->SPISize == 4 )
        *( (uint32_t *)p ) = htonl( (uint32_t)( prot->SPI & 0xFFFFFFFF ) );
      p += prot->SPISize;

      t = (struct IKEv2Transform *)p;

      struct Transform *trans = prot->transforms;
      uint8_t tcount = 0;
      while( trans )
      {
        tcount++;
        t->IsLast = trans->next ? 3 : 0;  // Encryption Transform hdr
        t->Reserved1 = 0;
        t->TransformLength = htons( sizeof( struct IKEv2Transform ) +
                           ( trans->keyLen ? sizeof( struct IKEv2Attribute ) : 0 ) );
        t->TransformType = trans->type;
        t->Reserved2 = 0;
        t->TransformID = htons( trans->id );
        p += sizeof( struct IKEv2Transform );
        pr->ProposalLength = htons( ntohs( pr->ProposalLength ) + sizeof( struct IKEv2Transform ) );
        if( trans->keyLen )    // key len attribute
        {
          att = (struct IKEv2Attribute *)p;
          att->TypeAF = htons( Attribute_GetTypeHF( IKEv2_ATT_KEY_LENGTH, 1 ) );
          att->Value = htons( trans->keyLen );
          p += sizeof( struct IKEv2Attribute );
          pr->ProposalLength = htons( ntohs( pr->ProposalLength ) + sizeof( struct IKEv2Attribute ) );
        }
        t = (struct IKEv2Transform *)p;
        trans = trans->next;
      }
      pr->TransformsCount = tcount;
      pr = (struct IKEv2Proposal *)p;
      prot = prot->next;
    }
    suppProp = suppProp->next;
  }

  plhdr->PayloadLength = htons( p - (uint8_t *)plhdr );
  return p - (uint8_t *)plhdr;
}

/**
 * @brief Creates a Nonce payload.
 *
 * As a side effect session->Ni (for initiator) or session->Nr (for responder)
 * field will be filled. Old value will be lost.
 *
 * @param next      next payload type. 
 * @param *s        current IKEv2 session data
 * @param len       length of nonce data 
 * @param *p        destination buffer
 *
 * @return size of created payload (in bytes)
 *
 * @see For more details about 'next' param usage see CreateKEPayload  description.
 */
uint32_t CreateNoncePayload( uint8_t **next, struct IKEv2Session *s, uint32_t len, uint8_t *p )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Creating Nonce payload.\n" );

  struct IKEv2PayloadHeader *plhdr = (struct IKEv2PayloadHeader *)p;
  
  // create payload hdr
  fillFixedPayloadHdr(next,IKEv2_NPT_NONCE,p);
  plhdr->PayloadLength = htons( sizeof( struct IKEv2PayloadHeader ) + len );
  
  p += sizeof( struct IKEv2PayloadHeader );

  RAND_bytes( p, len );        // generate nonce

  //fill correct session nonce value
  if(s->Type==IKEv2_STY_INITIATOR) {
      s->Ni=BN_bin2bn(p,len,s->Ni);
  } else {
      s->Nr=BN_bin2bn(p,len,s->Nr);
  }
  return sizeof( struct IKEv2PayloadHeader ) + len;
}

/**
 * @brief Creating IDR or IDI payload (depends on initiator param value)
 *
 * @param next      next payload type. 
 * @param initiator if true create IDI payload else IDR payload
 * @param type      type of ID
 * @param *id       ptr to ID data
 * @param idlen     ID data length
 * @param *p        destination buffer
 *
 * @return size of created payload (in bytes)
 *
 * @see For more details about 'next' param usage see CreateKEPayload  description.
 */
uint32_t CreateIDPayload( uint8_t **next, uint8_t initiator, uint8_t type, uint8_t *id, uint16_t idlen, uint8_t *p )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Creating ID payload.\n" );

  struct IKEv2IdentificationPayload *idp = (struct IKEv2IdentificationPayload *)p;
  if(initiator) {
      fillFixedPayloadHdr(next,IKEv2_NPT_IDI,p);
  } else {
      fillFixedPayloadHdr(next,IKEv2_NPT_IDR,p);
  }
  idp->IDType = type;
  idp->Reserved2 = 0;
  p += sizeof( struct IKEv2IdentificationPayload );
  switch( type )
  {
  case IKEv2_IDT_IPV4_ADDR:
    *( (uint32_t *)p ) = StringToIP( (char*)id );
    p += 4;
    break;

  case IKEv2_IDT_FQDN:          // only copy the id
  case IKEv2_IDT_RFC822_ADDR:
  case IKEv2_IDT_KEY_ID:
    memcpy( p, id, idlen );
    p += idlen;
    break;
  }
  
  idp->Length = htons( p - (uint8_t *)idp );
  return p - (uint8_t *)idp;
}

/**
 * @brief Creating an AUTH payload. 
 *
 * IKEv2 context date carry necessary info about used authentication type.
 *
 * @param *i2       IKEv2 context data
 * @param next      next payload type. 
 * @param *session  IKEv2 session data
 * @param *id       ptr to ID data
 * @param idlen     ID data length
 * @param *p        destination buffer
 *
 * @return size of created payload (in bytes)
 *
 * @see For more details about 'next' param usage see CreateKEPayload  description.
 */
uint32_t CreateAuthPayload( const struct ikev2_ctx *i2, uint8_t **next, struct IKEv2Session *session, uint8_t *id,
    uint16_t idlen, uint8_t *p )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Creating auth payload.\n" );
  int type;

  struct IKEv2AuthenticationPayload *ap = (struct IKEv2AuthenticationPayload *)p;
  type=i2->authtype;
  switch(i2->authtype) {
      case IKEv2_AUTH_SK:
	  type=IKEv2_AMT_SKMIC;
	  break;
      case IKEv2_AUTH_CERT:
	  type=GetCertAuthMethod(i2->rpk);
	  break;
  }

  // create header...
  fillFixedPayloadHdr(next,IKEv2_NPT_AUTH,p);
  ap->AuthMethod = type;
  ap->Reserved2 = 0;
  p += sizeof( struct IKEv2AuthenticationPayload );

  p += GenerateAuthData( i2, type, session, id, idlen, p );    // ...and auth data
  ap->Length = htons( p - (uint8_t *)ap );         // compute length of msg
  return p - (uint8_t *)ap;
}

/**
 * @brief Create an CERT payload
 *
 * @param *i2       IKEv2 context data
 * @param next      next payload type. 
 * @param type      type of certificate
 * @param *p        destination buffer
 *
 * @return size of created payload (in bytes)
 *
 * @see For more details about 'next' param usage see CreateKEPayload  description.
 */
uint32_t CreateCertPayload( const struct ikev2_ctx *i2, uint8_t **next, uint8_t type, uint8_t *p )
{
  unsigned char *cert;
  
  xlogf( I2L_DBG, "EAP-IKEv2: Creating cert payload.\n" );
  
  struct IKEv2PayloadHeader *pl = (struct IKEv2PayloadHeader *)p;
  fillFixedPayloadHdr(next,IKEv2_NPT_CERT,p);

  
  p += sizeof( struct IKEv2PayloadHeader );
  *( p++ ) = type;

  if( !i2->certlen ){
        xlogf( I2L_ERR, "EAP-IKEv2: Cert data not available.\n" );
        return 0;
  }
  memcpy( p, i2->cert, i2->certlen );
  cert=NULL;
  p += i2->certlen;

  pl->PayloadLength = htons( p - (uint8_t *)pl );         // compute length of msg
  return p - (uint8_t *)pl;
}


/**
 * @brief Creating an notify payload
 *
 * @param next      next payload type
 * @param type      type of notify payload  
 * @param *ndata    notify payload data
 * @param ndlen     ndata length
 * @param *p        destination buffer
 *
 * @return size of created payload (in bytes)
 *
 * @see For more details about 'next' param usage see CreateKEPayload  description.
 */
uint32_t CreateNotifyPayload( uint8_t **next, uint16_t type, uint8_t *ndata, uint16_t ndlen, uint8_t *p )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Creating notify payload.\n" );

  struct IKEv2NotifyPayload *np = (struct IKEv2NotifyPayload *)p;   
  // creating header
  fillFixedPayloadHdr(next,IKEv2_NPT_NOTIFY,p);
  np->SProtocolID = IKEv2_PID_IKE_SA;
  np->SPISize = 0;
  np->MsgType = htons( type );
  p += sizeof( struct IKEv2NotifyPayload );

  // create notification data, if there is any
  if( ndlen )
    memcpy( p, ndata, ndlen );  

  np->Length = htons( sizeof( struct IKEv2NotifyPayload ) + ndlen );     // compute length
  return sizeof( struct IKEv2NotifyPayload ) + ndlen;
}

/**
 * @brief Creating an DELETE payload
 *
 * @param next      next payload type
 * @param SProtocolID protocol for which sessions should be deleted
 * @param *SPIs     ptr to an array of SPIs
 * @param count     # of SPIs
 * @param *p        destination buffer
 *
 * @return size of created payload (in bytes)
 *
 * @see For more details about 'next' param usage see CreateKEPayload  description.
 */
uint32_t CreateDeletePayload( uint8_t **next, uint8_t SProtocolID,
    uint8_t *SPIs, uint16_t count, uint8_t *p )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Creating delete payload.\n" );

  int i;
  struct IKEv2DeletePayload *d = (struct IKEv2DeletePayload *)p;
  fillFixedPayloadHdr(next,IKEv2_NPT_DELETE,p);
  d->SProtocolID = SProtocolID;
  d->SPISize = SProtocolID == IKEv2_PID_IKE_SA ? 0 : 4;
  d->SPICount = htons( count );
  p += sizeof( struct IKEv2DeletePayload );
  for( i = 0; i < count; i++ )  // copy SPIs
  {
    if( SProtocolID == IKEv2_PID_IKE_SA )
    {
      *( (uint64_t *)p ) = hton64( ( (uint64_t *)SPIs )[i] );
      p += 8;
    }
    else
    {
      *( (uint32_t *)p ) = htonl( ( (uint32_t *)SPIs )[i] );
      p += 4;
    }
  }
  d->Length = htons( p - (uint8_t *)d );   // set length of payload
  return p - (uint8_t *)d;
}


/**
 * @brief Creates an Certification Request Payload (CERT REQ)
 * @param *i2       IKEv2 context data
 * @param next      next payload type
 * @param *p        destination buffer
 *
 * @return size of created payload (in bytes)
 *
 * @see For more details about 'next' param usage see CreateKEPayload  description.
 */
uint32_t CreateCertReqPayload(const struct ikev2_ctx *i2,uint8_t **next,uint8_t *p )
{
    xlogf(I2L_DBG,"EAP-IKEv2: Creating CERT REQ payload\n");
    assert(p);
    assert(i2 && i2->CAsHashStringLen>0);
    uint32_t len=0;
    struct IKEv2CertReqPayload *cr=(struct IKEv2CertReqPayload*)p;
    fillFixedPayloadHdr(next,IKEv2_NPT_CERTREQ,p);
    cr->Encoding=IKEv2_CET_X509_SIGNATURE;
    len=sizeof(struct IKEv2CertReqPayload);
    p+=sizeof(struct IKEv2CertReqPayload);
    memcpy(p,i2->CAsHashString,i2->CAsHashStringLen);
    len+=i2->CAsHashStringLen;
    cr->Length=htons(len);
    return len;
}

// Parse payloads

/**
 * @brief Parses an encryption payload (without integrity checksum!!)
 * @warning length field in header MUST be without integrity checksum!
 *
 * @param **data    decrypted payloads
 * @param *session  IKEv2 session data
 * @param *p        ptr to payload 
 *
 * @return length of decrypted data without padding and padlen field
 */
uint32_t ParseEncryptedPayload( uint8_t **data, struct IKEv2Session *session, uint8_t *p )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Parsing encrypted payload.\n" );

  struct IKEv2PayloadHeader *ph = (struct IKEv2PayloadHeader *)p; // parse header
  uint32_t bs = GetEncrBlockLen( GetTransformID( IKEv2_TRT_ENCRYPTION_ALGORITHM, IKEv2_PID_IKE_SA,
      session->UsedTransforms ),
      GetTransformKeyLen( IKEv2_TRT_ENCRYPTION_ALGORITHM, IKEv2_PID_IKE_SA, session->UsedTransforms ) );
  uint32_t enclen = ntohs( ph->PayloadLength ) - bs -
                     sizeof( struct IKEv2PayloadHeader );   // length of encrypted data
  p += sizeof( struct IKEv2PayloadHeader );
  uint8_t *iv = p;     // initialization vector
  p += bs;
  
  //decrypt
  IKEv2Crypt( p, enclen, iv, session->Type == IKEv2_STY_INITIATOR ? session->SK_er : session->SK_ei,
      GetTransformID( IKEv2_TRT_ENCRYPTION_ALGORITHM, IKEv2_PID_IKE_SA, session->UsedTransforms ),
      GetTransformKeyLen( IKEv2_TRT_ENCRYPTION_ALGORITHM, IKEv2_PID_IKE_SA, session->UsedTransforms ),
      IKEv2_DECRYPT );

  uint8_t padlen = p[enclen - 1];  // length of padding (8 bit!!)

  if( (uint32_t)( padlen + 1 ) >= enclen )
  {
    *data = NULL;
    xlogf( I2L_ERR, "EAP-IKEv2: Error: padlen has a wrong value, can't parse encrypted payload.\n" );
    return 0;
  }
  *data = Clone( p, enclen - padlen - 1 );  // return decrypted data without padding and padlen field
  return enclen - padlen - 1;
}

/**
 * @brief Parse SA payload
 *
 * @param *p        ptr to payload 
 *
 * @return Proposal structure which contains SA algorithms
 */
struct Proposal *ParseSAPayload( uint8_t *p )
{
  int i = 0, j, k;
  uint8_t *sp = p;
  uint64_t SPI;
  struct Proposal *prop = NULL, *ph=NULL;
  struct Protocol *prot;
  struct IKEv2PayloadHeader *pl = (struct IKEv2PayloadHeader *)p;
  
  xlogf( I2L_DBG, "EAP-IKEv2: Parsing SA payload.\n" );

  p += sizeof( struct IKEv2PayloadHeader );
  struct IKEv2Proposal *pr;

  for( ;; )  // protocols in proposals
  {
    pr = (struct IKEv2Proposal *)p;

    if( p + ntohs(pr->ProposalLength) > (uint8_t *)pl + ntohs(pl->PayloadLength) )
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Error: length fields inconsistent (1).\n" );
      DeleteProposalList( prop );
      return NULL;
    }
    
    if( pr->ProposalNr == i + 1 )   // new proposal?
    {
      i++;
      ph = AddProposal( &prop );
      if( !ph )
        return NULL;
    }
    
    if( !ph || pr->ProposalNr != i  )
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Error: ProposalNr wrong. Can't parse SA.\n" );
      return NULL;
    }
    if( pr->SPISize == 8 )
      SPI = ntoh64( *( (uint64_t *)p ) );
    else
      if( pr->SPISize == 4 )
        SPI = (uint64_t)ntohl( *( (uint32_t *)p ) );
      else
        SPI = 0;
    p += sizeof( struct IKEv2Proposal ) + pr->SPISize;
    prot = AddProtocol( ph, pr->ProtocolId, pr->SPISize, SPI );

    struct IKEv2Transform *tr;
    for( j = 1; j <= pr->TransformsCount; j++ )  // transforms
    {
      tr = (struct IKEv2Transform *)p;
      if( p + ntohs( tr->TransformLength ) > (uint8_t *)pr + ntohs(pr->ProposalLength) ) // check for inconsistent len
      {
        xlogf( I2L_ERR, "EAP-IKEv2: Error: length fields inconsistent (2).\n" );
        DeleteProposalList( prop );
        return NULL;
      }
      
      if( ( j < pr->TransformsCount ) ^ ( tr->IsLast == 3 ) )  // are IsLast and Transform# consistent?
      {
        xlogf( I2L_ERR, "EAP-IKEv2: Error: TransformCount and IsLast field are inconsistent. Can't parse SA.\n" );
        return NULL;
      }

      struct IKEv2Attribute *at;
      uint16_t KeyLen = 0;

      if( tr->TransformType != IKEv2_TRT_ENCRYPTION_ALGORITHM &&
          tr->TransformType != IKEv2_TRT_PSEUDO_RANDOM_FUNCTION &&
          tr->TransformType != IKEv2_TRT_INTEGRITY_ALGORITHM &&
          tr->TransformType != IKEv2_TRT_DIFFIE_HELLMAN_GROUP &&
          tr->TransformType != IKEv2_TRT_EXTENDED_SEQUENCE_NUMBERS )
      {
        xlogf( I2L_INFO, "EAP-IKEv2: Error: Unknown Transform type, ignoring...\n" );
        p += ntohs( tr->TransformLength );
        continue;
      }
      
      xlogf( I2L_DBG, "EAP-IKEv2: Got an transform: %s\n",
	      GetTransformName(tr->TransformType, ntohs( tr->TransformID )));
        
      at = (struct IKEv2Attribute *)( p + sizeof( struct IKEv2Transform ) );
      for( k = sizeof( struct IKEv2Transform ); k < ntohs( tr->TransformLength ); k += sizeof( struct IKEv2Attribute ) )
      {
        if( Attribute_GetType( ntohs( at->TypeAF ) ) == IKEv2_ATT_KEY_LENGTH )
        {
          if( Attribute_GetHF( ntohs( at->TypeAF ) ) )
          {
            if( ntohs( tr->TransformLength ) < k + sizeof( struct IKEv2Attribute ) )
            {
              xlogf( I2L_ERR, "EAP-IKEv2: Error: length fields inconsistent (3).\n" );
              DeleteProposalList( prop );
              return NULL;
            }
            KeyLen = ntohs( at->Value );
          }
          else
          { // wrong attribute format
            xlogf( I2L_ERR, "EAP-IKEv2: Got an wrong attribute format for KeyLength attribute... ignored\n" );
            k += ntohs( at->Length );
          }
        }
        else
        {       // ignore other attribute type
          xlogf( I2L_ERR, "EAP-IKEv2: Got an unknown attribute type... ignored\n" );
          if( !Attribute_GetHF( ntohs( at->TypeAF ) ) )
            k += ntohs( at->Length );
        }
        at = (struct IKEv2Attribute *)( (char *)at + sizeof( struct IKEv2Attribute ) +
              ( Attribute_GetHF( ntohs( at->TypeAF ) ) ? 0 : ntohs( at->Length ) ) );
      }
      if( !AddTransform( prot, tr->TransformType, ntohs( tr->TransformID ), KeyLen ) )
      {
        xlogf( I2L_INFO, "EAP-IKEv2: Key length attribute for transform or algorithm not valid..ignoring.\n" );
        KeyLen = 0;
      }

      if( ntohs( tr->TransformLength ) != k )
      {
        xlogf( I2L_ERR, "EAP-IKEv2: Error: length fields inconsistent (4).\n" );
        DeleteProposalList( prop );
        return NULL;
      }
      
      p += ntohs( tr->TransformLength );
    }

    if( ntohs( pr->ProposalLength ) != p - (uint8_t*)pr )
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Error: length fields inconsistent: %d %d.\n", ntohs( pr->ProposalLength ), p - (uint8_t *)pr );
      DeleteProposalList( prop );
      return NULL;
    }
    
    if( pr->IsLast != 2 )
    {
      if( ntohs( pl->PayloadLength ) != (uint16_t)( p - sp ) )
      {
        xlogf( I2L_ERR, "EAP-IKEv2: Error: length fields inconsistent (5).\n" );
        DeleteProposalList( prop );
        return NULL;
      }
      
      break;
    }
  }

  if( p - sp != ntohs( pl->PayloadLength ) )   // Check payloadhdr length field
  {
    DeleteProposalList( prop );
    xlogf( I2L_ERR, "EAP-IKEv2: Error: PayloadHeader length field not correct.\n" );
    return NULL;
  }
  return prop;
}

/**
 * @brief Parse KE Payload. Extract KE data an DH group information from
 * payload.
 *
 * @param *DHGroup  result ptr for DH group id
 * @param *p        ptr to payload 
 *
 * @return KE data
 */
BIGNUM *ParseKEPayload( uint16_t *DHGroup, uint8_t *p )
{
  BIGNUM *num; 
  struct IKEv2KeyExchangePayload *ke = (struct IKEv2KeyExchangePayload *)p;

  xlogf( I2L_DBG, "EAP-IKEv2: Parsing KE payload.\n" );

  if( DHGroup )            // return the DHGroup Nr, if desired
    *DHGroup = ntohs( ke->DHGroupNr );
  
  if( ! ( num = BN_bin2bn( p + sizeof( struct IKEv2KeyExchangePayload ),
      ntohs( ke->Length ) - sizeof( struct IKEv2KeyExchangePayload ), NULL ) ) )
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't convert KEPayload data to number." );
    return NULL;
  }
  return num;
}

/**
 * @brief Parse nonce payload and extract nonce data
 *
 * @param *p        ptr to payload 
 *
 * @return nonce data
 */
BIGNUM *ParseNoncePayload( uint8_t *p )
{
  BIGNUM *num;
  struct IKEv2PayloadHeader *ke = (struct IKEv2PayloadHeader *)p;

  xlogf( I2L_DBG, "EAP-IKEv2: Parsing Nonce payload.\n" );
  
  if( ! ( num = BN_bin2bn( p + sizeof( struct IKEv2PayloadHeader ),
      ntohs( ke->PayloadLength ) - sizeof( struct IKEv2PayloadHeader ), NULL ) ) )
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't convert nonce payload data to number." );
    return NULL;
  }
  return num;
}

/**
 * @brief Parse an ID payload
 *
 * @param **data    result ptr to ID data
 * @param *p        ptr to payload  
 *
 * @return length of payload body
 */
uint16_t ParseIDPayload( uint8_t **data, uint8_t *p )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Parsing identification payload.\n" );

  struct IKEv2PayloadHeader *idp = (struct IKEv2PayloadHeader*)p;

  uint16_t len = ntohs( idp->PayloadLength ) - sizeof( struct IKEv2PayloadHeader);
  if( !( *data = (uint8_t *)malloc( len ) ) )
  {
    xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't allocate memory for ID.\n" );
    *data = NULL;
    return 0;
  }
  memcpy( *data, p + sizeof( struct IKEv2PayloadHeader), len );
  return len;
}


/**
 * @brief Parses an AUTH payload
 *
 * @param *AuthMethod  result ptr for auth method
 * @param **data       result ptr for authentication data
 * @param *p           ptr to payload  
 *
 * @return length of data
 */
int ParseAuthPayload( uint8_t *AuthMethod, uint8_t **data, uint8_t *p )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Parsing auth payload.\n" );
  struct IKEv2AuthenticationPayload *ap = (struct IKEv2AuthenticationPayload *)p;
  if( ntohs( ap->Length ) <= sizeof( struct IKEv2AuthenticationPayload ) )   // no auth data?
    return 0;
  *AuthMethod = ap->AuthMethod;
  *data = Clone( p + sizeof( struct IKEv2AuthenticationPayload ), ntohs( ap->Length ) -
      sizeof( struct IKEv2AuthenticationPayload ) );
  return ntohs( ap->Length ) - sizeof( struct IKEv2AuthenticationPayload );
}

/**
 * @brief Parse certificate payload
 *
 * @param **cdata  result ptr for certificate data
 * @param *cdlen   result ptr for cdata length
 * @param *ctype   result ptr for certificate type
 * @param *p       ptr to payload  
 *
 * @return 0 on success, 1 otherwise
 */
int ParseCertPayload(uint8_t **cdata, uint32_t *cdlen, uint8_t *ctype, uint8_t *p )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Parsing cert payload.\n" );
  struct IKEv2PayloadHeader *pl = (struct IKEv2PayloadHeader *)p;
  if( ntohs( pl->PayloadLength ) <= sizeof( struct IKEv2PayloadHeader ) + 1 )   // no auth data?
    return 1;

  *cdlen = ntohs( pl->PayloadLength ) - sizeof( struct IKEv2PayloadHeader ) - 1;
        if( !( *cdata = (uint8_t*)malloc( *cdlen ) ) )
        {
                xlogf( I2L_ERR, "EAP-IKEv2: Error: Can't alloc mem for cert data.\n" );
                return 1;
        }
  *ctype=*(uint8_t*)(p+sizeof(struct IKEv2PayloadHeader));
  memcpy( *cdata, p + sizeof( struct IKEv2PayloadHeader ) + 1, *cdlen);

  return 0;
}

/**
 * @brief Parse notify payload
 *
 * @param *p       ptr to payload  
 * @param *data    result ptr for notification data
 *
 * @return 0 on success, 1 otherwise
 */
int ParseNotifyPayload( uint8_t *p, uint16_t *data )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Parsing notify payload.\n" );

  *data = 0;   // default
  struct IKEv2NotifyPayload *np = (struct IKEv2NotifyPayload *)p;
  if( ntohs( np->Length ) < sizeof( struct IKEv2NotifyPayload ) + np->SPISize )
  {
    xlogf( I2L_DBG, "EAP-IKEv2: Length field not correctly set.\n" );
    return 0;
  }
  if( np->SProtocolID != IKEv2_PID_IKE_SA || np->SPISize > 0 )
  {
    xlogf( I2L_DBG, "EAP-INFO: Unsupportet protocol in notify payload, ignoring...\n" );
    return 0;
  }
  p += sizeof( struct IKEv2NotifyPayload );

  switch( ntohs( np->MsgType ) )
  {
  case IKEv2_NMT_UNSUPPORTED_CRITICAL_PAYLOAD:
    if( ntohs( np->Length ) != np->SPISize + sizeof( struct IKEv2NotifyPayload ) + 1 )
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Length field not correctly set..\n" );
      return 0;
    }
    xlogf( I2L_ERR, "EAP-IKEv2: Error: got a 'unsupported critical payload: %d' error msg.\n", *p );
    break;

  case IKEv2_NMT_INVALID_IKE_SPI:
    xlogf( I2L_ERR, "EAP-IKEv2: Error: got a 'invalid ike spi' error msg.\n" );
    break;

  case IKEv2_NMT_INVALID_MAJOR_VERSION:
    xlogf( I2L_ERR, "EAP-IKEv2: Error: got a 'invalid major version' error msg.\n" );
    break;

  case IKEv2_NMT_INVALID_SYNTAX:
    xlogf( I2L_ERR, "EAP-IKEv2: Error: got a 'invalid syntax' error msg.\n" );
    break;

  case IKEv2_NMT_NO_PROPOSAL_CHOSEN:
    xlogf( I2L_ERR, "EAP-IKEv2: Error: got a 'no proposal chosen' error msg.\n" );
    break;

  case IKEv2_NMT_INVALID_KE_PAYLOAD:
    if( ntohs( np->Length ) != np->SPISize + sizeof( struct IKEv2NotifyPayload ) + 2 )
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Length field not correctly set..\n" );
      return 0;
    }
    xlogf( I2L_INFO, "EAP-IKEv2: Invalid KE payload, trying resend first msg.\n" );
    *data = ntohs( *( (uint16_t *)p ) );
    return 0;
    break;

  case IKEv2_NMT_AUTHENTICATION_FAILED:
    xlogf( I2L_ERR, "EAP-IKEv2: Error: authentication has failed.\n" );
    break;

  default:
    if( np->MsgType < 16383 )
      xlogf( I2L_ERR, "EAP-IKEv2: Got an unknown/unsupported notify msg with type id: 0x%x.\n",
          np->MsgType );
    break;
  }
  
  if( np->MsgType < 16383 )   // got an error msg?
    return 1;
  else
    return 0;
}

#if 0
/**
 * @brief Parse delete payload. After that delete IKEv2 session.
 *
 * @param *i2      IKEv2 context data
 * @param *p       ptr to payload  
 */

void ParseDeletePayload( const struct ikev2_ctx *i2, uint8_t *p )
{
  xlogf( I2L_DBG, "EAP-IKEv2: Parsing delete payload.\n" );

  int i;
  struct IKEv2Session *session;
  struct IKEv2DeletePayload *d = (struct IKEv2DeletePayload *)p;
  switch( d->SProtocolID )    // session for which protocol?
  {
  case IKEv2_PID_IKE_SA:
    if( d->SPISize != 0 || ntohs( d->Length ) != ntohs( d->SPICount ) * 8 ) // check format...
    {
      xlogf( I2L_ERR, "EAP-IKEv2: Got an delete payload with incorrect size or SPI size.\n" );
      return;
    }
    p += sizeof( struct IKEv2DeletePayload );
    for( i = 0; i < ntohs( d->SPICount ); i++ )      // delete sessions for given SPIs
    {
      session = FindSession( i2, ( (uint64_t *)p )[i] );
      if( session )                // if session not exists, ignore...
        DeleteSession( i2, session );
    }
    break;
  }
}
#endif

/**
 * @brief Parse Certificate Request Parse
 * @todo Current implementation only print payload data
 *
 * @param *i2      IKEv2 context data
 * @param *p       ptr to payload  
 */
void ParseCertReqPayload(const struct ikev2_ctx *i2, uint8_t *p)
{
    xlogf(I2L_DBG,"EAP-IKEv2: Parsing CERT REQ payload\n");
    struct IKEv2CertReqPayload *cr=(struct IKEv2CertReqPayload*)p;
    // testing implementation
    uint16_t len=ntohs(cr->Length);
    if(len<sizeof(struct IKEv2CertReqPayload)) {
	xlogf(I2L_ERR,"EAP-IKEv2: Packet length inconsistent\n");
	return;
    }
    xlogf(I2L_DBG,"EAP-IKEv2: Length=%d,Encoding=%d\n", len,cr->Encoding);
    len-=sizeof(struct IKEv2CertReqPayload);
    p+=sizeof(struct IKEv2CertReqPayload);
    if(len==i2->CAsHashStringLen && !memcmp(p,i2->CAsHashString,len)) {
	xlogf(I2L_DBG,"EAP-IKEv2: Request for certificate from our trusted CA\n");
    } else {
	xlogf(I2L_INFO,"EAP-IKEv2: Requset for certificate which could not be respected\n");
    }
    //PrintBuffer(p,len);
}

/**
 * @brief Auxiliary function used to set next field value of previous payload
 * and initialize current payload header.
 * 
 * @param **next *next must point to the previous payload 'next' field. After
 *                function execution *next will contain ptr to the current 
 *                payload 'next' * field.
 * @param type    payload type
 * @param *p      payload header in raw format
 */
static void fillFixedPayloadHdr(uint8_t **next, uint8_t type, uint8_t *p)
{
    assert(next && *next && p);
    struct IKEv2PayloadHeader *pl=(struct IKEv2PayloadHeader*)p;
    pl->NextPayload=IKEv2_NPT_NONE;
    pl->Reserved=0;
    pl->Reserved=0;
    **next=type;
    *next=&pl->NextPayload;
}
